#ifndef INCLUDED_ESDOM_API_hpp
#define INCLUDED_ESDOM_API_hpp

#include <ESDOM/ESDOMString.hpp>


namespace ESDOM
{

enum EncodingType{
	UTF_8			= 1,
	UTF_16BE		= 2,
	UTF_16LE		= 3,
	UTF_16BE_NONBOM	= 4,
	UTF_16LE_NONBOM	= 5
};

extern ESDOM_EXPORT ESDOMString	m_errorMessage;
extern const ESDOM_CHAR	ESDOM_NULL_STRING[2];

inline size_t strlen(const char *string)
{
	const char *ptr;
	for(ptr = string; *ptr; ptr++);
	return ptr-string;
}

inline size_t strlen(const ESDOM_CHAR *string)
{
	const ESDOM_CHAR *ptr;
	for(ptr = string; *ptr; ptr++);
	return ptr-string;
}

inline int strcmp(const ESDOM_CHAR *string1, const ESDOM_CHAR *string2)
{
	while(*string1 && *string1 == *string2){
		string1++;
		string2++;
	}
	return *string1 - *string2;
}

inline int strcmp(const char *string1, const ESDOM_CHAR *string2)
{
	while(*string1 && *string1 == *string2){
		string1++;
		string2++;
	}
	return *string1 - *string2;
}

inline int strcmp(const ESDOM_CHAR *string1, const char *string2)
{
	while(*string1 && *string1 == *string2){
		string1++;
		string2++;
	}
	return *string1 - *string2;
}

inline int strncmp(const ESDOM_CHAR *string1, const ESDOM_CHAR *string2, int n)
{
	for( ; n>0; n--, string1++, string2++){
		if( *string1 != *string2 )
			return *string1 - *string2;
	}
	return 0;
}

inline int strncmp(const char *string1, const ESDOM_CHAR *string2, int n)
{
	for( ; n>0; n--, string1++, string2++){
		if( *string1 != *string2 )
			return *string1 - *string2;
	}
	return 0;
}

inline int strncmp(const ESDOM_CHAR *string1, const char *string2, int n)
{
	for( ; n>0; n--, string1++, string2++){
		if( *string1 != *string2 )
			return *string1 - *string2;
	}
	return 0;
}

inline int strcasecmp(const ESDOM_CHAR *s1, const ESDOM_CHAR *s2)
{
	if( !s1 || !*s1 ){
		if( !s2 || !*s2 )
			return 0;
		return -1;
	}else{
		if( !s2 || !*s2 )
			return 1;
	}

	for( ; *s1; s1++, s2++ ){
		ESDOM_CHAR lch;
		ESDOM_CHAR rch;
		if( ( 'a' <= *s1 && *s1 <= 'z' )
			|| ( 0x03b1 <= *s1 && *s1 <= 0x03c9 )
//			|| ( 0x0430 <= *s1 && *s1 <= 0x044f ) )
			|| ( 0xff41 <= *s1 && *s1 <= 0xff5a ) )
			lch = (ESDOM_CHAR)(*s1 - 0x20);
//		else if()
//			lch = (ESDOM_CHAR)(*s1 - 0x20);
		else
			lch = *s1;

		if( ( 'a' <= *s2 && *s2 <= 'z' )
			|| ( 0x03b1 <= *s2 && *s2 <= 0x03c9 )
//			|| ( 0x0430 <= *s2 && *s2 <= 0x044f ) )
			|| ( 0xff41 <= *s2 && *s2 <= 0xff5a ) )
			rch = (ESDOM_CHAR)(*s2 - 0x20);
//		else if()
//			rch = (ESDOM_CHAR)(*s2 - 0x20);
		else
			rch = *s2;

		if( lch != rch )
			return lch - rch;
	}
	return *s1 - *s2;
}

inline int strcasecmp(const char *s1, const ESDOM_CHAR *s2)
{
	if( !s1 || !*s1 ){
		if( !s2 || !*s2 )
			return 0;
		return -1;
	}else{
		if( !s2 || !*s2 )
			return 1;
	}

	for( ; *s1; s1++, s2++ ){
		char lch;
		ESDOM_CHAR rch;
		if( 'a' <= *s1 && *s1 <= 'z' )
			lch = (char)(*s1 - 0x20);
		else
			lch = *s1;

		if( 'a' <= *s2 && *s2 <= 'z' )
			rch = (ESDOM_CHAR)(*s2 - 0x20);
		else
			rch = *s2;

		if( lch != rch )
			return lch - rch;
	}
	return *s1 - *s2;
}

inline int strcasecmp(const ESDOM_CHAR *s1, const char *s2)
{
	if( !s1 || !*s1 ){
		if( !s2 || !*s2 )
			return 0;
		return -1;
	}else{
		if( !s2 || !*s2 )
			return 1;
	}

	for( ; *s1; s1++, s2++ ){
		ESDOM_CHAR lch;
		char rch;
		if( 'a' <= *s1 && *s1 <= 'z' )
			lch = (ESDOM_CHAR)(*s1 - 0x20);
		else
			lch = *s1;

		if( 'a' <= *s2 && *s2 <= 'z' )
			rch = (char)(*s2 - 0x20);
		else
			rch = *s2;

		if( lch != rch )
			return lch - rch;
	}
	return *s1 - *s2;
}

inline ESDOM_CHAR * strcpy(ESDOM_CHAR *dst, const ESDOM_CHAR *src)
{
	ESDOM_CHAR *ret = dst;
	while(*src)
		*dst++ = *src++;
	*dst = 0;
	return ret;
}

inline ESDOM_CHAR * strncpy(ESDOM_CHAR *dst, const ESDOM_CHAR *src, int n)
{
	ESDOM_CHAR * const p_dst = dst;
	while(*src && n){
		*dst++ = *src++;
		n--;
	}
	while(n){
		*dst++ = '\0';
		n--;
	}
	return p_dst;
}

inline ESDOMString getLastError(void)
{
	return m_errorMessage;
}

inline void clearLastError(void)
{
	m_errorMessage = ESDOMString();
}


#ifdef ESDOM_DEBUG
ESDOM_EXPORT	unsigned long getTotalInstanceNum(void);
#endif

ESDOM_EXPORT	ESDOM_FILE	openFile(const char *filename, int mode);
ESDOM_EXPORT	ESDOM_FILE	openFile(const ESDOM_CHAR *filename, int mode);
ESDOM_EXPORT	bool		closeFile( ESDOM::ESDOM_FILE FileHandle );
ESDOM_EXPORT	bool		lockFile( ESDOM::ESDOM_FILE FileHandle, int mode );

ESDOM_EXPORT	bool initialize(unsigned long NodeListPageSize = 16,
								unsigned long NamedNodeMapHashTableSize = 31,
								unsigned long ESDOMStringPageSize = 32 );

class Document;

ESDOM_EXPORT Document createDocument(const ESDOM_CHAR *version,
									 const ESDOM_CHAR *rootTagName,
									 const ESDOM_CHAR *docutype = _ESDOM_TEXT(""));
ESDOM_EXPORT Document parse(ESDOM_FILE FileHandle);
ESDOM_EXPORT bool serialize(const Document &document, unsigned short encoding, ESDOM_FILE FileHandle);


}; /* namespace ESDOM */

#endif /* INCLUDED_ESDOM_API_hpp */
