#ifndef INCLUDED_ESDOM_Node_hpp
#define INCLUDED_ESDOM_Node_hpp

#include <ESDOM/ESDOMString.hpp>
#include <ESDOM/Interface.hpp>

#include <ESDOM/Element.hpp>
#include <ESDOM/Attr.hpp>
#include <ESDOM/Text.hpp>
#include <ESDOM/CDATASection.hpp>
#include <ESDOM/ProcessingInstruction.hpp>
#include <ESDOM/Comment.hpp>
#include <ESDOM/Document.hpp>
#include <ESDOM/DocumentType.hpp>

namespace ESDOM
{

class ESDOM_EXPORT NodeInst;
class ESDOM_EXPORT NodeList;
class ESDOM_EXPORT NamedNodeMap;

class ESDOM_EXPORT Node : public virtual Element, Attr, CDATASection,
 ProcessingInstruction, Comment, Document, DocumentType
{
public:
	enum ElementType
	{
		ELEMENT_NODE				= 1,
		ATTRIBUTE_NODE				= 2,
		TEXT_NODE					= 3,
		CDATA_SECTION_NODE			= 4,
		ENTITY_REFERENCE_NODE		= 5,
		ENTITY_NODE					= 6,
		PROCESSING_INSTRUCTION_NODE	= 7,
		COMMENT_NODE				= 8,
		DOCUMENT_NODE				= 9,
		DOCUMENT_TYPE_NODE			= 10,
		DOCUMENT_FRAGMENT_NODE		= 11,
		NOTATION_NODE				= 12
	};

	Node();
	Node(const Node &original);
	Node(const NodeInst *instance);
	virtual ~Node();

	Node &	operator = (const Node &node);

	// Element & Document
	NodeList		getElementsByTagName(const ESDOM_CHAR *name) const;

	// Element
	ESDOMString		getAttribute(const ESDOM_CHAR *name) const;
	void			setAttribute(const ESDOM_CHAR *name, const ESDOM_CHAR *value) throw(ESDOMException);
	void			removeAttribute(const ESDOM_CHAR *name) throw(ESDOMException);
	Attr			getAttributeNode(const ESDOM_CHAR *name) const;
	Attr			setAttributeNode(Attr newAttr) throw(ESDOMException);
	Attr			removeAttributeNode(Attr oldAttr) throw(ESDOMException);
	void			normalize(void);

	// Document
	DocumentType	getDoctype(void) const;
	bool			getStandalone(void) const;
	void			setStandalone(bool n);
	Element			getDocumentElement(void) const;
	Element					createElement(const ESDOM_CHAR *tagName) const throw(ESDOMException);
	Text					createTextNode(const ESDOM_CHAR *data) const;
	Comment					createComment(const ESDOM_CHAR *data) const;
	CDATASection			createCDATASection(const ESDOM_CHAR *data) const;
	ProcessingInstruction	createProcessingInstruction(const ESDOM_CHAR *target, const ESDOM_CHAR *data) const throw(ESDOMException);
	Attr					createAttribute(const ESDOM_CHAR *name) const throw(ESDOMException);
	Element			getElementById(const ESDOM_CHAR *id) const;
	ESDOMString		setDoctype(const ESDOM_CHAR *doctype);

protected:

	friend	class	NamedNodeMap;
	friend	class	NodeList;
	friend	class	NodeBase;
	};

}; /* namespace ESDOM */

#endif /* INCLUDED_ESDOM_Node_hpp */
