#ifndef INCLUDED_ESDOM__linux_hpp
#define INCLUDED_ESDOM__linux_hpp

#include <sys/types.h> 
#include <sys/stat.h>
#include <unistd.h>

namespace ESDOM
{

typedef int ESDOM_FILE;

inline bool ESDOM_isRegFile(ESDOM_FILE FileHandle)
{
	struct stat	buffer;
	fstat(FileHandle, &buffer);
	return S_ISREG(buffer.st_mode);
}

inline bool ESDOM_seek(ESDOM_FILE FileHandle,long offset, long whence)
{
	long w = 0;
	if( whence == FILE_SEEK_SET )
		w = SEEK_SET;
	else if( whence == FILE_SEEK_CUR )
		w = SEEK_CUR;
	else if( whence == FILE_SEEK_END )
		w = SEEK_END;
	if( lseek(FileHandle, offset, w) == offset -1 )
		return false;
	return true;
}

inline long ESDOM_tell(ESDOM_FILE FileHandle)
{
	return lseek(FileHandle, 0, SEEK_CUR);
}

inline bool ESDOM_truncate(ESDOM_FILE FileHandle)
{
	struct stat buffer;
	fstat(FileHandle, &buffer);
	if(!S_ISREG(buffer.st_mode))
		return false;
	return ftruncate(FileHandle,0) ? false : true;
}

inline unsigned long readFromFile(ESDOM_FILE FileHandle, unsigned char *buffer, unsigned long size)
{
	ssize_t result;
	unsigned long remain = size;
	while( remain )
	{
		result = read(FileHandle, buffer, remain);
		if( result == 0 || result == -1 )
			return 0;
		remain -= result;
	}
	return size;
}

inline unsigned long writeToFile(ESDOM_FILE FileHandle, const unsigned char *buffer, unsigned long size)
{
	ssize_t result;
	result = write(FileHandle, buffer, size);
	return ( result == -1 ) ? 0 : result;
}

}; /* namespace ESDOM */

#endif /* INCLUDED_ESDOM_linux_hpp */
