#include <ESDOM/ESDOM_define.hpp>
#include <ESDOM/Document.hpp>
#include "DocumentInst.hpp"
#include <ESDOM/Node.hpp>
#include <ESDOM/NodeList.hpp>
#include <ESDOM/Initializer.hpp>

using namespace ESDOM;

Document::Document()
{
}

Document::Document(const Document &original)
: NodeBase( original )
{
}

Document::Document(const DocumentInst *instance)
: NodeBase( instance )
{
}

Document::~Document()
{
}

Document & Document::operator = (const Document &node)
{
	Interface::operator = ( node );
	return *this;
}

DocumentType Document::getDoctype() const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return DocumentType( ((DocumentInst*)m_instance)->getDoctype() );
}

bool Document::getStandalone() const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return ((DocumentInst*)m_instance)->getStandalone();
}

void Document::setStandalone(bool n)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	((DocumentInst*)m_instance)->setStandalone(n);
}

Element Document::getDocumentElement() const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return Element( ((DocumentInst*)m_instance)->getDocumentElement() );
}

Element Document::createElement(const ESDOM_CHAR *tagName) const throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return Element( ((DocumentInst*)m_instance)->createElement(tagName) );
}

Text Document::createTextNode(const ESDOM_CHAR *data) const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return Text( ((DocumentInst*)m_instance)->createTextNode(data) );
}

Comment Document::createComment(const ESDOM_CHAR *data) const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return Comment( ((DocumentInst*)m_instance)->createComment(data) );
}

CDATASection Document::createCDATASection(const ESDOM_CHAR *data) const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return CDATASection( ((DocumentInst*)m_instance)->createCDATASection(data) );
}

ProcessingInstruction Document::createProcessingInstruction(const ESDOM_CHAR *target, const ESDOM_CHAR *data) const throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return ProcessingInstruction( ((DocumentInst*)m_instance)->createProcessingInstruction(target,data) );
}

Attr Document::createAttribute(const ESDOM_CHAR *name) const throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return Attr( ((DocumentInst*)m_instance)->createAttribute(name) );
}

NodeList Document::getElementsByTagName(const ESDOM_CHAR *name) const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return NodeList( ((DocumentInst*)m_instance)->getElementsByTagName(name) );
}

Element Document::getElementById(const ESDOM_CHAR *id) const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return Element( ((DocumentInst*)m_instance)->getElementById(id) );
}

ESDOMString Document::setDoctype(const ESDOM_CHAR *doctype)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return ((DocumentInst*)m_instance)->setDoctype(doctype);
}

//////////////////////////////////////////////////////////////////////////////////

Document Document::_createNewDocument(const ESDOM_CHAR *version, const ESDOM_CHAR *rootTagName, const ESDOM_CHAR *doctype)
{
	if( !Initializer::initialized() )
		throw ESDOMException(ESDOMException::UNINITIALIZED);

	DocumentInst		*document;
	try{
		document = new DocumentInst(); if(!document)throw 0;
	}catch(...){
		throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
	}

	try{
		if( version && *version )
			document->m_version = version;
		else
			document->m_version = "1.0";
		if( doctype && *doctype )
			document->setDoctype(doctype);

		document->_setDocumentElement(document->createElement(rootTagName));
		document->m_valid = true;
		return Document( document );
	}catch(ESDOMException e){
		document->removeRef();
		throw e;
	}
}
