#ifndef INCLUDED_ESDOM_DocumentInst_hpp
#define INCLUDED_ESDOM_DocumentInst_hpp

#include "NodeInst.hpp"
#include <ESDOM/Node.hpp>

namespace ESDOM
{

class ESDOM_EXPORT DocumentTypeInst;
class ESDOM_EXPORT ElementInst;
class ESDOM_EXPORT TextInst;
class ESDOM_EXPORT CommentInst;
class ESDOM_EXPORT CDATASectionInst;

class ESDOM_EXPORT DocumentInst : public NodeInst
{
public:
	explicit DocumentInst();
	~DocumentInst();

	NodeInst *			cloneNodeInst(bool deep) const;

	DocumentTypeInst *	getDoctype(void) const;
	bool				getStandalone(void) const;
	void				setStandalone(char n);
	ElementInst *		getDocumentElement(void) const;
	ElementInst	*				createElement(const ESDOM_CHAR *tagName) const;
	TextInst *					createTextNode(const ESDOM_CHAR *data) const;
	CommentInst *				createComment(const ESDOM_CHAR *data) const;
	CDATASectionInst *			createCDATASection(const ESDOM_CHAR *data) const;
	ProcessingInstructionInst *	createProcessingInstruction(const ESDOM_CHAR *target, const ESDOM_CHAR *data) const;
	AttrInst *					createAttribute(const ESDOM_CHAR *name) const;
	NodeListInst *		getElementsByTagName(const ESDOM_CHAR *name) const;
	ElementInst *		getElementById(const ESDOM_CHAR *id) const;
	ESDOMString			setDoctype(const ESDOM_CHAR *doctype);
	ESDOMString			getXmlVersion(void) const;
	bool				isValid(void) const;

	NodeInst *		insertBefore(NodeInst *newChild, NodeInst *refChild)
					{
						if( !newChild )return 0;
						_checkNewNodeType(newChild);
						return NodeInst::insertBefore(newChild,refChild);
					}
	NodeInst *		replaceChild(NodeInst *newChild, NodeInst *oldChild)
					{
						if( !newChild )return 0;
						_checkNewNodeType(newChild);
						return NodeInst::replaceChild(newChild,oldChild);
					}
	NodeInst *		appendChild(NodeInst *newChild)
					{
						if( !newChild )return 0;
						_checkNewNodeType(newChild);
						return NodeInst::appendChild(newChild);
					}

protected:
	explicit DocumentInst(const DocumentInst &original);

	ESDOMString			m_version;
	bool				m_valid;
	DocumentTypeInst*	m_doctype;
	ElementInst *		m_documentElement;
	char				m_standalone;
	void				_setDocumentElement(ElementInst* element);
	ElementInst *		_getElementById(const ElementInst *element, const ESDOM_CHAR *id) const;

	void	_checkNewNodeType(const NodeInst *newChild) const
			{
				unsigned short nodeType = newChild->getNodeType();
				if( nodeType == Node::ATTRIBUTE_NODE 
					|| nodeType == Node::TEXT_NODE 
					|| nodeType == Node::CDATA_SECTION_NODE 
					|| nodeType == Node::DOCUMENT_NODE 
					)
					throw ESDOMException(ESDOMException::HIERARCHY_REQUEST_ERR);
			}

	friend	class	Document;
	friend	class	Parser;
	friend	class	Serializer;
};

}; /* namespace ESDOM */

#endif /* INCLUDED_ESDOM_DocumentInst_hpp */
