#include <ESDOM/ESDOM_define.hpp>
#include <ESDOM/API.hpp>
#include <stdlib.h>
#include <string.h>
#include <ESDOM/ESDOMException.hpp>
#include <ESDOM/ESDOMString.hpp>
#include "ESDOMStringInst.hpp"

using namespace ESDOM;

#define EMPTY(string) (!(string).m_pInst || !(string).m_pInst->m_data || !*(string).m_pInst->m_data)

template<class _T_Char, class _T_Num>
inline unsigned make_form_num(_T_Num n, _T_Char * &str, unsigned add = 0){
	_T_Num buf = n;
	int length = 0;
	if(n < 0)
		length++;
	do{
		buf /= 10;
		length++;
	} while(buf);
	str = new _T_Char[length +1 +add];
	if(!str)
		throw 0;

	_T_Char * ptr = str + length;
	*ptr-- = '\0';
	if(n < 0){
		do{
			*ptr-- = (_T_Char)( '0' - n % 10 );
			n /= 10;
		} while(n);
		*ptr-- = '-';
	}else{
		do{
			*ptr-- = (_T_Char)( '0' + n % 10 );
			n /= 10;
		} while(n);
	}
	return strlen(ptr+1);
}

unsigned short		ESDOMString::FIGURE_NUMBER = 0;
const ESDOM_CHAR	ESDOMString::ESDOM_NULL_String[2] = {0,0};

ESDOMString::ESDOMString()
{
	m_pInst = 0;
}

ESDOMString::ESDOMString( const ESDOM_CHAR *string )
{
	make_from_ESDOM_CHAR( string );
}

ESDOMString::ESDOMString( const char *string )
{
	make_from_char( string );
}

ESDOMString::ESDOMString( ESDOM_CHAR ch )
{
	ESDOM_CHAR	string[2] = { ch, 0 };
	make_from_ESDOM_CHAR( string );
}

ESDOMString::ESDOMString( int n )
{
	make_from_int(n);
}

ESDOMString::ESDOMString( unsigned n )
{
	make_from_uint(n);
}

ESDOMString::ESDOMString( const ESDOMString &original )
{
	m_pInst = original.m_pInst;
	if( m_pInst )
		m_pInst->addRef();
}

ESDOMString::~ESDOMString()
{
	if( m_pInst )
		m_pInst->removeRef();
}

ESDOMString::operator const ESDOM_CHAR*() const
{
	if( EMPTY(*this) )
		return ESDOM_NULL_String;
	return m_pInst->m_data;
}

ESDOMString::operator bool() const
{
	return !EMPTY(*this);
}

ESDOMString & ESDOMString::operator = (const ESDOMString &string)
{
	if( &string == this )
		return *this;

	if( m_pInst )
		m_pInst->removeRef();

	m_pInst = string.m_pInst;

	if( m_pInst )
		m_pInst->addRef();

	return *this;
}

ESDOMString & ESDOMString::operator = (const ESDOM_CHAR * const string)
{
	if(!*string){
		if( m_pInst )
			m_pInst->removeRef();
		m_pInst = 0;
		return *this;
	}
	if(!m_pInst){
		make_from_ESDOM_CHAR(string);
		return *this;
	}

	if(m_pInst->m_data <= string  && string <= m_pInst->m_data + m_pInst->m_length){
		ESDOM_CHAR* str;
		try{
			str = new ESDOM_CHAR[strlen(string)+1];
			if(!str)
				throw 0;
		}catch(...){
			throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
		}
		strcpy(str, string);
		if( m_pInst ){
			m_pInst->removeRef();
			m_pInst = 0;
		}
		try{
			m_pInst = new ESDOMStringInst( str );if(!m_pInst)throw 0;
		}catch(...){
			delete[] str;
			throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
		}
		m_pInst->addRef();
		delete[] str;
		return *this;
	}

	if( m_pInst ){
		m_pInst->removeRef();
		m_pInst = 0;
	}

	make_from_ESDOM_CHAR( string );

	return *this;
}

ESDOMString & ESDOMString::operator = (const char * const string)
{
	if( m_pInst ){
		m_pInst->removeRef();
		m_pInst = 0;
	}

	make_from_char( string );

	return *this;
}

ESDOMString & ESDOMString::operator = (ESDOM_CHAR ch)
{
	if( m_pInst ){
		m_pInst->removeRef();
		m_pInst = 0;
	}

	ESDOM_CHAR	string[2] = { ch, 0 };
	make_from_ESDOM_CHAR( string );

	return *this;
}

ESDOMString & ESDOMString::operator = (int n)
{
	if( m_pInst ){
		m_pInst->removeRef();
		m_pInst = 0;
	}

	make_from_int(n);

	return *this;
}

ESDOMString & ESDOMString::operator = (unsigned n)
{
	if( m_pInst ){
		m_pInst->removeRef();
		m_pInst = 0;
	}

	make_from_uint(n);

	return *this;
}

bool ESDOMString::operator == (const ESDOMString &string) const
{
	if( string.m_pInst == m_pInst )
		return true;

	if(EMPTY(*this)){
		if(EMPTY(string))
			return true;
		else
			return false;
	}else{
		if(EMPTY(string))
			return false;
	}

	return !strcmp(m_pInst->m_data, string.m_pInst->m_data);
}

bool ESDOMString::operator == (const ESDOM_CHAR *string) const
{
	if(EMPTY(*this))
		return !strcmp(ESDOM_NULL_String, string);

	return !strcmp(m_pInst->m_data, string);
}

bool ESDOMString::operator == (const char *string) const
{
	if(EMPTY(*this))
		return !strcmp(ESDOM_NULL_String, string);

	return !strcmp(m_pInst->m_data, string);
}

bool ESDOMString::operator == (ESDOM_CHAR ch) const
{
	if(EMPTY(*this))
		return !ch;

	if( *m_pInst->m_data == ch && *(m_pInst->m_data+1) == 0 )
		return true;
	return false;
}

bool ESDOMString::operator == (int n) const
{
	if(EMPTY(*this))
		return false;
	unsigned 	absolute = abs(n);
	const ESDOM_CHAR	*ptr = m_pInst->m_data + m_pInst->m_length-1;

	do{
		if( *ptr != (ESDOM_CHAR)(absolute % 10 + '0') )
			return false;
		absolute /= 10;
		ptr--;
	}while( absolute && ptr >= m_pInst->m_data );

	if( n < 0 ){
		if( *ptr != '-' )
			return false;
		ptr--;
	}
	if( absolute || ptr >= m_pInst->m_data )
		return false;

	return true;
}

bool ESDOMString::operator == (unsigned n) const
{
	if(EMPTY(*this))
		return false;
	const ESDOM_CHAR	*ptr = m_pInst->m_data + m_pInst->m_length-1;

	do{
		if( *ptr != (ESDOM_CHAR)(n % 10 + '0') )
			return false;
		n /= 10;
		ptr--;
	}while( n && ptr >= m_pInst->m_data );

	if( n || ptr >= m_pInst->m_data )
		return false;

	return true;
}

bool ESDOM_EXPORT ESDOM::operator == (const ESDOM_CHAR *ls, const ESDOMString &rs) 
{
	if(EMPTY(rs))
		return !strcmp(ls, ESDOM_NULL_STRING);

	return !strcmp(ls, rs.m_pInst->m_data);
}

bool ESDOM_EXPORT ESDOM::operator == (const char *ls, const ESDOMString &rs) 
{
	if(EMPTY(rs))
		return !strcmp(ls, ESDOM_NULL_STRING);

	return !strcmp(ls, rs.m_pInst->m_data);
}

bool ESDOM_EXPORT ESDOM::operator == (ESDOM_CHAR lch, const ESDOMString &rs) 
{
	if(EMPTY(rs))
		return !lch;

	if( *rs.m_pInst->m_data == lch && *(rs.m_pInst->m_data+1) == 0 )
		return true;
	return false;
}

bool ESDOM_EXPORT ESDOM::operator == (int n, const ESDOMString &rs)
{
	if(EMPTY(rs))
		return false;

	unsigned 	absolute = abs(n);
	const ESDOM_CHAR	*ptr = rs.m_pInst->m_data + rs.getLength() -1;

	do{
		if( *ptr != (ESDOM_CHAR)(absolute % 10 + '0') )
			return false;
		absolute /= 10;
		ptr--;
	}while( absolute && ptr >= (const ESDOM_CHAR*)rs );

	if( n < 0 ){
		if( *ptr != '-' )
			return false;
		ptr--;
	}
	if( absolute || ptr >= (const ESDOM_CHAR*)rs )
		return false;

	return true;
}

bool ESDOM_EXPORT ESDOM::operator == (unsigned n, const ESDOMString &rs)
{
	if(EMPTY(rs))
		return false;

	const ESDOM_CHAR	*ptr = rs.m_pInst->m_data + rs.getLength() -1;

	do{
		if( *ptr != (ESDOM_CHAR)(n % 10 + '0') )
			return false;
		n /= 10;
		ptr--;
	}while( n && ptr >= (const ESDOM_CHAR*)rs );

	if( n || ptr >= (const ESDOM_CHAR*)rs )
		return false;

	return true;
}

bool ESDOMString::operator != (const ESDOMString &string) const
{
	if( string.m_pInst == m_pInst )
		return false;

	if(EMPTY(*this)){
		if(EMPTY(string))
			return false;
		else
			return true;
	}else{
		if(EMPTY(string))
			return true;
	}
	return strcmp(m_pInst->m_data, string.m_pInst->m_data) ? true : false;
}

bool ESDOMString::operator != (const ESDOM_CHAR *string) const
{
	if(EMPTY(*this))
		return strcmp(ESDOM_NULL_String, string) ? true : false;

	return strcmp(m_pInst->m_data, string) ? true : false;
}

bool ESDOMString::operator != (const char *string) const
{
	if(EMPTY(*this))
		return strcmp(ESDOM_NULL_String, string) ? true : false;

	return strcmp(m_pInst->m_data, string) ? true : false;
}

bool ESDOMString::operator != (ESDOM_CHAR ch) const
{
	if(EMPTY(*this))
		return ch ? true : false;

	if( *m_pInst->m_data == ch && *(m_pInst->m_data+1) == 0 )
		return false;
	return true;
}

bool ESDOMString::operator != (int n) const
{
	if(EMPTY(*this))
		return true;
	unsigned	absolute = abs(n);
	const ESDOM_CHAR	*ptr = m_pInst->m_data + m_pInst->m_length-1;

	do{
		if( *ptr != absolute % 10 + '0' )
			return true;
		absolute /= 10;
		ptr--;
	}while( absolute && ptr <= m_pInst->m_data );

	if( n < 0 ){
		if( *ptr != '-' )
			return true;
		ptr--;
	}
	if( absolute || ptr <= m_pInst->m_data )
		return true;

	return false;
}

bool ESDOMString::operator != (unsigned n) const
{
	if(EMPTY(*this))
		return true;
	const ESDOM_CHAR	*ptr = m_pInst->m_data + m_pInst->m_length-1;

	do{
		if( *ptr != (ESDOM_CHAR)(n % 10 + '0') )
			return true;
		n /= 10;
		ptr--;
	}while( n && ptr <= m_pInst->m_data );

	if( n || ptr <= m_pInst->m_data )
		return true;

	return false;
}

bool ESDOM_EXPORT ESDOM::operator != (const ESDOM_CHAR *ls, const ESDOMString &rs)
{
	if(EMPTY(rs))
		return strcmp(ls, ESDOM_NULL_STRING) ? true : false;

	return strcmp(ls, rs.m_pInst->m_data) ? true : false;
}

bool ESDOM_EXPORT ESDOM::operator != (const char *ls, const ESDOMString &rs)
{
	if(EMPTY(rs))
		return strcmp(ls, ESDOM_NULL_STRING) ? true : false;

	return strcmp(ls, rs.m_pInst->m_data) ? true : false;
}

bool ESDOM_EXPORT ESDOM::operator != (ESDOM_CHAR lch, const ESDOMString &rs)
{
	if(EMPTY(rs))
		return lch ? true : false;

	if( *rs.m_pInst->m_data == lch && *(rs.m_pInst->m_data+1) == 0 )
		return false;
	return true;
}

bool ESDOM_EXPORT ESDOM::operator != (int n, const ESDOMString &rs)
{
	if(EMPTY(rs))
		return true;

	unsigned 	absolute = abs(n);
	const ESDOM_CHAR	*ptr = rs.m_pInst->m_data + rs.getLength() -1;

	do{
		if( *ptr != (ESDOM_CHAR)(absolute % 10 + '0') )
			return true;
		absolute /= 10;
		ptr--;
	}while( absolute && ptr <= (const ESDOM_CHAR*)rs );

	if( n < 0 ){
		if( *ptr != '-' )
			return true;
		ptr--;
	}
	if( absolute || ptr <= (const ESDOM_CHAR*)rs )
		return true;

	return false;
}

bool ESDOM_EXPORT ESDOM::operator != (unsigned n, const ESDOMString &rs)
{
	if(EMPTY(rs))
		return true;
	const ESDOM_CHAR	*ptr = rs.m_pInst->m_data + rs.getLength() -1;

	do{
		if( *ptr != (ESDOM_CHAR)(n % 10 + '0') )
			return true;
		n /= 10;
		ptr--;
	}while( n && ptr <= (const ESDOM_CHAR*)rs );

	if( n || ptr <= (const ESDOM_CHAR*)rs )
		return true;

	return false;
}

ESDOMString & ESDOMString::operator += (const ESDOMString &string)
{
	if(EMPTY(string))
		return *this;

	if(!m_pInst){
		m_pInst = string.m_pInst;
		if( m_pInst )
			m_pInst->addRef();
		return *this;
	}
	if(this == &string){
		ESDOMString str(string);
		return *this += str;
	}
	if(!m_pInst->m_data || !*m_pInst->m_data){
		m_pInst->removeRef();
		m_pInst = string.m_pInst;
		if( m_pInst )
			m_pInst->addRef();
		return *this;
	}

	if( m_pInst->getRefCount() == 1 ){
		*m_pInst += string.m_pInst->m_data;
		return *this;
	}

	unsigned long length = m_pInst->m_length + string.m_pInst->m_length;

	ESDOMStringInst *newInst;
	try{ newInst = new ESDOMStringInst(length);if(!newInst)throw 0; }
	catch(...){ throw ESDOMException( ESDOMException::OUT_OF_MEMORY ); }

	memcpy( newInst->m_data, m_pInst->m_data, sizeof(ESDOM_CHAR)*m_pInst->m_length );
	memcpy( newInst->m_data+m_pInst->m_length, string.m_pInst->m_data, sizeof(ESDOM_CHAR)*string.m_pInst->m_length );
	newInst->m_data[length] = 0;
	m_pInst->removeRef();
	m_pInst = newInst;
	m_pInst->addRef();

	return *this;
}

ESDOMString & ESDOMString::operator += (const ESDOM_CHAR *string)
{
	if(!*string)
		return *this;

	if(!m_pInst){
		make_from_ESDOM_CHAR( string );
		return *this;
	}

	if(m_pInst->m_data <= string  && string <= m_pInst->m_data + m_pInst->m_length ){
		ESDOMString str(string);
		return *this += str;
	}

	if(!m_pInst->m_data || !*m_pInst->m_data){
		m_pInst->removeRef();
		make_from_ESDOM_CHAR(string);
		return *this;
	}

	if( m_pInst->getRefCount() == 1 ){
		*m_pInst += string;
		return *this;
	}

	unsigned long rsLength = strlen( string );
	unsigned long length = m_pInst->m_length + rsLength;

	ESDOMStringInst *newInst;
	try{ newInst = new ESDOMStringInst(length);if(!newInst)throw 0; }
	catch(...){ throw ESDOMException( ESDOMException::OUT_OF_MEMORY ); }
	memcpy( newInst->m_data, m_pInst->m_data, sizeof(ESDOM_CHAR)*m_pInst->m_length );
	memcpy( newInst->m_data + m_pInst->m_length, string, sizeof(ESDOM_CHAR)*rsLength );
	newInst->m_data[length] = 0;

	m_pInst->removeRef();
	m_pInst = newInst;
	m_pInst->addRef();

	return *this;
}

ESDOMString & ESDOMString::operator += (const char *string)
{
	if(!*string)
		return *this;
	if(!m_pInst){
		make_from_char(string);
		return *this;
	}
	if( !m_pInst->m_data || !*m_pInst->m_data ){
		m_pInst->removeRef();
		make_from_char(string);
		return *this;
	}

	if( m_pInst->getRefCount() == 1 ){
		*m_pInst += string;
		return *this;
	}

	unsigned long rsLength = ::strlen( string );
	unsigned long length = m_pInst->m_length + rsLength;

	ESDOMStringInst *newInst;
	try{ newInst = new ESDOMStringInst(length);if(!newInst)throw 0; }
	catch(...){ throw ESDOMException( ESDOMException::OUT_OF_MEMORY ); }
	memcpy( newInst->m_data, m_pInst->m_data, sizeof(ESDOM_CHAR)*m_pInst->m_length );
	for(ESDOM_CHAR *ptr = newInst->m_data + m_pInst->m_length; *string; string++, ptr++)
		*ptr = *string;
	newInst->m_data[length] = 0;

	m_pInst->removeRef();
	m_pInst = newInst;
	m_pInst->addRef();

	return *this;
}

ESDOMString & ESDOMString::operator += (ESDOM_CHAR ch)
{
	if( !m_pInst ){
		ESDOM_CHAR string[2] = {ch, 0};
		make_from_ESDOM_CHAR( string );
		return *this;
	}
	if( !m_pInst->m_data || !*m_pInst->m_data ){
		ESDOM_CHAR string[2] = {ch, 0};
		m_pInst->removeRef();
		make_from_ESDOM_CHAR( string );
		return *this;
	}

	if( m_pInst->getRefCount() == 1 ){
		*m_pInst += ch;
		return *this;
	}

	unsigned long length = m_pInst->m_length + 1;
	ESDOMStringInst *newInst;
	try{ newInst = new ESDOMStringInst(length);if(!newInst)throw 0; }
	catch(...){ throw ESDOMException( ESDOMException::OUT_OF_MEMORY ); }
	memcpy( newInst->m_data, m_pInst->m_data, sizeof(ESDOM_CHAR)*m_pInst->m_length );
	newInst->m_data[length-1] = ch;
	newInst->m_data[length] = 0;

	m_pInst->removeRef();
	m_pInst = newInst;
	m_pInst->addRef();

	return *this;
}

ESDOMString & ESDOMString::operator += (int n)
{
	if(!m_pInst){
		make_from_int(n);
		return *this;
	}
	if(!m_pInst->m_data || !*m_pInst->m_data){
		m_pInst->removeRef();
		make_from_int(n);
		return *this;
	}

	ESDOM_CHAR	*string;
	unsigned len_n;
	try	{
		len_n = make_form_num(n, string);
	}catch(...){
		throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
	}

	if( m_pInst->getRefCount() == 1 )
	{
		try{
			(*m_pInst)+= string;
		}catch(...){
			delete[] string;
			throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
		}
		delete[] string;
		return *this;
	}

	unsigned long length = m_pInst->m_length + len_n;
	ESDOMStringInst *newInst;
	try{
		newInst = new ESDOMStringInst(length);if(!newInst)throw 0;
	}catch(...){
		delete[] string;
		throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
	}
	memcpy( newInst->m_data, m_pInst->m_data, sizeof(ESDOM_CHAR)*m_pInst->m_length );
	memcpy( newInst->m_data + m_pInst->m_length, string, sizeof(ESDOM_CHAR)*(len_n +1) );

	m_pInst->removeRef();
	m_pInst = newInst;
	m_pInst->addRef();

	delete[] string;
	return *this;
}

ESDOMString & ESDOMString::operator += (unsigned n)
{
	if(!m_pInst){
		make_from_uint( n );
		return *this;
	}
	if(!m_pInst->m_data || !*m_pInst->m_data){
		m_pInst->removeRef();
		make_from_uint(n);
		return *this;
	}

	ESDOM_CHAR	*string;
	unsigned len_n;
	try	{
		len_n = make_form_num(n, string);
	}catch(...){
		throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
	}

	if( m_pInst->getRefCount() == 1 )
	{
		try{
			(*m_pInst)+= string;
		}catch(...){
			delete[] string;
			throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
		}
		delete[] string;
		return *this;
	}

	unsigned long length = m_pInst->m_length + len_n;
	ESDOMStringInst *newInst;
	try{
		newInst = new ESDOMStringInst(length);if(!newInst)throw 0;
	}catch(...){
		delete[] string;
		throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
	}
	memcpy( newInst->m_data, m_pInst->m_data, sizeof(ESDOM_CHAR)*m_pInst->m_length );
	memcpy( newInst->m_data + m_pInst->m_length, string, sizeof(ESDOM_CHAR)*(len_n +1) );

	m_pInst->removeRef();
	m_pInst = newInst;
	m_pInst->addRef();

	delete[] string;
	return *this;
}

ESDOMString ESDOM_EXPORT ESDOM::operator + (const ESDOMString &lstring, const ESDOMString &rstring)
{
	ESDOMString returnString( lstring );

	returnString += rstring;
	return returnString;
}

ESDOMString ESDOM_EXPORT ESDOM::operator + (const ESDOMString &lstring, const ESDOM_CHAR *rstring)
{
	ESDOMString returnString( lstring );

	returnString += rstring;
	return returnString;
}

ESDOMString ESDOM_EXPORT ESDOM::operator + (const ESDOMString &lstring, const char *rstring)
{
	ESDOMString returnString( lstring );

	returnString += rstring;
	return returnString;
}

ESDOMString ESDOM_EXPORT ESDOM::operator + (const ESDOMString &lstring, ESDOM_CHAR rchar)
{
	ESDOM_CHAR buf[2] = {rchar, 0};
	ESDOMString returnString( lstring );

	returnString += buf;
	return returnString;
}

ESDOMString ESDOM_EXPORT ESDOM::operator + (const ESDOMString &lstring, int rn)
{
	ESDOMString returnString( lstring );

	returnString += rn;
	return returnString;
}

ESDOMString ESDOM_EXPORT ESDOM::operator + (const ESDOMString &lstring, unsigned rn)
{
	ESDOMString returnString( lstring );

	returnString += rn;
	return returnString;
}

ESDOMString ESDOM_EXPORT ESDOM::operator + (const ESDOM_CHAR *lstring, const ESDOMString &rstring)
{
	ESDOMString returnString( lstring );

	returnString += rstring;
	return returnString;
}

ESDOMString ESDOM_EXPORT ESDOM::operator + (const char *lstring, const ESDOMString &rstring)
{
	ESDOMString returnString( lstring );

	returnString += rstring;
	return returnString;
}

ESDOMString ESDOM_EXPORT ESDOM::operator + (ESDOM_CHAR lch, const ESDOMString &rstring)
{
	ESDOM_CHAR buf[2] = {lch, 0};
	ESDOMString returnString( buf );

	returnString += rstring;
	return returnString;
}

ESDOMString ESDOM_EXPORT ESDOM::operator + (int ln, const ESDOMString &rstring)
{
	ESDOMString returnString( ln );

	returnString += rstring;
	return returnString;
}

ESDOMString ESDOM_EXPORT ESDOM::operator + (unsigned ln, const ESDOMString &rstring)
{
	ESDOMString returnString( ln );

	returnString += rstring;
	return returnString;
}

unsigned long ESDOMString::getLength() const
{
	if( m_pInst )
		return m_pInst->m_length;
	return 0;
}

int ESDOMString::Dec_to_Int() const
{
	if( !m_pInst || !m_pInst->m_data )
		return 0;
	const ESDOM_CHAR *ptr = m_pInst->m_data;
	bool	minus = false;
	int		n = 0;
	if( *ptr == '-' )
	{
		minus = true;
		ptr++;
	}
	while( *ptr )
	{
		if( *ptr < '0' || '9' < *ptr )
			return 0;
		n = n*10 + *ptr++ - '0';
	}
	if( minus )
		return -n;
	return n;
}

unsigned ESDOMString::Dec_to_UInt() const
{
	if( !m_pInst || !m_pInst->m_data )
		return 0;
	const ESDOM_CHAR *ptr = m_pInst->m_data;
	unsigned n = 0;
	while( *ptr )
	{
		if( *ptr < '0' || '9' < *ptr )
			return 0;
		n = n*10 + *ptr++ - '0';
	}
	return n;
}

unsigned ESDOMString::Hex_to_UInt() const
{
	if( !m_pInst || !m_pInst->m_data )
		return 0;
	const ESDOM_CHAR *ptr = m_pInst->m_data;
	unsigned n = 0;
	while( *ptr )
	{
		if( '0' <= *ptr && *ptr <= '9' )
			n = n*16 + *ptr++ - '0';
		else if( 'A' <= *ptr && *ptr <= 'F' )
			n = n*16 + *ptr++ - 'A' + 10;
		else if( 'a' <= *ptr && *ptr <= 'f' )
			n = n*16 + *ptr++ - 'a' + 10;
		else
			return 0;
	}
	return n;
}

unsigned long ESDOMString::index( const ESDOMString &string ) const
{
	if(&string == this)
		return 0;

	if( !m_pInst || !string.m_pInst || !m_pInst->m_length < !string.m_pInst->m_length
		|| !string.m_pInst->m_length )
		return (unsigned long)-1;

	unsigned long indexSuper;
	unsigned long indexSub;
	unsigned long length = m_pInst->m_length - string.m_pInst->m_length;
	for( indexSuper = 0; indexSuper <= length ; indexSuper++ )
	{
		indexSub = 0;
		for( unsigned long i = indexSuper; true; indexSub++, i++)
		{
			if( !string.m_pInst->m_data[indexSub] )
				return indexSuper;
			if( m_pInst->m_data[i] != string.m_pInst->m_data[indexSub] )
				break;
		}
	}

	return (unsigned long)-1;
}

unsigned long ESDOMString::index( const ESDOM_CHAR *substring ) const
{
	unsigned long subLength = strlen( substring );
	if( !m_pInst || !substring || !*substring || subLength > m_pInst->m_length || !subLength )
		return (unsigned long)-1;

	unsigned long indexSuper;
	unsigned long indexSub;
	unsigned long length = m_pInst->m_length - subLength;
	for( indexSuper = 0; indexSuper <= length ; indexSuper++ )
	{
		indexSub = 0;
		for( unsigned long i = indexSuper; 1; indexSub++, i++)
		{
			if( !substring[indexSub] )
				return indexSuper;
			if( m_pInst->m_data[i] != substring[indexSub] )
				break;
		}
	}

	return (unsigned long)-1;
}

ESDOMString ESDOMString::substr(unsigned long index, unsigned long length) const
{

	if( !m_pInst || index >= m_pInst->m_length )
		return ESDOMString();

	if( length == (unsigned long)-1 )
		length = m_pInst->m_length - index;
	else if( index + length > m_pInst->m_length )
		length = m_pInst->m_length - index;

	ESDOMStringInst *inst;
	try{ inst = new ESDOMStringInst(length);if(!inst)throw 0; }
	catch(...){ throw ESDOMException(ESDOMException::OUT_OF_MEMORY); }

	memcpy( inst->m_data, m_pInst->m_data+index, sizeof(ESDOM_CHAR)*(length));
	inst->m_data[length] = 0;
	ESDOMString returnString;
	returnString.m_pInst = inst;
	returnString.m_pInst->addRef();
	return returnString;
}

/////////////////////////////////////////////////////////////////////////

void ESDOMString::make_from_ESDOM_CHAR( const ESDOM_CHAR *string )
{
	if( !string || !*string ){
		m_pInst = 0;
		return;
	}

	try{
		m_pInst = new ESDOMStringInst( string );if(!m_pInst)throw 0;
	}catch(...){
		throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
	}
	m_pInst->addRef();
}

void ESDOMString::make_from_char( const char *string )
{
	if( !string || !*string ){
		m_pInst = 0;
		return;
	}

	try{
		m_pInst = new ESDOMStringInst( string );if(!m_pInst)throw 0;
	}catch(...){
		throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
	}
	m_pInst->addRef();
}

void ESDOMString::make_from_int(int n)
{
	ESDOM_CHAR	*string;
	try	{
		make_form_num(n, string);
	}catch(...){
		throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
	}

	try{
		make_from_ESDOM_CHAR( string );
	}catch(...){
		delete[] string;
		throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
	}
	delete[] string;
}

void ESDOMString::make_from_uint(unsigned n)
{
	ESDOM_CHAR	*string;
	try	{
		make_form_num(n, string);
	}catch(...){
		throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
	}

	try{
		make_from_ESDOM_CHAR( string );
	}catch(...){
		delete[] string;
		throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
	}
	delete[] string;
}


/////////////////////////////////////////////////////////////////////////

bool ESDOMString::isName() const
{
	if( !m_pInst || !m_pInst->m_length )
		return false;

	const ESDOM_CHAR *ptr = m_pInst->m_data;

	if(	!isLetter( *ptr )
		&& *ptr != '_'
		&& *ptr != ':' )
		return false;
	for( ptr++; *ptr; ptr++ )
	{
		if( !isLetter( *ptr )
			&& !isDigit( *ptr )
			&& *ptr != '.'
			&& *ptr != '-'
			&& *ptr != '_'
			&& *ptr != ':'
			&& !isCombiningChar( *ptr )
			&& !isExtender( *ptr ) )
			return false;
	}
	return true;
}

bool ESDOMString::isName(const ESDOM_CHAR *string)
{
	if(	!isLetter( *string )
		&& *string != '_'
		&& *string != ':' )
		return false;
	for( string++; *string; string++ )
	{
		if( !isLetter( *string )
			&& !isDigit( *string )
			&& *string != '.'
			&& *string != '-'
			&& *string != '_'
			&& *string != ':'
			&& !isCombiningChar( *string )
			&& !isExtender( *string ) )
			return false;
	}
	return true;
}

bool ESDOMString::isSpace( ESDOM_CHAR ch )
{
	if( ch == 0x20 || ch == 0x09 || ch == 0x0D || ch == 0x0A )
		return true;
	return false;
}

bool ESDOMString::isChar(ESDOM_CHAR ch)
{
	if(( ch == 0x09 )
	|| ( ch == 0x0A )
	|| ( ch == 0x0D )
	|| ( 0x0020 <= ch && ch <= 0xD7FF )
	|| ( 0xE000 <= ch && ch <= 0xFFFD )
	|| ( 0x10000 <= ch && ch <= 0x10FFFF ))
		return true;
	return false;
}

bool ESDOMString::isLetter( ESDOM_CHAR ch )
{
	return isBaseChar( ch ) || isIdeographic( ch );
}

bool ESDOMString::isBaseChar( ESDOM_CHAR ch )
{
	if( ch < 0x0BA8 ){
		if( ch < 0x06E5 ){
			if( ch < 0x03E2 ){
				if( ch < 0x01F4 ){
					if( ch < 0x0134 ){
						if(( 0x0041 <= ch && ch <= 0x005A )
						|| ( 0x0061 <= ch && ch <= 0x007A )
						|| ( 0x00C0 <= ch && ch <= 0x00D6 )
						|| ( 0x00D8 <= ch && ch <= 0x00F6 )
						|| ( 0x00F8 <= ch && ch <= 0x00FF )
						|| ( 0x0100 <= ch && ch <= 0x0131 ))
							return true;
					}else{
						if(( 0x0134 <= ch && ch <= 0x013E )
						|| ( 0x0141 <= ch && ch <= 0x0148 )
						|| ( 0x014A <= ch && ch <= 0x017E )
						|| ( 0x0180 <= ch && ch <= 0x01C3 )
						|| ( 0x01CD <= ch && ch <= 0x01F0 ))
							return true;
					}
				}else{
					if( ch < 0x038C ){
						if(( 0x01F4 <= ch && ch <= 0x01F5 )
						|| ( 0x01FA <= ch && ch <= 0x0217 )
						|| ( 0x0250 <= ch && ch <= 0x02A8 )
						|| ( 0x02BB <= ch && ch <= 0x02C1 )
						|| ( ch == 0x0386 )
						|| ( 0x0388 <= ch && ch <= 0x038A ))
							return true;
					}else{
						if(( ch == 0x038C )
						|| ( 0x038E <= ch && ch <= 0x03A1 )
						|| ( 0x03A3 <= ch && ch <= 0x03CE )
						|| ( 0x03D0 <= ch && ch <= 0x03D6 )
						|| ( ch == 0x03DA )
						|| ( ch == 0x03DC )
						|| ( ch == 0x03DE )
						|| ( ch == 0x03E0 ))
							return true;
					}
				}
			}else{
				if( ch < 0x0531 ){
					if( ch < 0x04C7 ){
						if(( 0x03E2 <= ch && ch <= 0x03F3 )
						|| ( 0x0401 <= ch && ch <= 0x040C )
						|| ( 0x040E <= ch && ch <= 0x044F )
						|| ( 0x0451 <= ch && ch <= 0x045C )
						|| ( 0x045E <= ch && ch <= 0x0481 )
						|| ( 0x0490 <= ch && ch <= 0x04C4 ))
							return true;
					}else{
						if(( 0x04C7 <= ch && ch <= 0x04C8 )
						|| ( 0x04CB <= ch && ch <= 0x04CC )
						|| ( 0x04D0 <= ch && ch <= 0x04EB )
						|| ( 0x04EE <= ch && ch <= 0x04F5 )
						|| ( 0x04F8 <= ch && ch <= 0x04F9 ))
							return true;
					}
				}else{
					if( ch < 0x0641 ){
						if(( 0x0531 <= ch && ch <= 0x0556 )
						|| ( ch == 0x0559 )
						|| ( 0x0561 <= ch && ch <= 0x0586 )
						|| ( 0x05D0 <= ch && ch <= 0x05EA )
						|| ( 0x05F0 <= ch && ch <= 0x05F2 )
						|| ( 0x0621 <= ch && ch <= 0x063A ))
							return true;
					}else{
						if(( 0x0641 <= ch && ch <= 0x064A )
						|| ( 0x0671 <= ch && ch <= 0x06B7 )
						|| ( 0x06BA <= ch && ch <= 0x06BE )
						|| ( 0x06C0 <= ch && ch <= 0x06CE )
						|| ( 0x06D0 <= ch && ch <= 0x06D3 )
						|| ( ch == 0x06D5 ))
							return true;
					}
				}
			}
		}else{
			if( ch < 0x0A8D ){
				if( ch < 0x09F0 ){
					if( ch < 0x0993 ){
						if(( 0x06E5 <= ch && ch <= 0x06E6 )
						|| ( 0x0905 <= ch && ch <= 0x0939 )
						|| ( ch == 0x093D )
						|| ( 0x0958 <= ch && ch <= 0x0961 )
						|| ( 0x0985 <= ch && ch <= 0x098C )
						|| ( 0x098F <= ch && ch <= 0x0990 ))
							return true;
					}else{
						if(( 0x0993 <= ch && ch <= 0x09A8 )
						|| ( 0x09AA <= ch && ch <= 0x09B0 )
						|| ( ch == 0x09B2 )
						|| ( 0x09B6 <= ch && ch <= 0x09B9 )
						|| ( 0x09DC <= ch && ch <= 0x09DD )
						|| ( 0x09DF <= ch && ch <= 0x09E1 ))
							return true;
					}
				}else{
					if( ch < 0x0A35 ){
						if(( 0x09F0 <= ch && ch <= 0x09F1 )
						|| ( 0x0A05 <= ch && ch <= 0x0A0A )
						|| ( 0x0A0F <= ch && ch <= 0x0A10 )
						|| ( 0x0A13 <= ch && ch <= 0x0A28 )
						|| ( 0x0A2A <= ch && ch <= 0x0A30 )
						|| ( 0x0A32 <= ch && ch <= 0x0A33 ))
							return true;
					}else{
						if(( 0x0A35 <= ch && ch <= 0x0A36 )
						|| ( 0x0A38 <= ch && ch <= 0x0A39 )
						|| ( 0x0A59 <= ch && ch <= 0x0A5C )
						|| ( ch == 0x0A5E )
						|| ( 0x0A72 <= ch && ch <= 0x0A74 )
						|| ( 0x0A85 <= ch && ch <= 0x0A8B ))
							return true;
					}
				}
			}else{
				if( ch < 0x0B32 ){
					if( ch < 0x0ABD ){
						if(( ch == 0x0A8D )
						|| ( 0x0A8F <= ch && ch <= 0x0A91 )
						|| ( 0x0A93 <= ch && ch <= 0x0AA8 )
						|| ( 0x0AAA <= ch && ch <= 0x0AB0 )
						|| ( 0x0AB2 <= ch && ch <= 0x0AB3 )
						|| ( 0x0AB5 <= ch && ch <= 0x0AB9 ))
							return true;
					}else{
						if(( ch == 0x0ABD )
						|| ( ch == 0x0AE0 )
						|| ( 0x0B05 <= ch && ch <= 0x0B0C )
						|| ( 0x0B0F <= ch && ch <= 0x0B10 )
						|| ( 0x0B13 <= ch && ch <= 0x0B28 )
						|| ( 0x0B2A <= ch && ch <= 0x0B30 ))
							return true;
					}
				}else{
					if( ch < 0x0B8E ){
						if(( 0x0B32 <= ch && ch <= 0x0B33 )
						|| ( 0x0B36 <= ch && ch <= 0x0B39 )
						|| ( ch == 0x0B3D )
						|| ( 0x0B5C <= ch && ch <= 0x0B5D )
						|| ( 0x0B5F <= ch && ch <= 0x0B61 )
						|| ( 0x0B85 <= ch && ch <= 0x0B8A ))
							return true;
					}else{
						if(( 0x0B8E <= ch && ch <= 0x0B90 )
						|| ( 0x0B92 <= ch && ch <= 0x0B95 )
						|| ( 0x0B99 <= ch && ch <= 0x0B9A )
						|| ( ch == 0x0B9C )
						|| ( 0x0B9E <= ch && ch <= 0x0B9F ))
							return true;
					}
				}
			}
		}
	}else{
		if( ch < 0x110E ){
			if( ch < 0x0E32 ){
				if( ch < 0x0C92 ){
					if( ch < 0x0C12 ){
						if(( 0x0BA3 <= ch && ch <= 0x0BA4 )
						|| ( 0x0BA8 <= ch && ch <= 0x0BAA )
						|| ( 0x0BAE <= ch && ch <= 0x0BB5 )
						|| ( 0x0BB7 <= ch && ch <= 0x0BB9 )
						|| ( 0x0C05 <= ch && ch <= 0x0C0C )
						|| ( 0x0C0E <= ch && ch <= 0x0C10 ))
							return true;
					}else{
						if(( 0x0C12 <= ch && ch <= 0x0C28 )
						|| ( 0x0C2A <= ch && ch <= 0x0C33 )
						|| ( 0x0C35 <= ch && ch <= 0x0C39 )
						|| ( 0x0C60 <= ch && ch <= 0x0C61 )
						|| ( 0x0C85 <= ch && ch <= 0x0C8C )
						|| ( 0x0C8E <= ch && ch <= 0x0C90 ))
							return true;
					}
				}else{
					if( ch < 0x0D0E ){
						if(( 0x0C92 <= ch && ch <= 0x0CA8 )
						|| ( 0x0CAA <= ch && ch <= 0x0CB3 )
						|| ( 0x0CB5 <= ch && ch <= 0x0CB9 )
						|| ( ch == 0x0CDE )
						|| ( 0x0CE0 <= ch && ch <= 0x0CE1 )
						|| ( 0x0D05 <= ch && ch <= 0x0D0C ))
							return true;
					}else{
						if(( 0x0D0E <= ch && ch <= 0x0D10 )
						|| ( 0x0D12 <= ch && ch <= 0x0D28 )
						|| ( 0x0D2A <= ch && ch <= 0x0D39 )
						|| ( 0x0D60 <= ch && ch <= 0x0D61 )
						|| ( 0x0E01 <= ch && ch <= 0x0E2E )
						|| ( ch == 0x0E30 ))
							return true;
					}
				}
			}else{
				if( ch < 0x0EB0 ){
					if( ch < 0x0E94 ){
						if(( 0x0E32 <= ch && ch <= 0x0E33 )
						|| ( 0x0E40 <= ch && ch <= 0x0E45 )
						|| ( 0x0E81 <= ch && ch <= 0x0E82 )
						|| ( ch == 0x0E84 )
						|| ( 0x0E87 <= ch && ch <= 0x0E88 )
						|| ( ch == 0x0E8A )
						|| ( ch == 0x0E8D ))
							return true;
					}else{
						if(( 0x0E94 <= ch && ch <= 0x0E97 )
						|| ( 0x0E99 <= ch && ch <= 0x0E9F )
						|| ( 0x0EA1 <= ch && ch <= 0x0EA3 )
						|| ( ch == 0x0EA5 )
						|| ( ch == 0x0EA7 )
						|| ( 0x0EAA <= ch && ch <= 0x0EAB )
						|| ( 0x0EAD <= ch && ch <= 0x0EAE ))
							return true;
					}
				}else{
					if( ch < 0x10D0 ){
						if(( ch == 0x0EB0 )
						|| ( 0x0EB2 <= ch && ch <= 0x0EB3 )
						|| ( ch == 0x0EBD )
						|| ( 0x0EC0 <= ch && ch <= 0x0EC4 )
						|| ( 0x0F40 <= ch && ch <= 0x0F47 )
						|| ( 0x0F49 <= ch && ch <= 0x0F69 )
						|| ( 0x10A0 <= ch && ch <= 0x10C5 ))
							return true;
					}else{
						if(( 0x10D0 <= ch && ch <= 0x10F6 )
						|| ( ch == 0x1100 )
						|| ( 0x1102 <= ch && ch <= 0x1103 )
						|| ( 0x1105 <= ch && ch <= 0x1107 )
						|| ( ch == 0x1109 )
						|| ( 0x110B <= ch && ch <= 0x110C ))
							return true;
					}
				}
			}
		}else{
			if( ch < 0x1F20 ){
				if( ch < 0x1175 ){
					if( ch < 0x115F ){
						if(( 0x110E <= ch && ch <= 0x1112 )
						|| ( ch == 0x113C )
						|| ( ch == 0x113E )
						|| ( ch == 0x1140 )
						|| ( ch == 0x114C )
						|| ( ch == 0x114E )
						|| ( ch == 0x1150 )
						|| ( 0x1154 <= ch && ch <= 0x1155 )
						|| ( ch == 0x1159 ))
							return true;
					}else{
						if(( 0x115F <= ch && ch <= 0x1161 )
						|| ( ch == 0x1163 )
						|| ( ch == 0x1165 )
						|| ( ch == 0x1167 )
						|| ( ch == 0x1169 )
						|| ( 0x116D <= ch && ch <= 0x116E )
						|| ( 0x1172 <= ch && ch <= 0x1173 ))
							return true;
					}
				}else{
					if( ch < 0x11BC ){
						if(( ch == 0x1175 )
						|| ( ch == 0x119E )
						|| ( ch == 0x11A8 )
						|| ( ch == 0x11AB )
						|| ( 0x11AE <= ch && ch <= 0x11AF )
						|| ( 0x11B7 <= ch && ch <= 0x11B8 )
						|| ( ch == 0x11BA ))
							return true;
					}else{
						if(( 0x11BC <= ch && ch <= 0x11C2 )
						|| ( ch == 0x11EB )
						|| ( ch == 0x11F0 )
						|| ( ch == 0x11F9 )
						|| ( 0x1E00 <= ch && ch <= 0x1E9B )
						|| ( 0x1EA0 <= ch && ch <= 0x1EF9 )
						|| ( 0x1F00 <= ch && ch <= 0x1F15 )
						|| ( 0x1F18 <= ch && ch <= 0x1F1D ))
							return true;
					}
				}
			}else{
				if( ch < 0x1FD6 ){
					if( ch < 0x1F80 ){
						if(( 0x1F20 <= ch && ch <= 0x1F45 )
						|| ( 0x1F48 <= ch && ch <= 0x1F4D )
						|| ( 0x1F50 <= ch && ch <= 0x1F57 )
						|| ( ch == 0x1F59 )
						|| ( ch == 0x1F5B )
						|| ( ch == 0x1F5D )
						|| ( 0x1F5F <= ch && ch <= 0x1F7D ))
							return true;
					}else{
						if(( 0x1F80 <= ch && ch <= 0x1FB4 )
						|| ( 0x1FB6 <= ch && ch <= 0x1FBC )
						|| ( ch == 0x1FBE )
						|| ( 0x1FC2 <= ch && ch <= 0x1FC4 )
						|| ( 0x1FC6 <= ch && ch <= 0x1FCC )
						|| ( 0x1FD0 <= ch && ch <= 0x1FD3 ))
							return true;
					}
				}else{
					if( ch < 0x212E ){
						if(( 0x1FD6 <= ch && ch <= 0x1FDB )
						|| ( 0x1FE0 <= ch && ch <= 0x1FEC )
						|| ( 0x1FF2 <= ch && ch <= 0x1FF4 )
						|| ( 0x1FF6 <= ch && ch <= 0x1FFC )
						|| ( ch == 0x2126 )
						|| ( 0x212A <= ch && ch <= 0x212B ))
							return true;
					}else{
						if(( ch == 0x212E )
						|| ( 0x2180 <= ch && ch <= 0x2182 )
						|| ( 0x3041 <= ch && ch <= 0x3094 )
						|| ( 0x30A1 <= ch && ch <= 0x30FA )
						|| ( 0x3105 <= ch && ch <= 0x312C )
						|| ( 0xAC00 <= ch && ch <= 0xD7A3 ))
							return true;
					}
				}
			}
		}
	}
	return false;
}

bool ESDOMString::isIdeographic( ESDOM_CHAR ch )
{
	if(( 0x4E00 <= ch && ch <= 0x9FA5 )
	|| ( ch == 0x3007 )
	|| ( 0x3021 <= ch && ch <= 0x3029 ))
		return true;
	return false;
}

bool ESDOMString::isCombiningChar( ESDOM_CHAR ch )
{
	if( ch < 0x0B4B ){
		if( ch < 0x09BC ){
			if( ch < 0x06DD ){
				if( ch < 0x05BB){
					if(( 0x0300 <= ch && ch <= 0x0345 )
					|| ( 0x0360 <= ch && ch <= 0x0361 )
					|| ( 0x0483 <= ch && ch <= 0x0486 )
					|| ( 0x0591 <= ch && ch <= 0x05A1 )
					|| ( 0x05A3 <= ch && ch <= 0x05B9 ))
						return true;
				}else{
					if(( 0x05BB <= ch && ch <= 0x05BD )
					|| ( ch == 0x05BF )
					|| ( 0x05C1 <= ch && ch <= 0x05C2 )
					|| ( ch == 0x05C4 )
					|| ( 0x064B <= ch && ch <= 0x0652 )
					|| ( ch == 0x0670 )
					|| ( 0x06D6 <= ch && ch <= 0x06DC ))
						return true;
				}
			}else{
				if( ch < 0x093C ){
					if(( 0x06DD <= ch && ch <= 0x06DF )
					|| ( 0x06E0 <= ch && ch <= 0x06E4 )
					|| ( 0x06E7 <= ch && ch <= 0x06E8 )
					|| ( 0x06EA <= ch && ch <= 0x06ED )
					|| ( 0x0901 <= ch && ch <= 0x0903 ))
						return true;
				}else{
					if(( ch == 0x093C )
					|| ( 0x093E <= ch && ch <= 0x094C )
					|| ( ch == 0x094D )
					|| ( 0x0951 <= ch && ch <= 0x0954 )
					|| ( 0x0962 <= ch && ch <= 0x0963 )
					|| ( 0x0981 <= ch && ch <= 0x0983 ))
						return true;
				}
			}
		}else{
			if( ch < 0x0A4B ){
				if( ch < 0x09E2 ){
					if(( ch == 0x09BC )
					|| ( ch == 0x09BE )
					|| ( ch == 0x09BF )
					|| ( 0x09C0 <= ch && ch <= 0x09C4 )
					|| ( 0x09C7 <= ch && ch <= 0x09C8 )
					|| ( 0x09CB <= ch && ch <= 0x09CD )
					|| ( ch == 0x09D7 ))
						return true;
				}else{
					if(( 0x09E2 <= ch && ch <= 0x09E3 )
					|| ( ch == 0x0A02 )
					|| ( ch == 0x0A3C )
					|| ( ch == 0x0A3E )
					|| ( ch == 0x0A3F )
					|| ( 0x0A40 <= ch && ch <= 0x0A42 )
					|| ( 0x0A47 <= ch && ch <= 0x0A48 ))
						return true;
				}
			}else{
				if( ch < 0x0ACB ){
					if(( 0x0A4B <= ch && ch <= 0x0A4D )
					|| ( 0x0A70 <= ch && ch <= 0x0A71 )
					|| ( 0x0A81 <= ch && ch <= 0x0A83 )
					|| ( ch == 0x0ABC )
					|| ( 0x0ABE <= ch && ch <= 0x0AC5 )
					|| ( 0x0AC7 <= ch && ch <= 0x0AC9 ))
						return true;
				}else{
					if(( 0x0ACB <= ch && ch <= 0x0ACD )
					|| ( 0x0B01 <= ch && ch <= 0x0B03 )
					|| ( ch == 0x0B3C )
					|| ( 0x0B3E <= ch && ch <= 0x0B43 )
					|| ( 0x0B47 <= ch && ch <= 0x0B48 ))
						return true;
				}
			}
		}
	}else{
		if( ch < 0x0D57 ){
			if( ch < 0x0C55 ){
				if( ch < 0x0BCA ){
					if(( 0x0B4B <= ch && ch <= 0x0B4D )
					|| ( 0x0B56 <= ch && ch <= 0x0B57 )
					|| ( 0x0B82 <= ch && ch <= 0x0B83 )
					|| ( 0x0BBE <= ch && ch <= 0x0BC2 )
					|| ( 0x0BC6 <= ch && ch <= 0x0BC8 ))
						return true;
				}else{
					if(( 0x0BCA <= ch && ch <= 0x0BCD )
					|| ( ch == 0x0BD7 )
					|| ( 0x0C01 <= ch && ch <= 0x0C03 )
					|| ( 0x0C3E <= ch && ch <= 0x0C44 )
					|| ( 0x0C46 <= ch && ch <= 0x0C48 )
					|| ( 0x0C4A <= ch && ch <= 0x0C4D ))
						return true;
				}
			}else{
				if( ch < 0x0CD5 ){
					if(( 0x0C55 <= ch && ch <= 0x0C56 )
					|| ( 0x0C82 <= ch && ch <= 0x0C83 )
					|| ( 0x0CBE <= ch && ch <= 0x0CC4 )
					|| ( 0x0CC6 <= ch && ch <= 0x0CC8 )
					|| ( 0x0CCA <= ch && ch <= 0x0CCD ))
						return true;
				}else{
					if(( 0x0CD5 <= ch && ch <= 0x0CD6 )
					|| ( 0x0D02 <= ch && ch <= 0x0D03 )
					|| ( 0x0D3E <= ch && ch <= 0x0D43 )
					|| ( 0x0D46 <= ch && ch <= 0x0D48 )
					|| ( 0x0D4A <= ch && ch <= 0x0D4D ))
						return true;
				}
			}
		}else{
			if( ch < 0x0F3F ){
				if( ch < 0x0EBB ){
					if(( ch == 0x0D57 )
					|| ( ch == 0x0E31 )
					|| ( 0x0E34 <= ch && ch <= 0x0E3A )
					|| ( 0x0E47 <= ch && ch <= 0x0E4E )
					|| ( ch == 0x0EB1 )
					|| ( 0x0EB4 <= ch && ch <= 0x0EB9 ))
						return true;
				}else{
					if(( 0x0EBB <= ch && ch <= 0x0EBC )
					|| ( 0x0EC8 <= ch && ch <= 0x0ECD )
					|| ( 0x0F18 <= ch && ch <= 0x0F19 )
					|| ( ch == 0x0F35 )
					|| ( ch == 0x0F37 )
					|| ( ch == 0x0F39 )
					|| ( ch == 0x0F3E ))
						return true;
				}
			}else{
				if( ch < 0x0FB1 ){
					if(( ch == 0x0F3F )
					|| ( 0x0F71 <= ch && ch <= 0x0F84 )
					|| ( 0x0F86 <= ch && ch <= 0x0F8B )
					|| ( 0x0F90 <= ch && ch <= 0x0F95 )
					|| ( ch == 0x0F97 )
					|| ( 0x0F99 <= ch && ch <= 0x0FAD ))
						return true;
				}else{
					if(( 0x0FB1 <= ch && ch <= 0x0FB7 )
					|| ( ch == 0x0FB9 )
					|| ( 0x20D0 <= ch && ch <= 0x20DC )
					|| ( ch == 0x20E1 )
					|| ( 0x302A <= ch && ch <= 0x302F )
					|| ( ch == 0x3099 )
					|| ( ch == 0x309A ))
						return true;
				}
			}
		}
	}
	return false;
}

bool ESDOMString::isDigit(ESDOM_CHAR ch)
{
	if( ch < 0x0BE7 ){
		if( ch < 0x09E6 ){
			if(( 0x0030 <= ch && ch <= 0x0039 )
			|| ( 0x0660 <= ch && ch <= 0x0669 )
			|| ( 0x06F0 <= ch && ch <= 0x06F9 )
			|| ( 0x0966 <= ch && ch <= 0x096F ))
				return true;
		}else{
			if(( 0x09E6 <= ch && ch <= 0x09EF )
			|| ( 0x0A66 <= ch && ch <= 0x0A6F )
			|| ( 0x0AE6 <= ch && ch <= 0x0AEF )
			|| ( 0x0B66 <= ch && ch <= 0x0B6F ))
				return true;
		}
	}else{
		if( ch < 0x0E50 ){
			if(( 0x0BE7 <= ch && ch <= 0x0BEF )
			|| ( 0x0C66 <= ch && ch <= 0x0C6F )
			|| ( 0x0CE6 <= ch && ch <= 0x0CEF )
			|| ( 0x0D66 <= ch && ch <= 0x0D6F ))
				return true;
		}else{
			if(( 0x0E50 <= ch && ch <= 0x0E59 )
			|| ( 0x0ED0 <= ch && ch <= 0x0ED9 )
			|| ( 0x0F20 <= ch && ch <= 0x0F29 ))
				return true;
		}
	}
	return false;
}

bool ESDOMString::isExtender(ESDOM_CHAR ch)
{
	if( ch < 0x3005 ){
		if(( ch == 0x00B7 )
		|| ( ch == 0x02D0 )
		|| ( ch == 0x02D1 )
		|| ( ch == 0x0387 )
		|| ( ch == 0x0640 )
		|| ( ch == 0x0E46 )
		|| ( ch == 0x0EC6 ))
			return true;
	}else{
		if(( ch == 0x3005 )
		|| ( 0x3031 <= ch && ch <= 0x3035 )
		|| ( 0x309D <= ch && ch <= 0x309E )
		|| ( 0x30FC <= ch && ch <= 0x30FE ))
			return true;
	}
	return false;
}
