#include <ESDOM/ESDOM_define.hpp>
#include <ESDOM/API.hpp>
#include <string.h>
#include <ESDOM/ESDOMException.hpp>
#include "ESDOMStringInst.hpp"

using namespace ESDOM;

unsigned long ESDOMStringInst::PAGE_SIZE = 0;

ESDOMStringInst::ESDOMStringInst( unsigned long length )
{
	m_refCount = 0;
	m_length = 0;
	m_data = 0;

	m_data = _allocate( length );
	m_length = length;
}

ESDOMStringInst::ESDOMStringInst( const ESDOM_CHAR *string )
{
	m_refCount = 0;
	m_length = 0;
	m_data = 0;

	if( !string )
	{
		m_data = _allocate( 0 );
		*m_data = 0;
		return;
	}
	unsigned long length = strlen( string );
	m_data = _allocate( length );
	m_length = length;
	memcpy( m_data, string, sizeof(ESDOM_CHAR)*(length+1) );
}

ESDOMStringInst::ESDOMStringInst( const char *string )
{
	m_refCount = 0;
	m_length = 0;
	m_data = 0;

	if( !string )
	{
		m_data = _allocate( 0 );
		*m_data = 0;
		return;
	}
	unsigned long length = ::strlen( string );
	m_data = _allocate( length );
	m_length = length;
	ESDOM_CHAR*	ptr = m_data;
	while(*string)
		*ptr++ = *string++;
	*ptr = 0;
}

ESDOMStringInst::~ESDOMStringInst()
{
	if( m_data )
		delete[] m_data;
}

ESDOM_CHAR* ESDOMStringInst::_allocate( unsigned long length )
{
	ESDOM_CHAR	*ptr;
	unsigned long size = ( ( length+1 )/PAGE_SIZE + 1 ) * PAGE_SIZE;
	try
	{
		ptr = new ESDOM_CHAR[size]; if(!ptr)throw 0;
	}
	catch(...)
	{
		throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
	}
	return ptr;
}

void ESDOMStringInst::addRef(void) const
{
	m_refCount++;
}

void ESDOMStringInst::removeRef(void) const
{
	if( --m_refCount < 1 )
		delete this;
}

unsigned long ESDOMStringInst::getRefCount(void) const
{
	return m_refCount;
}

void ESDOMStringInst::operator += (const ESDOM_CHAR *string)
{
	unsigned long length = strlen(string);
	if( PAGE_SIZE - m_length % PAGE_SIZE -1 >= length )
	{
		memcpy(m_data + m_length, string, sizeof(ESDOM_CHAR)*(length+1) );
		m_length += length;
	}
	else
	{
		ESDOM_CHAR*	oldData = m_data;
		try{
			m_data = _allocate(m_length+length);
		}catch(...){
			throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
		}
		memcpy( m_data, oldData, sizeof(ESDOM_CHAR)*m_length );
		memcpy( m_data + m_length, string, sizeof(ESDOM_CHAR)*(length+1) );
		m_length += length;
		delete[] oldData;
	}
}

void ESDOMStringInst::operator += (const char *string)
{
	unsigned long length = ::strlen(string);
	if( PAGE_SIZE - m_length % PAGE_SIZE -1 >= length )
	{
		ESDOM_CHAR *ptr = m_data + m_length;
		while(*string)
			*ptr++ = *string++;
		*ptr = 0;
		m_length += length;
	}
	else
	{
		ESDOM_CHAR*	oldData = m_data;
		try{
			m_data = _allocate(m_length+length);
		}catch(...){
			throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
		}
		memcpy( m_data, oldData, sizeof(ESDOM_CHAR)*m_length );
		ESDOM_CHAR *ptr = m_data + m_length;
		while(*string)
			*ptr++ = *string++;
		*ptr = 0;
		m_length += length;
		delete[] oldData;
	}
}

void ESDOMStringInst::operator += (ESDOM_CHAR ch)
{
	if( (m_length+1) % PAGE_SIZE )
	{
		m_data[m_length] = ch;
		m_data[m_length+1] = 0;
		m_length++;
	}
	else
	{
		ESDOM_CHAR	*oldData = m_data;
		try{
			m_data = _allocate(m_length+1);
		}catch(...){
			throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
		}
		memcpy( m_data, oldData, sizeof(ESDOM_CHAR)*m_length );
		m_data[m_length] = ch;
		m_data[m_length+1] = 0;
		m_length++;
		delete[] oldData;
	}
}
