#ifndef INCLUDED_ESDOM_ESDOM_vector_hpp
#define INCLUDED_ESDOM_ESDOM_vector_hpp

namespace ESDOM
{

template <class _Ty>
class ESDOM_EXPORT vector{
public:
	vector(unsigned pageSize = 16){
		m_pageSize = pageSize;
		m_unitNum = 0;
		m_length = 0;
		m_data = 0;
		m_data = new _Ty[m_pageSize];
		if(!m_data)
			throw 0;
		m_unitNum = 1;
	}
	~vector(){
		delete[] m_data;
	}
	vector & operator=(const vector &original){
		removeAll();
		m_data = _alloc(original.m_unitNum);
		memcpy(m_data, original.m_data, sizeof(_Ty)*original.m_length);
		return *this;
	}
	_Ty operator[](unsigned index) const{
		return (index >= m_length) ? 0 : m_data[index];
	}
	bool append(_Ty data){
		if(m_length >= m_pageSize*m_unitNum){
			_Ty * p = _alloc(m_unitNum+1);
			memcpy(p, m_data, sizeof(_Ty)*m_length);
			delete[]m_data;
			m_data = p;
			m_unitNum++;
		}
		m_data[m_length] = data;
		m_length++;
		return true;
	}
	bool insert(unsigned index, _Ty data){
		if(index >= m_length)
			return false;
		if(m_length >= m_pageSize*m_unitNum){
			_Ty * p = _alloc(m_unitNum+1);
			memcpy(p, m_data, sizeof(_Ty)*m_length);
			delete[]m_data;
			m_data = p;
			m_unitNum++;
		}
		memmove(&m_data[index+1], &m_data[index], sizeof(_Ty)*(m_length-index));
		m_data[index] = data;
		m_length++;
		return true;
	}
	bool replace(unsigned index, _Ty data){
		if(index >= m_length)
			return false;
		m_data[index] = data;
		return true;
	}
	bool remove(unsigned index){
		if(index >= m_length)
			return false;
		memmove(&m_data[index], &m_data[index+1], sizeof(_Ty)*(m_length-index-1));
		m_length--;
		if(m_unitNum > 1 && m_length <= m_pageSize*(m_unitNum-1) - m_pageSize/2){
			_Ty * p = _alloc(m_length/m_pageSize+1);
			memcpy(p, m_data, sizeof(_Ty)*m_length);
			delete[]m_data;
			m_data = p;
			m_unitNum--;
		}
		return true;
	}
	void removeAll(void){
		m_length = 0;
		delete[]m_data;
		m_unitNum = 0;
	}
	unsigned getLength(void) const{
		return m_length;
	}

private:
	_Ty			*m_data;
	unsigned	m_length;
	unsigned	m_pageSize;
	unsigned	m_unitNum;

	_Ty * _alloc(unsigned num){
		_Ty *ptr = new _Ty[m_pageSize*num];
		if(!ptr)
			throw 0;
		return ptr;
	}
};

} /* namespace ESDOM */

#endif /* INCLUDED_ESDOM_ESDOM_vector_hpp */
