#include <ESDOM/ESDOM_define.hpp>
#include <ESDOM/Element.hpp>
#include "ElementInst.hpp"
#include <ESDOM/Node.hpp>
#include <ESDOM/NodeList.hpp>

using namespace ESDOM;

Element::Element()
{
}

Element::Element(const Element &original)
: NodeBase( original )
{
}

Element::Element(const ElementInst *instance)
: NodeBase( instance )
{
}

Element::~Element()
{
}

Element & Element::operator = (const Element &node)
{
	Interface::operator = ( node );
	return *this;
}

ESDOMString Element::getAttribute(const ESDOM_CHAR *name) const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return ((ElementInst*)m_instance)->getAttribute(name);
}

void Element::setAttribute(const ESDOM_CHAR *name, const ESDOM_CHAR *value) throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	((ElementInst*)m_instance)->setAttribute(name, value);
}

void Element::removeAttribute(const ESDOM_CHAR *name) throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	((ElementInst*)m_instance)->removeAttribute(name);
}

Attr Element::getAttributeNode(const ESDOM_CHAR *name) const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return Attr( ((ElementInst*)m_instance)->getAttributeNode(name) );
}

Attr Element::setAttributeNode(Attr newAttr) throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	AttrInst *oldAttrInst = ((ElementInst*)m_instance)->getAttributeNode( newAttr.getNodeName() );
	if( !oldAttrInst )
	{
		((ElementInst*)m_instance)->appendAttributeNode((AttrInst*)newAttr.m_instance);
		return newAttr;
	}
	Attr returnAttr(oldAttrInst);
	((ElementInst*)m_instance)->replaceAttrbuteNode((AttrInst*)newAttr.m_instance, oldAttrInst);

	return returnAttr;
}

Attr Element::removeAttributeNode(Attr oldAttr) throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	((ElementInst*)m_instance)->removeAttributeNode((AttrInst*)oldAttr.m_instance);
	return oldAttr;
}

NodeList Element::getElementsByTagName(const ESDOM_CHAR *name) const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return NodeList( ((ElementInst*)m_instance)->getElementsByTagName(name) );
}

void Element::normalize()
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	((ElementInst*)m_instance)->normalize();
}
