#include <ESDOM/ESDOM_define.hpp>
#include <ESDOM/ESDOMException.hpp>
#include "ElementInst.hpp"
#include "NodeInst.hpp"
#include "NodeListInst.hpp"
#include "NamedNodeMapInst.hpp"
#include "AttrInst.hpp"
#include <string.h>

#include <ESDOM/Node.hpp>

using namespace ESDOM;

ElementInst::ElementInst(const ESDOM_CHAR *tagName)
 : NodeInst(tagName, Node::ELEMENT_NODE, true, true)
{
}

ElementInst::ElementInst( const ElementInst &original )
 : NodeInst( original )
{
}

ElementInst::~ElementInst()
{
}

NodeInst * ElementInst::cloneNodeInst(bool deep) const
{
	ElementInst *newNodeInst;

	try{
		newNodeInst = new ElementInst( *this ); if(!newNodeInst)throw 0;
	}catch(...){
		throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
	}

	if( !deep )
		return newNodeInst;

//	newNodeInst->m_readonly = false;
	NodeListInst *newChildNodes = m_childNodes->clone(true);
	for( unsigned long index = 0, length = newChildNodes->getLength(); index < length; index++ )
		newChildNodes->item(index)->m_parentNode = (NodeInst*)newNodeInst;
	newNodeInst->m_childNodes->removeRef();
	newNodeInst->m_childNodes = newChildNodes;
	newNodeInst->m_childNodes->addRef();

//	newNodeInst->m_readonly = m_readonly;
	return newNodeInst;
}

ESDOMString	ElementInst::getAttribute(const ESDOM_CHAR *name) const
{
	if( !name )
		return ESDOMString();
	NodeInst *attr = m_attributes->getNamedItem( name );
	if( !attr )
		return ESDOMString();
	return attr->getNodeValue();
}

void ElementInst::setAttribute(const ESDOM_CHAR *name, const ESDOM_CHAR *value)
{
	// check >>
	if( m_flags & READ_ONLY )
		throw ESDOMException( ESDOMException::NO_MODIFICATION_ALLOWED_ERR );

	if( !name || !ESDOMString::isName(name) )
		throw ESDOMException(ESDOMException::INVALID_CHARACTER_ERR);
	// <<
	AttrInst *newAttr;
	try{
		newAttr = new AttrInst(name); if(!newAttr)throw 0;
	}catch(...){
		throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
	}
	if( value )
		newAttr->setNodeValue(value);

	NodeInst *oldNode =  m_attributes->getNamedItem(name);
	if( !oldNode )
	{
		m_attributes->appendNamedItem(newAttr);
		newAttr->m_ownerElement = this;
		return;
	}
	oldNode->addRef();
	m_attributes->replaceNamedItem( newAttr, oldNode );
	newAttr->m_ownerElement = this;
	AttrInst *oldAttr = dynamic_cast<AttrInst*>( oldNode );
	if( !oldAttr )
	{
		oldNode->removeRef();
		throw ESDOMException(ESDOMException::BAD_CAST);
	}
	oldAttr->m_ownerElement = 0;
	oldNode->removeRef();
}

void ElementInst::removeAttribute(const ESDOM_CHAR *name)
{
	// check >>
	if( m_flags & READ_ONLY )
		throw ESDOMException( ESDOMException::NO_MODIFICATION_ALLOWED_ERR );
	// <<
	if( !name )
		return;
	NodeInst *oldNode = m_attributes->getNamedItem(name);
	if( !oldNode )
		return;

	oldNode->addRef();
	m_attributes->removeNamedItem(oldNode);
	AttrInst *oldAttr = dynamic_cast<AttrInst*>( oldNode );
	if( !oldAttr )
	{
		oldNode->removeRef();
		throw ESDOMException(ESDOMException::BAD_CAST);
	}
	oldAttr->m_ownerElement = 0;
	oldNode->removeRef();
}

AttrInst* ElementInst::getAttributeNode(const ESDOM_CHAR *name) const
{
	if( !name )
		return 0;
	NodeInst *node = m_attributes->getNamedItem(name);
	if( !node )
		return 0;
	AttrInst *attr = dynamic_cast<AttrInst*>(node);
	if( !attr )
	{
		throw ESDOMException(ESDOMException::BAD_CAST);
	}
	return attr;
}

bool ElementInst::appendAttributeNode(AttrInst *newAttr)
{
	if( !newAttr )
		return false;
	// check >>
	if( m_flags & READ_ONLY )
		throw ESDOMException( ESDOMException::NO_MODIFICATION_ALLOWED_ERR );

	if( newAttr->m_ownerElement )
		throw ESDOMException(ESDOMException::INUSE_ATTRIBUTE_ERR);
	// <<
	if( m_attributes->appendNamedItem(newAttr) )
	{
		newAttr->m_ownerElement = this;
		return true;
	}
	return false;
}

bool ElementInst::replaceAttrbuteNode(AttrInst *newAttr, AttrInst *oldAttr)
{
	if( !newAttr || !oldAttr )
		return false;
	// check >>
	if( m_flags & READ_ONLY )
		throw ESDOMException( ESDOMException::NO_MODIFICATION_ALLOWED_ERR );

	if( newAttr->m_ownerElement )
		throw ESDOMException(ESDOMException::INUSE_ATTRIBUTE_ERR);
	// <<
	if( m_attributes->replaceNamedItem( newAttr, oldAttr ) )
	{
		newAttr->m_ownerElement = this;
		oldAttr->m_ownerElement = 0;
		return true;
	}
	return false;
}

void ElementInst::removeAttributeNode(AttrInst *oldAttr)
{
	// check >>
	if( !oldAttr || oldAttr->m_ownerElement != this )
		throw ESDOMException(ESDOMException::NOT_FOUND_ERR);

	if( m_flags & READ_ONLY )
		throw ESDOMException( ESDOMException::NO_MODIFICATION_ALLOWED_ERR );
	// <<
	if( !m_attributes->removeNamedItem(oldAttr) )
		throw ESDOMException(ESDOMException::NOT_FOUND_ERR);
	oldAttr->m_ownerElement = 0;
}

NodeListInst* ElementInst::getElementsByTagName(const ESDOM_CHAR *tagname) const
{
	NodeListInst *nodeList;
	try
	{
		nodeList = new NodeListInst;
		if( !nodeList )throw 0;
	}
	catch(...)
	{
		throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
	}
	if( !tagname || !*tagname )
		return nodeList;

	for( unsigned long index = 0; index < m_childNodes->getLength(); index++ )
	{
		if( m_childNodes->item(index)->getNodeType() == Node::ELEMENT_NODE )
		{
			if( m_childNodes->item(index)->getNodeName() == tagname
				|| !memcmp(tagname,_ESDOM_TEXT("*"),sizeof(ESDOM_CHAR)*2) )
				nodeList->appendNode( m_childNodes->item(index) );
			NodeListInst *descendantNodeList = ((ElementInst*)m_childNodes->item(index))->getElementsByTagName(tagname);
			for( unsigned long j = 0; j < descendantNodeList->getLength(); j++ )
				nodeList->appendNode( descendantNodeList->item(j) );
			descendantNodeList->removeRef();
		}
	}
	return nodeList;
}

void ElementInst::normalize()
{
	for( unsigned long i = 0, length = m_childNodes->getLength(); i < length; i++ )
	{
		NodeInst *nodeInst = m_childNodes->item(i);
		switch( nodeInst->getNodeType() )
		{
		case Node::ELEMENT_NODE:
			((ElementInst*)nodeInst)->normalize();
			break;

		case Node::TEXT_NODE:
			while(1)
			{
				NodeInst *nextInst = m_childNodes->item(i+1);
				if( !nextInst || nextInst->getNodeType() != Node::TEXT_NODE )
					break;
				nodeInst->m_nodeValue += nextInst->m_nodeValue;
				nextInst->addRef();
				removeChild( nextInst );
				nextInst->removeRef();
				length--;
			}
			break;
		}
	}
}

///////////////////////////////////////////////////////////

