#ifndef INCLUDED_ESDOM_ElementInst_hpp
#define INCLUDED_ESDOM_ElementInst_hpp

#include "NodeInst.hpp"
#include <ESDOM/Node.hpp>

namespace ESDOM
{

class ESDOM_EXPORT AttrInst;

class ESDOM_EXPORT ElementInst : public NodeInst
{
public:
	explicit ElementInst(const ESDOM_CHAR *tagName);
	~ElementInst();

	NodeInst *	cloneNodeInst(bool deep) const;

	ESDOMString		getAttribute(const ESDOM_CHAR *name) const;
	void			setAttribute(const ESDOM_CHAR *name, const ESDOM_CHAR *value);
	void			removeAttribute(const ESDOM_CHAR *name);
	AttrInst *		getAttributeNode(const ESDOM_CHAR *name) const;
	bool			appendAttributeNode(AttrInst *newAttr);
	bool			replaceAttrbuteNode(AttrInst *newAttr, AttrInst *oldAttr);
	void			removeAttributeNode(AttrInst *oldAttr);
	NodeListInst *	getElementsByTagName(const ESDOM_CHAR *name) const;
	void			normalize(void);

	NodeInst *		insertBefore(NodeInst *newChild, NodeInst *refChild)
					{
						if( !newChild )return 0;
						_checkNewNodeType(newChild);
						return NodeInst::insertBefore(newChild,refChild);
					}
	NodeInst *		replaceChild(NodeInst *newChild, NodeInst *oldChild)
					{
						if( !newChild )return 0;
						_checkNewNodeType(newChild);
						return NodeInst::replaceChild(newChild,oldChild);
					}
	NodeInst *		appendChild(NodeInst *newChild)
					{
						if( !newChild )return 0;
						_checkNewNodeType(newChild);
						return NodeInst::appendChild(newChild);
					}

protected:
	explicit ElementInst(const ElementInst &original);
	void	_checkNewNodeType(const NodeInst *newChild) const
			{
				unsigned short nodeType = newChild->getNodeType();
				if( nodeType == Node::ATTRIBUTE_NODE 
					|| nodeType == Node::PROCESSING_INSTRUCTION_NODE 
					|| nodeType == Node::DOCUMENT_NODE 
					|| nodeType == Node::DOCUMENT_TYPE_NODE 
					)
					throw ESDOMException(ESDOMException::HIERARCHY_REQUEST_ERR);
			}
};

}; /* namespace ESDOM */

#endif /* INCLUDED_ESDOM_ElementInst_hpp */
