#include <ESDOM/ESDOM_define.hpp>
#include <ESDOM/ESDOMException.hpp>
#include <ESDOM/NamedNodeMap.hpp>
#include <ESDOM/Node.hpp>

#include "NamedNodeMapInst.hpp"
#include "AttrInst.hpp"

using namespace ESDOM;

NamedNodeMap::NamedNodeMap()
{
}

NamedNodeMap::NamedNodeMap(const NamedNodeMap &original)
 : Interface(original)
{
}

NamedNodeMap::NamedNodeMap(const NamedNodeMapInst *instance)
 : Interface(instance)
{
}

NamedNodeMap::~NamedNodeMap()
{
}

NamedNodeMap & NamedNodeMap::operator = (const NamedNodeMap &newNamedNodeMap)
{
	if( m_instance )
		m_instance->removeRef();
	
	m_instance = newNamedNodeMap.m_instance;
	if( m_instance )
		m_instance->addRef();
	return *this;
}

unsigned long NamedNodeMap::getLength(void) const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return ((NamedNodeMapInst*)m_instance)->getLength();
}

Node NamedNodeMap::getNamedItem(const ESDOM_CHAR *name) const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( !name )
		return 0;
	return Node( ((NamedNodeMapInst*)m_instance)->getNamedItem( name ) );
}

Node NamedNodeMap::setNamedItem(const Node arg) throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( arg.getNodeType() == Node::ATTRIBUTE_NODE )
	{
		if( ((AttrInst*)arg.m_instance)->getOwnerElement() )
			throw ESDOMException(ESDOMException::INUSE_ATTRIBUTE_ERR);
	}
	NodeInst *oldNode = ((NamedNodeMapInst*)m_instance)->getNamedItem( arg.getNodeName() );
	
	if( oldNode )
	{
		Node returnNode( oldNode );
		((NamedNodeMapInst*)m_instance)->replaceNamedItem( (NodeInst*)arg.m_instance, oldNode );
		return returnNode;
	}
	((NamedNodeMapInst*)m_instance)->appendNamedItem((NodeInst*)arg.m_instance);

	return 0;
}

Node NamedNodeMap::removeNamedItem(const ESDOM_CHAR *name) throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);

	NodeInst *node = ((NamedNodeMapInst*)m_instance)->getNamedItem( name );
	if( !node )
		throw ESDOMException(ESDOMException::NOT_FOUND_ERR);

	Node ret( node );
	((NamedNodeMapInst*)m_instance)->removeNamedItem(node);
	return ret;
}

Node NamedNodeMap::item(unsigned long index) const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return Node( ((NamedNodeMapInst*)m_instance)->item(index) );
}

