#include <ESDOM/ESDOM_define.hpp>
#include <ESDOM/API.hpp>
#include <ESDOM/ESDOMException.hpp>
#include <ESDOM/ESDOMString.hpp>
#include "NamedNodeMapInst.hpp"
#include "NodeInst.hpp"
#include "AttrInst.hpp"
#include <string.h>

#include <ESDOM/Node.hpp>

using namespace ESDOM;

unsigned long NamedNodeMapInst::TABLE_SIZE = 0;

NamedNodeMapInst::NamedNodeMapInst()
{
	m_nodes = vector<LISTITEM*>(TABLE_SIZE/2);
	m_hashTable = 0;
	try{
		m_hashTable = new LISTITEM*[TABLE_SIZE]; if(!m_hashTable)throw 0;
	}catch(...){
		throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
	}
	memset( m_hashTable, 0, sizeof(LISTITEM*)*TABLE_SIZE );
}

NamedNodeMapInst::NamedNodeMapInst(const NamedNodeMapInst &original)
{
	m_nodes = vector<LISTITEM*>(TABLE_SIZE/2);
	m_hashTable = 0;
	try{
		m_hashTable = new LISTITEM*[TABLE_SIZE]; if(!m_hashTable)throw 0;
	}catch(...){
		throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
	}
	memset( m_hashTable, 0, sizeof(LISTITEM*)*TABLE_SIZE );
	for( unsigned long index = 0, length = original.m_nodes.getLength(); index < length; index++ )
	{
		appendNamedItem( original.m_nodes[index]->inst->cloneNodeInst(false) );
	}
}

NamedNodeMapInst::~NamedNodeMapInst()
{
	if( m_hashTable )
	{
		for( unsigned long index = 0, length = m_nodes.getLength(); index < length; index++ )
		{
			m_nodes[index]->inst->removeRef();
			delete[]m_nodes[index];
		}	
		delete[] m_hashTable;
	}
}

unsigned long NamedNodeMapInst::getLength(void) const
{
	return m_nodes.getLength();
}

NodeInst * NamedNodeMapInst::item( unsigned long index ) const
{
	return m_nodes[index]->inst;
}

NodeInst * NamedNodeMapInst::getNamedItem( const ESDOM_CHAR *name ) const
{
	if( !name )
		return 0;

	LISTITEM	*pItem = _findItem( name );
	if( pItem )
		return pItem->inst;

	return 0;
}

bool NamedNodeMapInst::appendNamedItem(const NodeInst *newNode)
{
	if( !newNode )
		return false;

	ESDOMString	name = newNode->getNodeName();

	if( _findItem( name ) )
		return false;

	LISTITEM	*pNewItem;
	try
	{
		pNewItem = new LISTITEM; if(!pNewItem)throw 0;
	}
	catch(...)
	{
		throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
	}

	pNewItem->inst = const_cast<NodeInst*>(newNode);
	pNewItem->previous = 0;
	unsigned long hash = _makeHashCode( (const ESDOM_CHAR*)name );
	pNewItem->next = m_hashTable[hash];
	if( m_hashTable[hash] )
		m_hashTable[hash]->previous = pNewItem;
	m_hashTable[hash] = pNewItem;

	m_nodes.append(pNewItem);
	newNode->addRef();
	return true;
}

bool NamedNodeMapInst::replaceNamedItem(const NodeInst *newNode, const NodeInst *oldNode)
{
	if( !newNode || !oldNode )
		return false;

	LISTITEM *pItem = _findItem( oldNode->getNodeName() );
	if( !pItem || pItem->inst != oldNode )
		return false;

	pItem->inst = const_cast<NodeInst*>(newNode);
	oldNode->removeRef();
	newNode->addRef();
	return true;
}

bool NamedNodeMapInst::removeNamedItem(const NodeInst *oldNode)
{
	if( !oldNode )
		return false;

	for( unsigned long index = 0, length = m_nodes.getLength(); index < length; index++ )
	{
		if( m_nodes[index]->inst == oldNode )
		{
			if(m_nodes[index]->next)
				m_nodes[index]->next->previous = m_nodes[index]->previous;
			if(m_nodes[index]->previous)
				m_nodes[index]->previous->next = m_nodes[index]->next;
			unsigned long hash = _makeHashCode( (const ESDOM_CHAR*)oldNode->getNodeName() );
			if(m_hashTable[hash] == m_nodes[index])
				m_hashTable[hash] = m_nodes[index]->next;
			delete[]m_nodes[index];
			m_nodes.remove(index);
			oldNode->removeRef();
			return true;
		}
	}
	return false;
}

//////////////////////////////////////////////

NamedNodeMapInst * NamedNodeMapInst::clone() const
{
	NamedNodeMapInst *clone;
	try
	{
		clone = new NamedNodeMapInst( *this ); if(!clone)throw 0;
	}
	catch(...)
	{
		throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
	}
	return clone;
}

unsigned long NamedNodeMapInst::_makeHashCode( const ESDOM_CHAR *name ) const
{
	const ESDOM_CHAR	*ptr;
	unsigned long	hash = 0;

	for( ptr = name; *ptr; ptr++ )
	{
		hash = hash * 37 + *ptr;
	}
	return hash % TABLE_SIZE;
}

NamedNodeMapInst::LISTITEM * NamedNodeMapInst::_findItem( const ESDOM_CHAR *name ) const
{
	LISTITEM	*pItem;
	pItem = m_hashTable[_makeHashCode(name)];
	while(pItem)
	{
		if( !strcmp( name, pItem->inst->getNodeName() ) )
			return pItem;
		pItem = pItem->next;
	}
	return 0;
}
