#ifndef INCLUDED_ESDOM_NamedNodeMapInst_hpp
#define INCLUDED_ESDOM_NamedNodeMapInst_hpp

#include "Instance.hpp"
#include "ESDOM_vector.hpp"

namespace ESDOM
{

class ESDOM_EXPORT NodeInst;

class ESDOM_EXPORT NamedNodeMapInst : public Instance
{
	struct LISTITEM
	{
		NodeInst	*inst;
		LISTITEM	*next;
		LISTITEM	*previous;
	};

public:
	explicit NamedNodeMapInst();
	~NamedNodeMapInst();

	unsigned long		getLength(void) const;
	NodeInst *			getNamedItem(const ESDOM_CHAR *name) const;
	bool				appendNamedItem(const NodeInst *newNode);
	bool				replaceNamedItem(const NodeInst *newNode, const NodeInst *oldNode);
	bool				removeNamedItem(const NodeInst *oldNode);
	NodeInst *			item(unsigned long index) const;
	NamedNodeMapInst *	clone(void) const;

protected:
	explicit NamedNodeMapInst(const NamedNodeMapInst &original);

	static unsigned long	TABLE_SIZE;

	friend	class	Initializer;

private:
#ifdef NEED_EXPLICIT_INSTANTIATION
template class ESDOM_EXPORT ::ESDOM::vector<LISTITEM*>;
#endif

	LISTITEM			**m_hashTable;
	vector<LISTITEM*>	m_nodes;

	unsigned long	_makeHashCode(const ESDOM_CHAR *name) const;
	LISTITEM *		_findItem( const ESDOM_CHAR *name ) const;
};

}; /* namespace ESDOM */

#endif /* INCLUDED_ESDOM_NamedNodeMapInst_hpp */
