#include <ESDOM/ESDOM_define.hpp>
#include <ESDOM/Node.hpp>
#include "NodeInst.hpp"
#include <ESDOM/NodeList.hpp>
#include <ESDOM/Attr.hpp>

using namespace ESDOM;

Node::Node()
{
}

Node::Node(const Node &original)
: NodeBase( original )
{
}

Node::Node(const NodeInst *instance)
: NodeBase( instance )
{
}

Node::~Node()
{
}

Node & Node::operator = (const Node &newNode)
{
	Interface::operator = (newNode);
	return *this;
}

/////////////////////////////////////////////////////////////
// Element & Document

NodeList Node::getElementsByTagName(const ESDOM_CHAR *name) const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	unsigned short nodeType = ((NodeInst*)m_instance)->getNodeType();
	if( nodeType == Node::DOCUMENT_NODE )
		return Document::getElementsByTagName(name);
	else if ( nodeType == Node::ELEMENT_NODE )
		return Element::getElementsByTagName(name);
	throw ESDOMException(ESDOMException::METHOD_IS_NOT_MEMBER);
}

/////////////////////////////////////////////////////////////
// Element

ESDOMString Node::getAttribute(const ESDOM_CHAR *name) const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( ((NodeInst*)m_instance)->getNodeType() != Node::ELEMENT_NODE )
		throw ESDOMException(ESDOMException::METHOD_IS_NOT_MEMBER);
	return Element::getAttribute(name);
}

void Node::setAttribute(const ESDOM_CHAR *name, const ESDOM_CHAR *value) throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( ((NodeInst*)m_instance)->getNodeType() != Node::ELEMENT_NODE )
		throw ESDOMException(ESDOMException::METHOD_IS_NOT_MEMBER);
	Element::setAttribute(name, value);
}

void Node::removeAttribute(const ESDOM_CHAR *name) throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( ((NodeInst*)m_instance)->getNodeType() != Node::ELEMENT_NODE )
		throw ESDOMException(ESDOMException::METHOD_IS_NOT_MEMBER);
	Element::removeAttribute(name);
}

Attr Node::getAttributeNode(const ESDOM_CHAR *name) const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( ((NodeInst*)m_instance)->getNodeType() != Node::ELEMENT_NODE )
		throw ESDOMException(ESDOMException::METHOD_IS_NOT_MEMBER);
	return Element::getAttributeNode(name);
}

Attr Node::setAttributeNode(Attr newAttr) throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( ((NodeInst*)m_instance)->getNodeType() != Node::ELEMENT_NODE )
		throw ESDOMException(ESDOMException::METHOD_IS_NOT_MEMBER);
	return Element::setAttributeNode(newAttr);
}

Attr Node::removeAttributeNode(Attr oldAttr) throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( ((NodeInst*)m_instance)->getNodeType() != Node::ELEMENT_NODE )
		throw ESDOMException(ESDOMException::METHOD_IS_NOT_MEMBER);
	return Element::removeAttributeNode(oldAttr);
}

void Node::normalize()
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( ((NodeInst*)m_instance)->getNodeType() != Node::ELEMENT_NODE )
		throw ESDOMException(ESDOMException::METHOD_IS_NOT_MEMBER);
	Element::normalize();
}

/////////////////////////////////////////////////////////////
// Document
DocumentType Node::getDoctype() const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( ((NodeInst*)m_instance)->getNodeType() != Node::DOCUMENT_NODE )
		throw ESDOMException(ESDOMException::METHOD_IS_NOT_MEMBER);
	return Document::getDoctype();
}

bool Node::getStandalone() const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( ((NodeInst*)m_instance)->getNodeType() != Node::DOCUMENT_NODE )
		throw ESDOMException(ESDOMException::METHOD_IS_NOT_MEMBER);
	return Document::getStandalone();
}

void Node::setStandalone(bool n)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( ((NodeInst*)m_instance)->getNodeType() != Node::DOCUMENT_NODE )
		throw ESDOMException(ESDOMException::METHOD_IS_NOT_MEMBER);
	Document::setStandalone(n);
}

Element Node::getDocumentElement() const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( ((NodeInst*)m_instance)->getNodeType() != Node::DOCUMENT_NODE )
		throw ESDOMException(ESDOMException::METHOD_IS_NOT_MEMBER);
	return Document::getDocumentElement();
}

Element Node::createElement(const ESDOM_CHAR *tagName) const throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( ((NodeInst*)m_instance)->getNodeType() != Node::DOCUMENT_NODE )
		throw ESDOMException(ESDOMException::METHOD_IS_NOT_MEMBER);
	return Document::createElement(tagName);
}

Text Node::createTextNode(const ESDOM_CHAR *data) const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( ((NodeInst*)m_instance)->getNodeType() != Node::DOCUMENT_NODE )
		throw ESDOMException(ESDOMException::METHOD_IS_NOT_MEMBER);
	return Document::createTextNode(data);
}

Comment Node::createComment(const ESDOM_CHAR *data) const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( ((NodeInst*)m_instance)->getNodeType() != Node::DOCUMENT_NODE )
		throw ESDOMException(ESDOMException::METHOD_IS_NOT_MEMBER);
	return Document::createComment(data);
}

CDATASection Node::createCDATASection(const ESDOM_CHAR *data) const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( ((NodeInst*)m_instance)->getNodeType() != Node::DOCUMENT_NODE )
		throw ESDOMException(ESDOMException::METHOD_IS_NOT_MEMBER);
	return Document::createCDATASection(data);
}

ProcessingInstruction Node::createProcessingInstruction(const ESDOM_CHAR *target, const ESDOM_CHAR *data) const throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( ((NodeInst*)m_instance)->getNodeType() != Node::DOCUMENT_NODE )
		throw ESDOMException(ESDOMException::METHOD_IS_NOT_MEMBER);
	return Document::createProcessingInstruction(target, data);
}

Attr Node::createAttribute(const ESDOM_CHAR *name) const throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( ((NodeInst*)m_instance)->getNodeType() != Node::DOCUMENT_NODE )
		throw ESDOMException(ESDOMException::METHOD_IS_NOT_MEMBER);
	return Document::createAttribute(name);
}

Element Node::getElementById(const ESDOM_CHAR *id) const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( ((NodeInst*)m_instance)->getNodeType() != Node::DOCUMENT_NODE )
		throw ESDOMException(ESDOMException::METHOD_IS_NOT_MEMBER);
	return Document::getElementById(id);
}

ESDOMString Node::setDoctype(const ESDOM_CHAR *doctype)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	if( ((NodeInst*)m_instance)->getNodeType() != Node::DOCUMENT_NODE )
		throw ESDOMException(ESDOMException::METHOD_IS_NOT_MEMBER);
	return Document::setDoctype(doctype);
}
