#include <ESDOM/ESDOM_define.hpp>
#include <ESDOM/NodeBase.hpp>
#include <ESDOM/NodeList.hpp>
#include <ESDOM/NamedNodeMap.hpp>
#include <ESDOM/Node.hpp>

#include "NodeInst.hpp"

using namespace ESDOM;

NodeBase::NodeBase()
{
}

NodeBase::NodeBase(const NodeBase &original)
 : Interface( original )
{
}

NodeBase::NodeBase(const NodeInst *instance)
 : Interface( instance )
{
}

NodeBase::~NodeBase()
{
}

NodeBase & NodeBase::operator = (const NodeBase &newNode)
{
	Interface::operator = (newNode);
	return *this;
}

ESDOMString NodeBase::getNodeName() const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return ((NodeInst*)m_instance)->getNodeName();
}

ESDOMString NodeBase::getNodeValue() const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return ((NodeInst*)m_instance)->getNodeValue();
}

void NodeBase::setNodeValue(const ESDOM_CHAR *value) throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	((NodeInst*)m_instance)->setNodeValue(value);
}

unsigned short NodeBase::getNodeType() const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return ((NodeInst*)m_instance)->getNodeType();
}

Node NodeBase::getParentNode() const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return Node( ((NodeInst*)m_instance)->getParentNode() );
}

NodeList NodeBase::getChildNodes() const
{
	if( m_instance )
		return NodeList( ((NodeInst*)m_instance)->getChildNodes() );
	return 0;
}

Node NodeBase::getFirstChild() const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return Node( ((NodeInst*)m_instance)->getFirstChild() );
}

Node NodeBase::getLastChild() const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return Node( ((NodeInst*)m_instance)->getLastChild() );
}

Node NodeBase::getPreviousSibling() const
{
	if( m_instance )
		return Node( ((NodeInst*)m_instance)->getPreviousSibling() );
	return 0;
}

Node NodeBase::getNextSibling() const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return Node( ((NodeInst*)m_instance)->getNextSibling() );
}

NamedNodeMap NodeBase::getAttributes() const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return NamedNodeMap( ((NodeInst*)m_instance)->getAttributes() );
}

bool NodeBase::hasChildNodes() const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return ((NodeInst*)m_instance)->hasChildNodes();
}

Node NodeBase::insertBefore(Node newChild, Node refChild) throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return Node( ((NodeInst*)m_instance)->insertBefore( (NodeInst*)newChild.m_instance, (NodeInst*)refChild.m_instance ) );
}

Node NodeBase::replaceChild(Node newChild, Node oldChild) throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return Node( ((NodeInst*)m_instance)->replaceChild( (NodeInst*)newChild.m_instance, (NodeInst*)oldChild.m_instance ) );
}

Node NodeBase::removeChild(Node oldChild) throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return Node( ((NodeInst*)m_instance)->removeChild( (NodeInst*)oldChild.m_instance ) );
}

Node NodeBase::appendChild(Node newChild) throw(ESDOMException)
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return Node( ((NodeInst*)m_instance)->appendChild( (NodeInst*)newChild.m_instance ) );
}

Node NodeBase::cloneNode(bool deep)const
{
	if( !m_instance )
		throw ESDOMException(ESDOMException::NULL_OBJECT);
	return Node( ((NodeInst*)m_instance)->cloneNodeInst(deep) );
}

NodeBase::operator Element () const
{
	if( !m_instance )
		return Element();
	if( ((NodeInst*)m_instance)->getNodeType() != Node::ELEMENT_NODE )
		throw ESDOMException(ESDOMException::BAD_CAST);
	return Element( (ElementInst*)m_instance );
}

NodeBase::operator Attr () const
{
	if( !m_instance )
		return Attr();
	if( ((NodeInst*)m_instance)->getNodeType() != Node::ATTRIBUTE_NODE )
		throw ESDOMException(ESDOMException::BAD_CAST);
	return Attr( (AttrInst*)m_instance );
}

NodeBase::operator Text () const
{
	if( !m_instance )
		return Text();
	if( ((NodeInst*)m_instance)->getNodeType() != Node::TEXT_NODE )
		throw ESDOMException(ESDOMException::BAD_CAST);
	return Text( (TextInst*)m_instance );
}

NodeBase::operator CDATASection () const
{
	if( !m_instance )
		return CDATASection();
	if( ((NodeInst*)m_instance)->getNodeType() != Node::CDATA_SECTION_NODE )
		throw ESDOMException(ESDOMException::BAD_CAST);
	return CDATASection( (CDATASectionInst*)m_instance );
}

NodeBase::operator ProcessingInstruction () const
{
	if( !m_instance )
		return ProcessingInstruction();
	if( ((NodeInst*)m_instance)->getNodeType() != Node::PROCESSING_INSTRUCTION_NODE )
		throw ESDOMException(ESDOMException::BAD_CAST);
	return ProcessingInstruction( (ProcessingInstructionInst*)m_instance );
}

NodeBase::operator Comment () const
{
	if( !m_instance )
		return Comment();
	if( ((NodeInst*)m_instance)->getNodeType() != Node::COMMENT_NODE )
		throw ESDOMException(ESDOMException::BAD_CAST);
	return Comment( (CommentInst*)m_instance );
}

NodeBase::operator Document () const
{
	if( !m_instance )
		return Document();
	if( ((NodeInst*)m_instance)->getNodeType() != Node::DOCUMENT_NODE )
		throw ESDOMException(ESDOMException::BAD_CAST);
	return Document( (DocumentInst*)m_instance );
}

NodeBase::operator DocumentType () const
{
	if( !m_instance )
		return DocumentType();
	if( ((NodeInst*)m_instance)->getNodeType() != Node::DOCUMENT_TYPE_NODE )
		throw ESDOMException(ESDOMException::BAD_CAST);
	return DocumentType( (DocumentTypeInst*)m_instance );
}

NodeBase::operator Node () const
{
	return Node( ((NodeInst*)m_instance) );
}
