#include <ESDOM/ESDOM_define.hpp>
#include <ESDOM/ESDOMException.hpp>
#include "NodeInst.hpp"
#include "NodeListInst.hpp"
#include "NamedNodeMapInst.hpp"
//#include "DocumentInst.hpp"

using namespace ESDOM;

NodeInst::NodeInst( const ESDOM_CHAR *nodeName, unsigned short nodeType, bool hasAttributes, bool canHaveChild )
{
	m_nodeName		= nodeName;
//	m_nodeValue		= ;
	m_nodeType		= nodeType;
	m_parentNode	= 0;

	try{
		m_childNodes = new NodeListInst; if(!m_childNodes)throw 0;
	}catch(...){
		throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
	}
	m_childNodes->addRef();

	if( hasAttributes )
	{
		try	{
			m_attributes = new NamedNodeMapInst; if(!m_attributes)throw 0;
		}catch(...){
			throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
		}
		m_attributes->addRef();
	}
	else
		m_attributes = 0;
//	m_ownerDocument	= 0;
	m_flags = 0;
	if( canHaveChild )
		m_flags |= CAN_HAVE_CHILD;
}

NodeInst::NodeInst( const NodeInst &original )
{
	m_nodeName		= original.m_nodeName;
	m_nodeValue		= original.m_nodeValue;
	m_nodeType		= original.m_nodeType;
	m_parentNode	= 0;

	try{
		m_childNodes = new NodeListInst; if(!m_childNodes)throw 0;
	}catch(...){
		throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
	}

	if( original.m_attributes )
	{
		m_attributes = original.m_attributes->clone();
		m_attributes->addRef();
	}
	else
		m_attributes = 0;

//	m_ownerDocument	= 0;
	m_flags = original.m_flags;
}

NodeInst::~NodeInst()
{
	m_childNodes->removeRef();
	if( m_attributes )
		m_attributes->removeRef();
}

ESDOMString NodeInst::getNodeName() const
{
	return m_nodeName;
}

ESDOMString NodeInst::getNodeValue() const
{
	return m_nodeValue;
}

void NodeInst::setNodeValue( const ESDOM_CHAR *value )
{
	if( m_flags & READ_ONLY )
		throw ESDOMException( ESDOMException::NO_MODIFICATION_ALLOWED_ERR );
	m_nodeValue = value;
}

unsigned short NodeInst::getNodeType() const
{
	return m_nodeType;
}

NodeInst * NodeInst::getParentNode() const
{
	return m_parentNode;
}

NodeListInst * NodeInst::getChildNodes() const
{
	return m_childNodes;
}

NodeInst * NodeInst::getFirstChild() const
{
	return m_childNodes->item(0);
}

NodeInst * NodeInst::getLastChild() const
{
	return m_childNodes->item( m_childNodes->getLength()-1 );
}

NodeInst * NodeInst::getPreviousSibling() const
{
	if( !m_parentNode )
		return 0;

	NodeInst *previous = 0;
	for( unsigned long index = 0, length = m_parentNode->m_childNodes->getLength();
		index < length;
		previous = m_parentNode->m_childNodes->item(index++) )
	{
		if( m_parentNode->m_childNodes->item(index) == this )
			return previous;
	}
	return 0;
}

NodeInst * NodeInst::getNextSibling() const
{
	if( !m_parentNode )
		return 0;

	NodeInst *next = 0;
	for( unsigned long index = m_parentNode->m_childNodes->getLength()-1;
		index >= 0;
		next = m_parentNode->m_childNodes->item(index--) )
	{
		if( m_parentNode->m_childNodes->item(index) == this )
			return next;
	}
	return 0;
}

NamedNodeMapInst * NodeInst::getAttributes() const
{
	return m_attributes;
}
/*
DocumentInst * NodeInst::getOwnerDocument()
{
	m_ownerDocument->addRef;
	return m_ownerDocument;
}
*/
NodeInst * NodeInst::insertBefore( NodeInst *newChild, NodeInst *refChild )
{
	// Check >>
	if( m_flags & READ_ONLY )
		throw ESDOMException( ESDOMException::NO_MODIFICATION_ALLOWED_ERR );

	if( !newChild || !(m_flags&CAN_HAVE_CHILD)  || newChild->_isAncestorOf( this ) )
		throw ESDOMException( ESDOMException::HIERARCHY_REQUEST_ERR );

	if( refChild->m_parentNode != this )
		throw ESDOMException( ESDOMException::NOT_FOUND_ERR );

	//	DocumentInst *owner = newChild->getOwnerDocument();
//	if( owner != this->m_ownerDocument )
//	{
//		owner->removeRef();
//		throw ESDOMException( ESDOMException::WRONG_DOCUMENT_ERR );
//	}
//	owner->removeRef();
	// <<

	if( !refChild )
		return appendChild( newChild );


	if( !m_childNodes->insertNode( newChild, refChild ) )
		throw ESDOMException( ESDOMException::NOT_FOUND_ERR );

	if( newChild->m_parentNode )
		newChild->m_parentNode->m_childNodes->removeNode( newChild );

	newChild->m_parentNode = this;
	return newChild;
}

NodeInst * NodeInst::replaceChild( NodeInst *newChild, NodeInst *oldChild )
{
	// ErrorCheck >>
	if( m_flags & READ_ONLY )
		throw ESDOMException( ESDOMException::NO_MODIFICATION_ALLOWED_ERR );

	if( !newChild || !(m_flags&CAN_HAVE_CHILD) || newChild->_isAncestorOf( this ) )
		throw ESDOMException( ESDOMException::HIERARCHY_REQUEST_ERR );

	if( !oldChild || oldChild->m_parentNode != this )
		throw ESDOMException( ESDOMException::NOT_FOUND_ERR );

//	DocumentInst *owner = newChild->getOwnerDocument();
//	if( owner != this->m_ownerDocument )
//	{
//		owner->removeRef();
//		throw ESDOMException( ESDOMException::WRONG_DOCUMENT_ERR );
//	}
//	owner->removeRef();
	// <<

	if( !m_childNodes->replaceNode( newChild, oldChild ) )
		throw ESDOMException( ESDOMException::NOT_FOUND_ERR );

	if( newChild->m_parentNode )
		newChild->m_parentNode->m_childNodes->removeNode( newChild );

	newChild->m_parentNode = this;
	oldChild->m_parentNode = 0;
	return oldChild;
}

NodeInst * NodeInst::removeChild( NodeInst *oldChild )
{
	// ErrorCheck >>
	if( m_flags & READ_ONLY )
		throw ESDOMException( ESDOMException::NO_MODIFICATION_ALLOWED_ERR );

	if( !oldChild || oldChild->m_parentNode != this )
		throw ESDOMException( ESDOMException::NOT_FOUND_ERR );
	// <<

	if( !m_childNodes->removeNode( oldChild ) )
		throw ESDOMException( ESDOMException::NOT_FOUND_ERR );

	oldChild->m_parentNode = 0;
	return oldChild;
}

NodeInst * NodeInst::appendChild( NodeInst *newChild )
{
	// ErrorCheck >>
	if( m_flags & READ_ONLY )
		throw ESDOMException( ESDOMException::NO_MODIFICATION_ALLOWED_ERR );

	if( !newChild || !(m_flags&CAN_HAVE_CHILD) || newChild->_isAncestorOf( this ) )
		throw ESDOMException( ESDOMException::HIERARCHY_REQUEST_ERR );

//	DocumentInst *owner = newChild->getOwnerDocument();
//	if( owner != this->m_ownerDocument )
//	{
//		owner->removeRef();
//		throw ESDOMException( ESDOMException::WRONG_DOCUMENT_ERR );
//	}
//	owner->removeRef();
	// <<

	m_childNodes->appendNode( newChild );

	if( newChild->m_parentNode )
		newChild->m_parentNode->removeChild( newChild );

	newChild->m_parentNode = this;
	return newChild;
}

bool NodeInst::hasChildNodes() const
{
	return m_childNodes->getLength() ? true : false;
}

/////////////////////////////////////////////////////////////////////////////

bool NodeInst::_isAncestorOf( const NodeInst *testNode ) const
{
	const NodeInst *parent = testNode;
	do
	{
		if( parent == this )
			return true;
		parent = parent->m_parentNode;
	}while( parent );
	return false;
}
