#ifndef INCLUDED_ESDOM_NodeInst_hpp
#define INCLUDED_ESDOM_NodeInst_hpp

#include <ESDOM/ESDOMString.hpp>
#include "Instance.hpp"

namespace ESDOM
{

class ESDOM_EXPORT NodeListInst;
class ESDOM_EXPORT NamedNodeMapInst;

class ESDOM_EXPORT NodeInst : public Instance
{
public:
	explicit NodeInst(const ESDOM_CHAR* nodeName, unsigned short nodeType, bool hasAttributes, bool canHaveChild);
	virtual	~NodeInst();

	virtual	ESDOMString			getNodeName(void) const;
	virtual	ESDOMString			getNodeValue(void) const;
	virtual	void				setNodeValue(const ESDOM_CHAR *value);
	virtual	unsigned short		getNodeType(void) const;

	virtual NodeInst *			getParentNode(void) const;
	virtual	NodeListInst *		getChildNodes(void) const;
	virtual	NodeInst *			getFirstChild(void) const;
	virtual	NodeInst *			getLastChild(void) const;
	virtual	NodeInst *			getPreviousSibling(void) const;
	virtual	NodeInst *			getNextSibling(void) const;
	virtual	NamedNodeMapInst *	getAttributes(void) const;

	virtual	NodeInst*			insertBefore(NodeInst *newChildInst, NodeInst *refChildInst);
	virtual	NodeInst*			replaceChild(NodeInst *newChildInst, NodeInst *oldChildInst);
	virtual	NodeInst*			removeChild(NodeInst *oldChildInst);
	virtual	NodeInst*			appendChild(NodeInst *newChildInst);
	virtual	bool				hasChildNodes(void) const;
	virtual	NodeInst*			cloneNodeInst(bool deep) const =0;

protected:
	enum flagValue{
		READ_ONLY		= 0x01,
		CAN_HAVE_CHILD	= 0x02
	};

	explicit NodeInst(const NodeInst &original);

	unsigned short		m_flags;
	ESDOMString			m_nodeName;
	ESDOMString			m_nodeValue;
	unsigned short		m_nodeType;
	NodeInst			*m_parentNode;
	NodeListInst		*m_childNodes;
	NamedNodeMapInst	*m_attributes;
//	DocumentInst		*m_ownerDocument;
//	bool				m_readonly;

	virtual bool		_isAncestorOf(const NodeInst *testNode) const;

	friend	class	ElementInst;
	friend	class	DocumentInst;
};

}; /* namespace ESDOM */

#endif /* INCLUDED_ESDOM_NodeInst_hpp */
