#include <ESDOM/ESDOM_define.hpp>
#include <ESDOM/ESDOMException.hpp>
#include "NodeListInst.hpp"
#include "NodeInst.hpp"
#include <string.h>

using namespace ESDOM;

unsigned long NodeListInst::ITEMS_PAR_UNIT = 0;

NodeListInst::NodeListInst()
{
	m_length = 0;
}

NodeListInst::NodeListInst(const NodeListInst &original)
{
	for( unsigned long pageIndex = 0, length = original.m_page.getLength(); pageIndex <= length; pageIndex++)
	{
		m_page.append(_allocate());
	}

	m_length = original.m_length;
	for( unsigned long index = 0, page = 0; index < m_length; page++ )
	{
		for( unsigned long offset = 0; offset < ITEMS_PAR_UNIT && index < m_length; offset++, index++)
			( m_page[page][offset] = original.m_page[page][offset] )->addRef();
	}
}

NodeListInst::NodeListInst(const NodeListInst &original, bool duplicateNode)
{
	for( unsigned long pageIndex = 0, length = original.m_page.getLength(); pageIndex <= length; pageIndex++)
	{
		m_page.append(_allocate());
	}

	m_length = original.m_length;
	for( unsigned long index = 0, page = 0; index < m_length; page++ )
	{
		for( unsigned long offset = 0; offset < ITEMS_PAR_UNIT && index < m_length; offset++, index++)
		{
			if( duplicateNode )
				m_page[page][offset] = original.m_page[page][offset]->cloneNodeInst(true);
			else
				m_page[page][offset] = original.m_page[page][offset];
			m_page[page][offset]->addRef();
		}
	}
}

NodeListInst::~NodeListInst()
{
	for( unsigned long index = 0, page = 0; index < m_length; page++ )
	{
		for( unsigned long offset = 0; offset < ITEMS_PAR_UNIT && index < m_length; offset++, index++)
			m_page[page][offset]->removeRef();
	}

	for( unsigned long i = 0, length = m_page.getLength(); i < length; i++ )
	{
		delete[]m_page[i];
	}
}

NodeInst * NodeListInst::item( unsigned long index ) const
{
	if( index >= m_length )
		return 0;
	return m_page[index/ITEMS_PAR_UNIT][index%ITEMS_PAR_UNIT];
}

unsigned long NodeListInst::getLength( void ) const
{
	return m_length;
}

NodeListInst * NodeListInst::clone(bool duplicateNode) const
{
	NodeListInst *newInst;
	try{
		newInst = new NodeListInst(*this, duplicateNode); if(!newInst)throw 0;
	}catch(...){
		throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
	}
	return newInst;
}

bool NodeListInst::appendNode( const NodeInst *newNodeInst )
{
	if( !newNodeInst )
		return false;

	if( m_length >= ITEMS_PAR_UNIT * m_page.getLength() )
		m_page.append(_allocate());

	m_page[m_length/ITEMS_PAR_UNIT][m_length%ITEMS_PAR_UNIT] = const_cast<NodeInst*>(newNodeInst);
	newNodeInst->addRef();
	m_length++;
	return true;
}

bool NodeListInst::insertNode( const NodeInst *newNodeInst, const NodeInst *refNodeInst )
{
	if( !newNodeInst || !refNodeInst )
		return false;

	unsigned long index, page, offset = 0;
	for( index = 0, page = 0; index < m_length; page++ )
	{
		bool fBreak = false;
		for( offset = 0; offset < ITEMS_PAR_UNIT && index < m_length; offset++, index++)
		{
			if( m_page[page][offset] == refNodeInst )
			{
				fBreak = true;
				break;
			}
		}
		if(fBreak)
			break;
	}
	if( index == m_length )
		return false;

	if( m_length >= ITEMS_PAR_UNIT * m_page.getLength() )
		m_page.append(_allocate());


	if( m_length < (page+1)*ITEMS_PAR_UNIT )
	{
		memmove( &m_page[page][offset+1], &m_page[page][offset], sizeof(NodeInst*)*(m_length-index) );
		m_page[page][offset] = const_cast<NodeInst*>(newNodeInst);
	}
	else
	{
		NodeInst	*buf, *buf2;
		buf = m_page[page][ITEMS_PAR_UNIT-1];
		memmove( &m_page[page][offset+1], &m_page[page][offset], sizeof(NodeInst*)*(ITEMS_PAR_UNIT-offset-1) );
		m_page[page][offset] = const_cast<NodeInst*>(newNodeInst);

		index += ITEMS_PAR_UNIT - offset;
		page++;
		offset = 0;
		for( ; 1; index += ITEMS_PAR_UNIT, page++ )
		{
			if( m_length - index < ITEMS_PAR_UNIT )
			{
				memmove( &m_page[page][1], &m_page[page][0], sizeof(NodeInst*)*(m_length-index) );
				m_page[page][0] = buf;
				break;
			}
			buf2 = m_page[page][ITEMS_PAR_UNIT-1];
			memmove( &m_page[page][1], &m_page[page][0], sizeof(NodeInst*)*(ITEMS_PAR_UNIT-1) );
			m_page[page][0] = buf;
			buf = buf2;
		}
	}

	newNodeInst->addRef();
	m_length++;
	return true;
}

bool NodeListInst::removeNode( const NodeInst *oldNodeInst )
{
	if( !oldNodeInst )
		return false;

	unsigned long index, page, offset = 0;
	for( index = 0, page = 0; index < m_length; page++ )
	{
		bool fBreak = false;
		for( offset = 0; offset < ITEMS_PAR_UNIT && index < m_length; offset++, index++)
		{
			if( m_page[page][offset] == oldNodeInst )
			{
				fBreak = true;
				break;
			}
		}
		if(fBreak)
			break;
	}
	if( index == m_length )
		return false;

	if( m_length <= (page+1)*ITEMS_PAR_UNIT )
	{
		memmove( &m_page[page][offset], &m_page[page][offset+1], sizeof(NodeInst*)*(m_length-index-1) );
	}
	else
	{
		if( offset < ITEMS_PAR_UNIT-1 )
			memmove( &m_page[page][offset], &m_page[page][offset+1], sizeof(NodeInst*)*(ITEMS_PAR_UNIT-offset-1) );
		m_page[page][ITEMS_PAR_UNIT-1] = m_page[page+1][0];

		index += ITEMS_PAR_UNIT - offset;
		page++;
		offset = 0;
		for( ; 1; index += ITEMS_PAR_UNIT, page++ )
		{
			if( m_length - index <= ITEMS_PAR_UNIT )
			{
				memmove( &m_page[page][0], &m_page[page][1], sizeof(NodeInst*)*(m_length-index-1) );
				break;
			}
			memmove( &m_page[page][0], &m_page[page][1], sizeof(NodeInst*)*(ITEMS_PAR_UNIT-1) );
			m_page[page][ITEMS_PAR_UNIT-1] = m_page[page+1][0];
		}
	}

	m_length--;
	oldNodeInst->removeRef();

	if( m_page.getLength() > 1 && m_length < ITEMS_PAR_UNIT * (m_page.getLength()-1) - ITEMS_PAR_UNIT/2 )
	{
		unsigned long removeIndex = m_length/ITEMS_PAR_UNIT+1;
		while(m_page[removeIndex])
		{
			delete[]m_page[removeIndex];
			m_page.remove(removeIndex);
		}
	}
	return true;
}

bool NodeListInst::replaceNode( const NodeInst *newNodeInst, const NodeInst *oldNodeInst )
{
	if( !newNodeInst || !oldNodeInst )
		return false;

	for( unsigned long index = 0, page = 0; index < m_length; page++ )
	{
		for( unsigned long offset = 0; offset < ITEMS_PAR_UNIT && index < m_length; offset++, index++)
		{
			if( m_page[page][offset] == oldNodeInst )
			{
				m_page[page][offset] = const_cast<NodeInst*>(newNodeInst);
				newNodeInst->addRef();
				oldNodeInst->removeRef();
				return true;
			}
		}
	}
	return false;
}

///////////////////////////////////////////////////////////////////

inline NodeInst ** NodeListInst::_allocate()
{
	NodeInst	**nodes;
	try
	{
		nodes = new NodeInst*[ITEMS_PAR_UNIT]; if(!nodes)throw 0;
	}
	catch(...)
	{
		throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
	}
	return nodes;
}

