#include "Parser.hpp"
#include <ESDOM/API.hpp>
#include <ESDOM/Document.hpp>
#include <ESDOM/NodeList.hpp>
#include <ESDOM/NamedNodeMap.hpp>
#include <ESDOM/ProcessingInstruction.hpp>
#include <ESDOM/Text.hpp>
#include <ESDOM/Comment.hpp>
#include <ESDOM/CDATASection.hpp>
#include <ESDOM/Initializer.hpp>
#include <stdio.h>
#include <string.h>

#include "DocumentInst.hpp"
#include "DocumentTypeInst.hpp"
#include "ElementInst.hpp"
#include "CommentInst.hpp"
#include "TextInst.hpp"
#include "CDATASectionInst.hpp"
#include "ProcessingInstructionInst.hpp"

using namespace ESDOM;

inline void Parser::SetErrorMessage(const char* string)
{
	m_errorMessage = (string);
	m_errorMessage += " : line ";
	m_errorMessage += (unsigned)m_lineNumber;
	m_errorMessage += " column ";
	m_errorMessage += (unsigned)m_columnNumber;
}

Document Parser::parse(ESDOM_FILE FileHandle)
{
	if( !Initializer::initialized() )
	{
		m_errorMessage = "ESDOM has not initialized";
		return 0;
	}

	if( !FileHandle )
		return 0;

	if( ESDOM_isRegFile(FileHandle) )
	{
		long fptr = ESDOM_tell(FileHandle);
		ESDOM_seek( FileHandle, 0, FILE_SEEK_SET );
		m_src = FileHandle;
		Document document = parse();
		ESDOM_seek( FileHandle, fptr, FILE_SEEK_SET );
		return document;
	}
	m_src = FileHandle;
	return parse();
}

/*
Document Parser::parseFromBuffer(const unsigned char *data, unsigned long size)
{
	if( !data )
		return 0;
//	return _parseXML( (void*)FileHandle, SRC_BUFFER, size );
}

*/
Document Parser::parse()
{
	ESDOM_CHAR	buffer[5];
	ESDOM_CHAR	ch;

	m_lineNumber = 1;
	m_columnNumber = 1;
	m_charPrev = '\0';

	try
	{
		Document document( new DocumentInst );
		unsigned char sample[5];
		read( sample, 2 );
		if( sample[0] == 0xfe && sample[1] == 0xff )
		{ // UTF-16 BigEndian
			m_encoding = UTF_16BE;
			m_columnNumber++;
			readChars( buffer, 5 );
			if( memcmp( buffer, _ESDOM_TEXT("<?xml"), sizeof(ESDOM_CHAR)*5 ) )
			{
				m_errorMessage = "document is not XML";
				return 0;
			}
		}
		else if( sample[0] == 0xff && sample[1] == 0xfe )
		{ // UTF-16 LittleEndian
			m_encoding = UTF_16LE;
			m_columnNumber++;
			readChars( buffer, 5 );
			if( memcmp( buffer, _ESDOM_TEXT("<?xml"), sizeof(ESDOM_CHAR)*5 ) )
			{
				m_errorMessage = "document is not XML";
				return 0;
			}
		}
		else if( sample[0] == 0x00 && sample[1] == 0x3c )
		{ // UTF-16 BigEndian without BOM
			m_encoding = UTF_16BE;
			m_columnNumber++;
			readChars( buffer, 4 );
			if( memcmp( buffer, _ESDOM_TEXT("?xml"), sizeof(ESDOM_CHAR)*4 ) )
			{
				m_errorMessage = "document is not XML";
				return 0;
			}
		}
		else if( sample[0] == 0x3c && sample[1] == 0x00 )
		{ // UTF-16 LittleEndian without BOM
			m_encoding = UTF_16LE;
			m_columnNumber++;
			readChars( buffer, 4 );
			if( memcmp( buffer, _ESDOM_TEXT("?xml"), sizeof(ESDOM_CHAR)*4 ) )
			{
				m_errorMessage = "document is not XML";
				return 0;
			}
		}
		else if( sample[0] == 0xef && sample[1] == 0xbb )
		{ // UTF-8 with BOM ?
			read( sample, 1 );
			if( sample[0] != 0xbf )
			{
				m_errorMessage = "unknown BOM";
				return 0;
			}
			m_encoding = UTF_8;
			read( sample, 5 );
			if( memcmp( sample, "<?xml", 5 ) )
			{
				m_errorMessage = "document is not XML";
				return 0;
			}
			m_columnNumber += 6;
		}
		else if( sample[0] == 0x3c && sample[1] == 0x3f )
		{ // UTF-8 without BOM
			m_encoding = UTF_8;
			read( sample, 3 );
			if( memcmp( sample, "xml", 3 ) )
			{
				m_errorMessage = "document is not XML";
				return 0;
			}
			m_columnNumber += 5;
		}
		else
		{
			m_errorMessage = "unknown BOM";
			return 0;
		}

		readChars( buffer, 1 );
		if( !ESDOMString::isSpace(buffer[0]) )
		{
			m_errorMessage = "document is not XML";
			return 0;
		}

		// XMLDecl
		Element	dummyNode( new ElementInst(_ESDOM_TEXT("dummy")) );
		ch =skipSpace();
		while( ch != '?' ) {
			parseAttribute( dummyNode, ch );
			ch =skipSpace();
		}
		((DocumentInst*)document.m_instance)->m_version = dummyNode.getAttribute(_ESDOM_TEXT("version"));
		if( m_encoding == UTF_8 )
		{
			ESDOMString enc = dummyNode.getAttribute(_ESDOM_TEXT("encoding"));
			if( (bool)enc && strcasecmp( enc, _ESDOM_TEXT("UTF-8") ) )
			{
				m_errorMessage = "unknown encoding type";
				return 0;
			}
		}
		if( dummyNode.getAttribute(_ESDOM_TEXT("standalone")) == _ESDOM_TEXT("yes") )
		{
			document.setStandalone(true);
		}
		else if( dummyNode.getAttribute(_ESDOM_TEXT("standalone")) == _ESDOM_TEXT("no") )
		{
			document.setStandalone(false);
		}
		readChars( &ch, 1 );
		if( ch != '>' )
		{
			m_errorMessage = "document is not XML";
			return 0;
		}

		while(1)
		{
			try{
				ch =skipSpace();
			}
			catch(ESDOM_EOF)
			{
				if( document.getDocumentElement() )
				{
					((DocumentInst*)document.m_instance)->m_valid = true;
					return document;
				}
				return 0;
			}
			if( ch != '<' )
			{
				SetErrorMessage("document is ungrammatical");
				return 0;
			}
			readChars( &ch, 1 );
			if( ch == '?' )
			{ // Processing instructions
				document.appendChild( parseProcessingInstruction());
			}
			else if( ch == '!' )
			{
				ESDOM_CHAR	name[6];
				readChars( name, 2 );
				if( name[0] == '-' && name[1] == '-' )
				{ // Comment
					document.appendChild( parseComment() );
				}
				else if(  name[0] == 'D' && name[1] == 'O' )
				{
					readChars( name, 6 );
					if( name[0] == 'C' && name[1] == 'T' && name[2] == 'Y'
						&& name[3] == 'P' && name[4] == 'E' && ESDOMString::isSpace(name[5]) )
					{ // DocumentType
						if( document.getDoctype() )
						{
							SetErrorMessage("document type already exist");
							return 0;
						}
						document.setDoctype(parseDTD());
					}
					else
					{
						SetErrorMessage("unknown node");
						return 0;
					}
				}
				else
				{
					SetErrorMessage("unknown node");
					return 0;
				}
			}
			else if( ESDOMString::isLetter(ch) || ch == '_' || ch == ':' )
			{
				if( document.getDocumentElement() )
				{
					SetErrorMessage("document element is already exist");
					return 0;
				}
				Element documentElement = parseElement(ch);
				((DocumentInst*)document.m_instance)->_setDocumentElement( (ElementInst*)documentElement.m_instance );
			}
			else
			{
				SetErrorMessage("unknown node ");
				return 0;
			}
		}
	}
	catch(ESDOMException e)
	{
		throw e;
	}
	catch(ESDOM_EOF)
	{
		return 0;
	}
	catch(...)
	{
		return 0;
	}
	return 0;
}

ESDOMString Parser::parseDTD(){
	ESDOMString doctype;
	ESDOM_CHAR	ch;

	try{
		ch = skipSpace();

		ESDOMString name;
		ch = parseName(name, ch);
		if(!(bool)name)
			throw 0;
		doctype = name;
		if(ESDOMString::isSpace(ch))
			ch = skipSpace();
		if(ch == '>')
			return doctype;
		if(ch != '['){
			ch = parseName(name, ch);
			if(name == _ESDOM_TEXT("PUBLIC")){
				if(!ESDOMString::isSpace(ch))
					throw 0;
				doctype += _ESDOM_TEXT(" PUBLIC ");
				ch = skipSpace();
				if(ch != '"' && ch != '\'')
					throw 0;
				ESDOM_CHAR quot = ch;
				doctype += ch;
				readChars(&ch, 1);
				while(ch != quot){
					if(ch != 0x20 && ch != 0xD && ch != 0xA &&
						!( 'a' <= ch && ch <= 'z') && !( '?' <= ch && ch <= 'Z') && 
						!( '\'' <= ch && ch <= ';' ) && ch != '!' && ch != '=' &&
						ch != '#' && ch != '$' && ch != '%' && ch != '_' )
						throw 0;
					doctype += ch;
					readChars(&ch, 1);
				}
				doctype += ch;
				readChars(&ch, 1);
				doctype += _ESDOM_TEXT(" ");
			}
			else if(name != _ESDOM_TEXT("SYSTEM"))
				throw 0;
			if(!ESDOMString::isSpace(ch))
				throw 0;
			ch = skipSpace();
			if(ch != '\'' && ch != '"')
				throw 0;
			ESDOM_CHAR quot = ch;
			doctype += ch;
			readChars(&ch, 1);
			while(ch != quot){
				doctype += ch;
				readChars(&ch, 1);
			}
			doctype += quot;
			ch = skipSpace();
			if(ch == '>')
				return doctype;
			if(ch != '[')
				throw;
		}
		doctype += _ESDOM_TEXT("[");
		ch = skipSpace();
		while(ch != ']'){
			if(ch == '%'){
				readChars(&ch, 1);
				ch = parseName(name, ch);
				if(!(bool)name || ch != ';')
					throw 0;
				doctype += _ESDOM_TEXT("\r\n\t%") + name + _ESDOM_TEXT(";");
			}else if(ch == '<'){
				readChars(&ch, 1);
				if(ch == '?'){
					ProcessingInstruction pi = parseProcessingInstruction();
					doctype += _ESDOM_TEXT("\r\n\t<?") + pi.getNodeName();
					if(pi.getNodeValue())
						doctype += _ESDOM_TEXT(" ") + pi.getNodeValue();
					doctype += _ESDOM_TEXT("?>");
				}else if(ch == '!'){
					readChars(&ch, 1);
					if(ch == '-'){
						readChars(&ch, 1);
						if(ch == '-'){
							Comment comment = parseComment();
							doctype += _ESDOM_TEXT("\r\n\t<!--")
								+ comment.getNodeValue() + _ESDOM_TEXT("-->");
						}else
							throw 0;
					}else{
						ch = parseName(name, ch);
						if(!ESDOMString::isSpace(ch) || !(bool)name)
							throw 0;
						ch = skipSpace();
						doctype += _ESDOM_TEXT("\r\n\t<!") + name + _ESDOM_TEXT(" ");
						if(name == _ESDOM_TEXT("ELEMENT")){
							while(ch != '>'){
								doctype += ch;
								readChars(&ch, 1);
							}
							doctype += ch;

						}else if(name == _ESDOM_TEXT("ATTLIST")){
							while(ch != '>'){
								doctype += ch;
								readChars(&ch, 1);
							}
							doctype += ch;

						}else if(name == _ESDOM_TEXT("ENTITY")){
							bool fEntityDef = false;
							if(ch == '%'){
								readChars(&ch, 1);
								if(!ESDOMString::isSpace(ch))
									throw 0;
								ch = skipSpace();
								doctype += _ESDOM_TEXT("% ");
								fEntityDef = true;
							}
							ch = parseName(name, ch);
							if(!ESDOMString::isSpace(ch))
								throw 0;
							ch = skipSpace();
							doctype += name;
							doctype += _ESDOM_TEXT(" ");
							if(ch == '"' || ch == '\''){
								ESDOM_CHAR quot = ch;
								do{
									doctype += ch;
									readChars(&ch, 1);
								}while(ch != quot);
								doctype += quot;
								ch = skipSpace();
								if(ch != '>')
									throw 0;
							}else{
								if(ch != 'S' && ch != 'D')
									throw 0;
								ch = parseName(name, ch);
								if(!ESDOMString::isSpace(ch))
									throw 0;
								doctype += name;
								doctype += _ESDOM_TEXT(" ");
								if(name == _ESDOM_TEXT("PUBLIC")){
									ch = skipSpace();
									if(ch != '"' && ch != '\'')
										throw 0;
									ESDOM_CHAR quot = ch;
									doctype += ch;
									readChars(&ch, 1);
									while(ch != quot){
										if(ch != 0x20 && ch != 0xD && ch != 0xA &&
											!( 'a' <= ch && ch <= 'z') && !( '?' <= ch && ch <= 'Z') && 
											!( '\'' <= ch && ch <= ';' ) && ch != '!' && ch != '=' &&
											ch != '#' && ch != '$' && ch != '%' && ch != '_' )
											throw 0;
										doctype += ch;
										readChars(&ch, 1);
									}
									doctype += ch;
									readChars(&ch, 1);
									if(!ESDOMString::isSpace(ch))
										throw 0;
									doctype += _ESDOM_TEXT(" ");
								}else if(name != _ESDOM_TEXT("SYSTEM"))
									throw 0;
								ch = skipSpace();
								if(ch != '\'' && ch != '"')
									throw 0;
								ESDOM_CHAR quot = ch;
								doctype += ch;
								readChars(&ch, 1);
								while(ch != quot){
									doctype += ch;
									readChars(&ch, 1);
								}
								doctype += quot;
								if(ESDOMString::isSpace(ch))
									ch = skipSpace();
								if(fEntityDef){
									while(ch != '>'){
										doctype += ch;
										readChars(&ch, 1);
									}
								}
								if(ch != '>')
									throw 0;
								doctype += _ESDOM_TEXT(">");
							}
							doctype += _ESDOM_TEXT(">");

						}else if(name == _ESDOM_TEXT("NOTATION")){
							ch = parseName(name, ch);
							if(!ESDOMString::isSpace(ch) || !(bool)name)
								throw 0;
							doctype += name + _ESDOM_TEXT(" ");
							ch = parseName(name, ch);
							if(!ESDOMString::isSpace(ch) || !(bool)name)
								throw 0;
							ch = skipSpace();
							doctype += name + _ESDOM_TEXT(" ");
							if(name == _ESDOM_TEXT("PUBLIC")){
								if(ch != '"' && ch != '\'')
									throw 0;
								ESDOM_CHAR quot = ch;
								doctype += ch;
								readChars(&ch, 1);
								while(ch != quot){
									if(ch != 0x20 && ch != 0xD && ch != 0xA &&
										!( 'a' <= ch && ch <= 'z') && !( '?' <= ch && ch <= 'Z') && 
										!( '\'' <= ch && ch <= ';' ) && ch != '!' && ch != '=' &&
										ch != '#' && ch != '$' && ch != '%' && ch != '_' )
										throw 0;
									doctype += ch;
									readChars(&ch, 1);
								}
								doctype += ch;
								ch = skipSpace();
								if(ch == '\'' || ch == '"'){
									doctype += _ESDOM_TEXT(" ");
									doctype += ch;
									quot = ch;
									do{
										readChars(&ch, 1);
										doctype += ch;
									}while(ch != quot);
									ch = skipSpace();
								}
								if(ch != '>')
									throw 0;
							}else if(name == _ESDOM_TEXT("SYSTEM")){
								if(ch != '"' && ch != '\'')
									throw 0;
								ESDOM_CHAR quot = ch;
								do{
									readChars(&ch, 1);
									doctype += ch;
								}while(ch != quot);
							}
							doctype += _ESDOM_TEXT(">");
						}else
							throw 0;
					}
				} // end of >> else if(else if(ch == '!'){ <<
			}else
				throw 0;
			ch = skipSpace();
		}
		ch = skipSpace();
		doctype += _ESDOM_TEXT("\r\n]");
	}catch(int){
		SetErrorMessage("Invalid DTD");
		throw 0;
	}
	return doctype;
}

ProcessingInstruction Parser::parseProcessingInstruction()
{
	ESDOM_CHAR		ch;
	ESDOMString	target;

	readChars(&ch, 1);
	ch = parseName(target, ch);

	if(!(bool)target)
	{
		SetErrorMessage("Invalid ProcessingInstructionInst target");
		throw 0;
	}

	if( !strcasecmp( target, _ESDOM_TEXT("xml") ) )
	{
		SetErrorMessage("invalid ProcessingInstructionInst");
		throw 0;
	}

	ProcessingInstruction newPI( new ProcessingInstructionInst(target) );

	ESDOMString data;
	if(ESDOMString::isSpace(ch))
		ch = skipSpace();

	while(true)
	{
		if( ch == '?' )
		{
			readChars(&ch, 1);
			if( ch == '>' )
				break;
			data += (ESDOM_CHAR)'?';
		}
//		if( !String::isChar(ch) )
//		{
//			SetErrorMessage("Invalid PI value");
//			throw 0;
//		}
		data += ch;
		readChars(&ch, 1);
	};

	newPI.setNodeValue( data );
	return newPI;
}

Element Parser::parseElement(ESDOM_CHAR tagInitial)
{
	if( !ESDOMString::isLetter( tagInitial ) && tagInitial != '_' && tagInitial != ':' )
	{
		SetErrorMessage("invalid tag name");
		throw 0;
	}

	ESDOMString	tagName;
	ESDOM_CHAR		ch;
	ch = parseName(tagName, tagInitial);

	if( !(bool)tagName || ( !ESDOMString::isSpace(ch) && ch != '/' && ch != '>' ) )
	{
		SetErrorMessage("invalid element tagname");
		throw 0;
	}

	Element element( new ElementInst(tagName) );

	if( ESDOMString::isSpace(ch) )
		ch = skipSpace();
	while( ch != '/' && ch != '>' )
	{
		parseAttribute( element, ch);
		ch = skipSpace();
	}
	if( ch == '/' )
	{
		readChars(&ch, 1);
		if( ch == '>' )
			return element;
		SetErrorMessage("invalid element tag");
		throw 0;
	}
	readChars( &ch, 1 );
	while(true)
	{
		if( ch == '<' )
		{
			readChars( &ch, 1 );
			if( ch == '/' )
			{
				ESDOM_CHAR	*tagBuffer;
				unsigned long length = tagName.getLength();
				try{
					tagBuffer = new ESDOM_CHAR[length+1];
					if(!tagBuffer)throw 0;
				}catch(...){
					throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
				}
				readChars(tagBuffer, length);
				tagBuffer[length] = 0;
				ch = skipSpace();
				if( tagName != tagBuffer || ch != '>' )
				{
					delete[] tagBuffer;
					SetErrorMessage("invalid close tag");
					throw 0;
				}
				delete[] tagBuffer;
				return element;
			}
			else if( ch == '?' )
			{ // Processing instructions
				element.appendChild( parseProcessingInstruction() );
			}
			else if( ch == '!' )
			{
				ESDOM_CHAR	name[5];
				readChars( name, 2 );
				if( name[0] == '-' && name[1] == '-' )
				{ // Comment
					element.appendChild( parseComment() );
				}
				else if(  name[0] == '[' && name[1] == 'C' )
				{
					readChars( name, 5 );
					if( name[0] == 'D' && name[1] == 'A' && name[2] == 'T'
						&& name[3] == 'A' && name[4] == '['  )
					{ // CDATASection
						element.appendChild( parseCDATASection() );
					}
					else
					{
						SetErrorMessage("unknown node");
						throw 0;
					}
				}
				else
				{
					SetErrorMessage("unknown node");
					throw 0;
				}
			}
			else if( ESDOMString::isLetter(ch) || ch == '_' || ch == ':' )
			{
				element.appendChild( parseElement(ch) );
			}
			else
			{
				SetErrorMessage("unknown node");
				throw 0;
			}
		}
		else
		{
			ESDOMString	textdata;
			
			do
			{
				if( ch == '&' )
				{
					ESDOMString ret;
					ch = entityReference2char( ret );
					textdata += ret;
				}
				else
				{
					textdata += ch;
					readChars( &ch, 1 );
				}
			}while( ch != '<' );
			Text text( new TextInst );
			text.setNodeValue(textdata);
			element.appendChild( text );
			continue;
		}
		readChars( &ch, 1 );
	}
	return 0;
}

void Parser::parseAttribute(Element &element, ESDOM_CHAR initial)
{
	if( !ESDOMString::isLetter( initial ) && initial != '_' && initial != ':' )
	{
		SetErrorMessage("invalid attribute name");
		throw 0;
	}

	ESDOM_CHAR	ch, quot;
	ESDOMString	name, value;
	ch = parseName(name, initial);

	if(!(bool)name || ( !ESDOMString::isSpace(ch) && ch != '=' ))
	{
		SetErrorMessage("invalid attribute name");
		throw 0;
	}

	if( ch != '=' )
		ch = skipSpace();
	if( ch != '=' )
	{
		SetErrorMessage("invalid attribute");
		throw 0;
	}
	quot = ch = skipSpace();
	if( quot != '\'' && quot != '"' )
	{
		SetErrorMessage("invalid attribute value");
		throw 0;
	}
	readChars( &ch, 1 );
	while( ch != quot )
	{
		if( ch == '<' )
		{
			SetErrorMessage("invalid attribute value");
			throw 0;
		}
		if( ch == '&' )
		{
			ESDOMString ret;
			ch = entityReference2char( ret );
			value += ret;
			continue;
		}
		value += ch;
		readChars( &ch, 1 );
	}

	element.setAttribute( name, value );
}

Comment Parser::parseComment()
{
	ESDOMString	data;
	ESDOM_CHAR		chs[3];

	readChars( chs, 3 );

	while(chs[0] != '-' || chs[1] != '-' || chs[2] != '>')
	{
		data += chs[0];
		chs[0] = chs[1];
		chs[1] = chs[2];
		readChars( &chs[2], 1 );
	}
	Comment comment( new CommentInst );
	comment.setNodeValue(data);
	return comment;
}

CDATASection Parser::parseCDATASection()
{
	ESDOMString	data;
	ESDOM_CHAR		CDEnd[3] = {0,0,0};

	readChars( CDEnd, 3 );

	while(CDEnd[0] != ']' || CDEnd[1] != ']' || CDEnd[2] != '>')
	{
		data += CDEnd[0];
		CDEnd[0] = CDEnd[1];
		CDEnd[1] = CDEnd[2];
		readChars( &CDEnd[2], 1 );
	}
	CDATASection cdata( new CDATASectionInst );
	cdata.setNodeValue(data);
	return cdata;
}

ESDOM_CHAR Parser::skipSpace()
{
	ESDOM_CHAR	buffer;
	do
		readChars( &buffer, 1 );
	while( ESDOMString::isSpace(buffer) );
	return buffer;
}

ESDOM_CHAR Parser::parseName(ESDOMString &name, ESDOM_CHAR initial)
{
	ESDOM_CHAR	ch;
	name = initial;
	if( !ESDOMString::isLetter(initial) && initial != '_' && initial != ':' )
		return initial;

	readChars(&ch, 1);
	while( ESDOMString::isNameChar(ch) )
	{
		name += ch;
		readChars(&ch, 1);
	}
	return ch;
}

ESDOM_CHAR Parser::entityReference2char(ESDOMString &ret)
{
	ESDOM_CHAR	ch;
	readChars(&ch,1);
	if( ch == '#' )
	{
		readChars(&ch, 1);
		if(ch == 'x')
		{
			ESDOM_CHAR	num = 0;
			unsigned	length = 0;
			readChars(&ch, 1);
			while(ch != ';')
			{
				if( '0' <= ch && ch <= '9' )
					num = (ESDOM_CHAR)(num*16 + ( ch - '0' ));
				else if( 'a' <= ch && ch <= 'f' )
					num = (ESDOM_CHAR)(num*16 + ( ch +10 -'a' ));
				else if( 'A' <= ch && ch <= 'F' )
					num = (ESDOM_CHAR)(num*16 + ( ch +10 -'A' ));
				else
				{
					SetErrorMessage("invalid EntityReference");
					break;
				}
				length++;
				readChars(&ch, 1);
				if(length == 4)
				{
					readChars(&ch, 1);
					break;
				}
			}
			ret = num;
			return ch;
		}
		else
		{
			ESDOM_CHAR	num = 0;
			unsigned	length = 0;
			readChars(&ch, 1);
			while(ch != ';')
			{
				if( '0' <= ch && ch <= '9' )
					num = (ESDOM_CHAR)(num*10 + ( ch - '0' ));
				else
				{
					SetErrorMessage("invalid EntityReference");
					break;
				}
				length++;
				readChars(&ch, 1);
				if(length == 5)
				{
					readChars(&ch, 1);
					break;
				}
			}
			ret = num;
			return ch;
		}
	}
	else if( ESDOMString::isLetter(ch) || ch == '_' || ch == ':' )
	{
		ESDOMString name = ch;
		readChars(&ch, 1);
		while(ch != ';')
		{
			if( !ESDOMString::isNameChar(ch) )
			{
				SetErrorMessage("invalid EntityReference");
				ret = name;
				return ch;
			}
			name += ch;
			readChars(&ch, 1);
		}
		char *name_ansi;
		try{
			name_ansi = new char[name.getLength()+1];
			if(!name_ansi)throw 0;
		}catch(...){
			throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
		}
		const ESDOM_CHAR* name_ptr = name;
		char* name_ansi_ptr = name_ansi;
		while(*name_ptr)
		{
			if( *name_ptr > 127)
				break;
			*name_ansi_ptr++ = (char)*name_ptr++;
		}
		ENTITIES	*eptr;
		if( *name_ptr )
			eptr = 0;
		else
		{
			*name_ansi_ptr = 0;
			eptr = in_word_set(name_ansi, name.getLength());
		}

		delete[] name_ansi;
		if( eptr )
		{
			ret = eptr->value;
		}
		else
		{
			ret = _ESDOM_TEXT("&");
			ret += name;
			ret += _ESDOM_TEXT(";");
		}
		readChars(&ch, 1);
		return ch;
	}
	else
	{
		SetErrorMessage("invalid EntityReference");
		throw 0;
	}
}

void Parser::readChars(ESDOM_CHAR *buffer, unsigned long length)
{
	unsigned char chars[3];
//	unsigned char chars[6]; // for UCS4

	switch(m_encoding)
	{
	case UTF_8:
		for( ; length; length--, buffer++ )
		{
			read( &chars[0], 1 );
			if( chars[0] < 0x80 )
			{
				*buffer = chars[0];
				if(m_charPrev == '\r')
				{
					m_lineNumber++;
					m_columnNumber = 1;
				}
				else
				{
					if(*buffer == '\n')
					{
						m_lineNumber++;
						m_columnNumber = 1;
					}
					else
						m_columnNumber++;
				}
				m_charPrev = *buffer;
				continue;
			}

			if( chars[0] < 0xc2 )
				throw 0;

			read( &chars[1], 1 );
			if( chars[0] < 0xe0 )
			{
				if( chars[1] < 0x80 || 0xbf < chars[1] )
					throw 0;
				*buffer = (ESDOM_CHAR)(( ( chars[0] & 0x1f ) << 6 ) + ( chars[1] & 0x3f ));
				if(m_charPrev == '\r')
				{
					m_lineNumber++;
					m_columnNumber = 1;
				}
				else
					m_columnNumber++;
				m_charPrev = *buffer;
				continue;
			}

			read( &chars[2], 1 );
			if( chars[0] < 0xf0 )
			{
				if( ( chars[0] == 0xe0 && chars[1] < 0xa0 ) || ( chars[0] != 0xe0 && chars[1] < 0x80 ) )
					throw 0;
				if( 0xbf < chars[1] || chars[2] < 0x80 || 0xbf < chars[2] )
					throw 0;
				*buffer = (ESDOM_CHAR)(( ( chars[0] & 0x0f ) << 12 ) + ( ( chars[1] & 0x3f ) << 6 ) + ( chars[2] & 0x3f ));
				if(m_charPrev == '\r')
				{
					m_lineNumber++;
					m_columnNumber = 1;
				}
				else
					m_columnNumber++;
				m_charPrev = *buffer;
				continue;
			}

/*			read( &chars[3], 1 );
			if( chars[0] == 0xf7 )
			{
				*buffer = (ESDOM_CHAR)(( ( chars[0] & 0x0c ) << 18 ) + ( ( chars[1] & 0x3f ) << 12 )
						+ ( ( chars[2] & 0x3f ) << 6 ) + ( chars[3] & 0x3f ));
				if(m_charPrev == '\r')
				{
					m_lineNumber++;
					m_columnNumber = 1;
				}
				else
					m_columnNumber++;
				m_charPrev = *buffer;
				continue;
			}

			read( &chars[4], 1 );
			if( chars[0] < 0xfc )
			{
				*buffer = (ESDOM_CHAR)(( ( chars[0] & 0x03 ) << 24 ) + ( ( chars[1] & 0x3f ) << 18 )
						+ ( ( chars[2] & 0x3f ) << 12 ) + ( ( chars[3] & 0x3f ) << 6 ) + ( chars[4] & 0x3f );
				if(m_charPrev == '\r')
				{
					m_lineNumber++;
					m_columnNumber = 1;
				}
				else
					m_columnNumber++;
				m_charPrev = *buffer;
				continue;
			}

			read( &chars[5], 1 );
			if( chars[0] < 0xfe )
			{
				*buffer = (ESDOM_CHAR)(( ( chars[0] & 0x01 ) << 30 ) + ( ( chars[1] & 0x3f ) << 24 ) + ( ( chars[2] & 0x3f ) << 18 )
						+ ( ( chars[3] & 0x3f ) << 12 ) + ( ( chars[4] & 0x3f ) << 6 ) + ( chars[5] & 0x3f ));
				if(m_charPrev == '\r')
				{
					m_lineNumber++;
					m_columnNumber = 1;
				}
				else
					m_columnNumber++;
				m_charPrev = *buffer;
				continue;
			}
*/
			throw 0;
		}
		return;

	case UTF_16BE:
		for( ; length; length--, buffer++ )
		{
			read( chars, 2 );
			*buffer = (ESDOM_CHAR)(chars[0]*0x100 + chars[1]);
			if(m_charPrev == '\r')
			{
				m_lineNumber++;
				m_columnNumber = 1;
			}
			else
			{
				if(*buffer == '\n')
				{
					m_lineNumber++;
					m_columnNumber = 1;
				}
				else
					m_columnNumber++;
			}
			m_charPrev = *buffer;
		}
		return;

	case UTF_16LE:
		for( ; length; length--, buffer++ )
		{
			read( chars, 2 );
			*buffer = (ESDOM_CHAR)(chars[1]*0x100 + chars[0]);
			if(m_charPrev == '\r')
			{
				m_lineNumber++;
				m_columnNumber = 1;
			}
			else
			{
				if(*buffer == '\n')
				{
					m_lineNumber++;
					m_columnNumber = 1;
				}
				else
					m_columnNumber++;
			}
			m_charPrev = *buffer;
		}
		return;
	}

	throw 0;
}

void Parser::read(unsigned char *buffer, unsigned long size)
{
	if( (FILE*)m_src == stdin )
	{
		if( fread( buffer, 1, size, (FILE*)m_src ) == size )
			return;
		throw ESDOM_EOF();
	}
	if( readFromFile( (ESDOM_FILE)m_src, buffer, size ) == size )
		return;
	throw ESDOM_EOF();
}

//////////////////////////////////////////////////////////////////////////////////////////

/* C code produced by gperf version 2.7.2 */
/* Command-line: gperf -ptD word.key  */

#define TOTAL_KEYWORDS 253
#define MIN_WORD_LENGTH 2
#define MAX_WORD_LENGTH 8
#define MIN_HASH_VALUE 5
#define MAX_HASH_VALUE 508
/* maximum key range = 504, duplicates = 43 */

#ifdef __GNUC__
__inline
#else
#ifdef __cplusplus
inline
#endif
#endif
unsigned int
ESDOM::Parser::hash (register const char *str, register unsigned int len)
{
  static unsigned short asso_values[] =
    {
      509, 509, 509, 509, 509, 509, 509, 509, 509, 509,
      509, 509, 509, 509, 509, 509, 509, 509, 509, 509,
      509, 509, 509, 509, 509, 509, 509, 509, 509, 509,
      509, 509, 509, 509, 509, 509, 509, 509, 509, 509,
      509, 509, 509, 509, 509, 509, 509, 509, 509,  10,
      110,  45,   0, 509, 509, 509, 509, 509, 509, 509,
      509, 509, 509, 509, 509, 185,  30,  75, 125,  85,
      509,  30,  10, 205, 509,  70,  15,   0, 145, 250,
        0, 509,   5,  95,  85, 190, 509, 509,   5,  20,
       10, 509, 509, 509, 509, 509, 509, 155, 225, 252,
      165,   0,   5, 245, 122,  75,   0,  90, 105, 242,
       10,  45,  55,  15, 230, 130,  15, 200,  10,   0,
       15, 195,  15, 509, 509, 509, 509, 509, 509, 509,
      509, 509, 509, 509, 509, 509, 509, 509, 509, 509,
      509, 509, 509, 509, 509, 509, 509, 509, 509, 509,
      509, 509, 509, 509, 509, 509, 509, 509, 509, 509,
      509, 509, 509, 509, 509, 509, 509, 509, 509, 509,
      509, 509, 509, 509, 509, 509, 509, 509, 509, 509,
      509, 509, 509, 509, 509, 509, 509, 509, 509, 509,
      509, 509, 509, 509, 509, 509, 509, 509, 509, 509,
      509, 509, 509, 509, 509, 509, 509, 509, 509, 509,
      509, 509, 509, 509, 509, 509, 509, 509, 509, 509,
      509, 509, 509, 509, 509, 509, 509, 509, 509, 509,
      509, 509, 509, 509, 509, 509, 509, 509, 509, 509,
      509, 509, 509, 509, 509, 509, 509, 509, 509, 509,
      509, 509, 509, 509, 509, 509
    };
  return len + asso_values[(unsigned char)str[len - 1]] + asso_values[(unsigned char)str[0]];
}

#ifdef __GNUC__
__inline
#endif
ESDOM::Parser::ENTITIES *
ESDOM::Parser::in_word_set (register const char *str, register unsigned int len)
{
  static ENTITIES
 wordlist[] =
    {
		{"Prime",	0x2033},	{"egrave",	0x00e8},	{"eacute",	0x00e9},	{"frac14",	0x00bc},
		{"frac34",	0x00be},	{"ne",		0x2260},	{"fnof",	0x0192},	{"equiv",	0x2261},
		{"ntilde",	0x00f1},	{"epsilon",	0x03b5},	{"zwj",		0x200d},	{"zwnj",	0x200c},
		{"tilde",	0x02dc},	{"exist",	0x2203},	{"trade",	0x2122},	{"there4",	0x2234},
		{"notin",	0x2209},	{"Yacute",	0x00dd},	{"not",		0x00ac},	{"thorn",	0x00fe},
		{"quot",	0x0022},	{"euro",	0x20ac},	{"oline",	0x203e},	{"ograve",	0x00f2},
		{"otilde",	0x00f5},	{"oacute",	0x00f3},	{"Rho",		0x03a1},	{"ordf",	0x00aa},
		{"ensp",	0x2002},	{"emsp",	0x2003},	{"prime",	0x2032},	{"weierp",	0x2118},
		{"omicron",	0x03bf},	{"piv",		0x03d6},	{"nbsp",	0x00a0},	{"plusmn",	0x00b1},
		{"part",	0x2202},	{"thinsp",	0x2009},	{"Pi",		0x03a0},	{"Phi",		0x03a6},
		{"Psi",		0x03a8},	{"image",	0x2111},	{"igrave",	0x00ec},	{"iacute",	0x00ed},
		{"Xi",		0x039e},	{"ni",		0x220b},	{"isin",	0x2208},	{"infin",	0x221e},
		{"Egrave",	0x00c8},	{"Eacute",	0x00c9},	{"xi",		0x03be},	{"int",		0x222b},
		{"iquest",	0x00bf},	{"ETH",		0x00d0},	{"Epsilon",	0x0395},	{"le",		0x2264},
		{"euml",	0x00eb},	{"Scaron",	0x0160},	{"prop",	0x221d},	{"perp",	0x22a5},
		{"frasl",	0x2044},	{"forall",	0x2200},	{"frac12",	0x00bd},	{"lt",		0x003c},
		{"loz",		0x25ca},	{"eth",		0x00f0},	{"lowast",	0x2217},	{"Yuml",	0x0178},
		{"pi",		0x03c0},	{"phi",		0x03c6},	{"psi",		0x03c8},	{"sube",	0x2286},
		{"supe",	0x2287},	{"ndash",	0x2013},	{"sigmaf",	0x03c2},	{"sup1",	0x00b9},
		{"scaron",	0x0161},	{"sect",	0x00a7},	{"sdot",	0x22c5},	{"times",	0x00d7},
		{"Ntilde",	0x00d1},	{"Chi",		0x03a7},	{"ouml",	0x00f6},	{"laquo",	0x00ab},
		{"ldquo",	0x201c},	{"lsquo",	0x2018},	{"lsaquo",	0x2039},	{"eta",		0x03b7},
		{"acute",	0x00b4},	{"agrave",	0x00e0},	{"atilde",	0x00e3},	{"aacute",	0x00e1},
		{"permil",	0x2030},	{"Zeta",	0x0396},	{"nabla",	0x2207},	{"divide",	0x00f7},
		{"oslash",	0x00f8},	{"zeta",	0x03b6},	{"theta",	0x03b8},	{"Lambda",	0x039b},
		{"sup3",	0x00b3},	{"sbquo",	0x201a},	{"oplus",	0x2295},	{"otimes",	0x2297},
		{"hellip",	0x2026},	{"iuml",	0x00ef},	{"iexcl",	0x00a1},	{"Ccedil",	0x00c7},
		{"sup",		0x2283},	{"Beta",	0x0392},	{"Gamma",	0x0393},	{"Agrave",	0x00c0},
		{"Atilde",	0x00c3},	{"Aacute",	0x00c1},	{"Euml",	0x00cb},	{"Ugrave",	0x00d9},
		{"Uacute",	0x00da},	{"empty",	0x2205},	{"yacute",	0x00fd},	{"Mu",		0x039c},
		{"omega",	0x03c9},	{"ugrave",	0x00f9},	{"uacute",	0x00fa},	{"Upsilon",	0x03a5},
		{"yen",		0x00a5},	{"Igrave",	0x00cc},	{"Iacute",	0x00cd},	{"nu",		0x03bd},
		{"amp",		0x0026},	{"para",	0x00b6},	{"asymp",	0x2248},	{"lceil",	0x2308},
		{"upsilon",	0x03c5},	{"tau",		0x03c4},	{"prod",	0x220f},	{"pound",	0x00a3},
		{"Kappa",	0x039a},	{"iota",	0x03b9},	{"THORN",	0x00de},	{"nsub",	0x2284},
		{"Eta",		0x0397},	{"sup2",	0x00b2},	{"Theta",	0x0398},	{"ge",		0x2265},
		{"kappa",	0x03ba},	{"Sigma",	0x03a3},	{"Ograve",	0x00d2},	{"Otilde",	0x00d5},
		{"Oacute",	0x00d3},	{"ecirc",	0x00ea},	{"hearts",	0x2665},	{"gt",		0x003e},
		{"middot",	0x00b7},	{"auml",	0x00e4},	{"thetasym",0x03d1},	{"talefsym",0x2135},
		{"lambda",	0x03bb},	{"spades",	0x2660},	{"Omicron",	0x039f},	{"curren",	0x00a4},
		{"cent",	0x00a2},	{"bdquo",	0x201e},	{"or",		0x2228},	{"rho",		0x03c1},
		{"raquo",	0x00bb},	{"rdquo",	0x201d},	{"rsquo",	0x2019},	{"rsaquo",	0x203a},
		{"Delta",	0x0394},	{"Tau",		0x03a4},	{"apos",	0x0027},	{"sigma",	0x03c3},
		{"ordm",	0x00ba},	{"micro",	0x00b5},	{"Auml",	0x00c4},	{"oelig",	0x0153},
		{"Uuml",	0x00dc},	{"diams",	0x2666},	{"ocirc",	0x00f4},	{"yuml",	0x00ff},
		{"uml",		0x00a8},	{"uuml",	0x00fc},	{"cap",		0x2229},	{"cup",		0x222a},
		{"Iuml",	0x00cf},	{"alpha",	0x03b1},	{"and",		0x2227},	{"delta",	0x03b4},
		{"upsih",	0x03d2},	{"shy",		0x00ad},	{"chi",		0x03c7},	{"icirc",	0x00ee},
		{"bull",	0x2022},	{"real",	0x211c},	{"lArr",	0x21d0},	{"larr",	0x2190},
		{"rceil",	0x2309},	{"lfloor",	0x230a},	{"Ecirc",	0x00ca},	{"Alpha",	0x0391},
		{"Nu",		0x039d},	{"lrm",		0x200e},	{"lang",	0x2329},	{"harr",	0x2194},
		{"hArr",	0x21d4},	{"sub",		0x2282},	{"Ouml",	0x00d6},	{"Dagger",	0x2021},
		{"cedil",	0x00b8},	{"ccedil",	0x00e7},	{"Iota",	0x0399},	{"mdash",	0x2014},
		{"sum",		0x2211},	{"sim",		0x223c},	{"minus",	0x2212},	{"Oslash",	0x00d8},
		{"szlig",	0x00df},	{"beta",	0x03b2},	{"clubs",	0x2663},	{"darr",	0x2193},
		{"dArr",	0x21d3},	{"dagger",	0x2020},	{"ang",		0x2220},	{"aring",	0x00e5},
		{"gamma",	0x03b3},	{"aelig",	0x00e6},	{"Omega",	0x03a9},	{"acirc",	0x00e2},
		{"deg",		0x00b0},	{"uarr",	0x2191},	{"uArr",	0x21d1},	{"Aring",	0x00c5},
		{"AElig",	0x00c6},	{"Acirc",	0x00c2},	{"mu",		0x03bc},	{"Ucirc",	0x00db},
		{"copy",	0x00a9},	{"ucirc",	0x00fb},	{"brvbar",	0x00a6},	{"Icirc",	0x00ce},
		{"rarr",	0x2192},	{"rArr",	0x21d2},	{"rfloor",	0x230b},	{"rlm",		0x200f},
		{"macr",	0x00af},	{"reg",		0x00ae},	{"rang",	0x232a},	{"crarr",	0x21b5},
		{"radic",	0x221a},	{"OElig",	0x0152},	{"cong",	0x2245},	{"Ocirc",	0x00d4},
		{"circ",	0x02c6}
    };

  static short lookup[] =
    {
        -1,   -1,   -1,   -1,   -1,    0, -261, -252,
        -2,   -1,   -1, -285,    5,   -1,    6,    7,
         8,    9,   10,   11, -276,   15, -241,   -3,
        -1,   16,   17,   -1,   18,   -1,   19, -250,
        -2,   -1,   20,   -1,   -1,   -1,   -1,   -1,
        -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
        -1,   21,   22, -309,   -1,   26,   27, -230,
        -3,   -1,   -1, -317,   30,   31,   32, -225,
        -2,   -1,   -1,   -1,   33,   34,   -1,   35,
        -1,   -1,   36,   -1,   37,   38, -339,   -1,
        41, -337,   44, -211,   -2, -214,   -2,   45,
        -1,   46,   47, -348,   50,   51, -205,   -2,
        52,   -1,   53,   -1,   -1,   -1,   54,   -1,
        -1,   -1,   -1,   55,   -1,   56,   -1,   57,
        -1,   -1, -371,   60,   61, -195,   -2,   -1,
        -1,   62,   63,   64,   -1,   65,   66,   -1,
        -1,   67,   -1,   -1,   68, -392, -389, -182,
        -2,   73, -184,   -2,   -1,   74,   -1,   -1,
        75,   -1,   76,   -1,   -1, -421,   79,   80,
        -1,   81,   82, -418,   86,   -1,   87,   -1,
        88, -416, -164,   -3, -170,   -3,   92, -176,
        -2,   93,   94,   95,   -1,   96,   97,   98,
        99,   -1,   -1,  100, -457,  103,   -1,  104,
       105,  106,  107,   -1,  108,  109,  110, -446,
      -142,   -3,  114,   -1, -451, -138,   -2,   -1,
       117,  118,  119, -152,   -2,  120, -463,  123,
       124, -132,   -2, -475,  127,  128,  129, -473,
        -1,  132,  133, -123,   -2, -128,   -2,   -1,
       134,  135,   -1,   -1,   -1,   -1,  136,   -1,
        -1,   -1,  137,  138,   -1,   -1,   -1,  139,
        -1,   -1,   -1,  140,  141,  142,   -1,  143,
        -1,   -1,  144,   -1,   -1,   -1,   -1,  145,
      -513,  149,  150, -107,   -3,   -1,  151,  152,
       153, -526, -523,  158,  159,  -97,   -2,  160,
       -99,   -2,   -1,  161,   -1,  162,  163,   -1,
      -536,  167,  -89,   -3,   -1,  168,   -1,   -1,
       169,  170,  171,  172,  173,   -1,  174,  175,
        -1,   -1,   -1,  176,  177,   -1,  178,   -1,
       179,   -1,   -1,   -1,  180,  181, -565,  -71,
        -2,   -1,  184,  185,   -1,   -1,   -1,   -1,
        -1,   -1,   -1,  186,   -1,  187,   -1,  188,
       189,   -1,  190,   -1,  191,   -1,  192,   -1,
        -1,   -1,   -1, -597,  196,  197,  198,  -60,
        -3,  199,   -1,  200,   -1,   -1,  201,   -1,
        -1,   -1,  202,   -1, -619,   -1,  205,  206,
        -1,  207,  208,  209,  210,  -50,   -2,   -1,
        -1,  211,   -1,   -1,   -1,   -1,   -1, -635,
        -1,  214,  215,   -1,  216,  -41,   -2,   -1,
       217,   -1,   -1,  218,   -1,   -1,   -1,   -1,
        -1,   -1,   -1,   -1,   -1,   -1,   -1, -662,
        -1,  221,   -1,  222,   -1, -660,  -30,   -3,
       -34,   -2,  226,   -1,  227,  228,   -1,   -1,
        -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
        -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
        -1,   -1, -692, -690,  -22,   -2,  -24,   -2,
        -1,   -1,  233,   -1,  234,   -1,   -1,  235,
        -1,   -1,   -1,  236,   -1,   -1,   -1,   -1,
        -1,  237,   -1,   -1,   -1,  238,  239,   -1,
      -721,   -1,  242,  -13,   -2,   -1,   -1,   -1,
        -1,   -1,   -1,  243,  244,   -1,  245,  246,
        -1,   -1,   -1,   -1,   -1,   -1,   -1, -742,
        -6,   -2,   -1,   -1,   -1,   -1,   -1,   -1,
        -1,   -1,   -1,   -1,  249,  250,   -1,   -1,
        -1,   -1,   -1,  251,  252
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      register int key = hash (str, len);

      if (key <= MAX_HASH_VALUE && key >= 0)
        {
          register int index = lookup[key];

          if (index >= 0)
            {
              register const char *s = wordlist[index].name;

              if (*str == *s && !::strcmp (str + 1, s + 1))
                return &wordlist[index];
            }
          else if (index < -TOTAL_KEYWORDS)
            {
              register int offset = - 1 - TOTAL_KEYWORDS - index;
              register ENTITIES *wordptr = &wordlist[TOTAL_KEYWORDS + lookup[offset]];
              register ENTITIES *wordendptr = wordptr + -lookup[offset + 1];

              while (wordptr < wordendptr)
                {
                  register const char *s = wordptr->name;

                  if (*str == *s && !::strcmp (str + 1, s + 1))
                    return wordptr;
                  wordptr++;
                }
            }
        }
    }
  return 0;
}
