#ifndef INCLUDED_ESDOM_Parser_hpp
#define INCLUDED_ESDOM_Parser_hpp

#include <ESDOM/ESDOM_define.hpp>
#include <ESDOM/API.hpp>

namespace ESDOM
{

class ESDOM_EXPORT Element;
class ESDOM_EXPORT ProcessingInstruction;
class ESDOM_EXPORT Comment;
class ESDOM_EXPORT Text;
class ESDOM_EXPORT CDATASection;
class ESDOM_EXPORT Document;
class ESDOMString;

class ESDOM_EXPORT Parser
{
public:
	Document		parse(ESDOM_FILE FileHandle);
//	Document		parseFromBuffer(const unsigned char *data);

private:
	struct ENTITIES
	{
		const char	*name;
		ESDOM_CHAR	value;
	};
	struct ESDOM_EOF{
	};

	Document				parse();
	ESDOMString				parseDTD();
	Element					parseElement(ESDOM_CHAR initial);
	void					parseAttribute(Element &element, ESDOM_CHAR initial);
	Comment					parseComment();
	ProcessingInstruction	parseProcessingInstruction();
	CDATASection			parseCDATASection();
	ESDOM_CHAR				skipSpace();
	ESDOM_CHAR				parseName(ESDOMString &name, ESDOM_CHAR initial);
	ESDOM_CHAR				entityReference2char(ESDOMString &ret);

	static unsigned int		hash (register const char *str, register unsigned int len);
	static ENTITIES *		in_word_set (register const char *str, register unsigned int len);

	void			readChars(ESDOM_CHAR *buffer, unsigned long length);
	void			read(unsigned char *buffer, unsigned long size);

	inline void		SetErrorMessage(const char* string);

private:
	ESDOM_FILE		m_src;
	unsigned short	m_encoding;
	unsigned long	m_lineNumber;
	unsigned long	m_columnNumber;
	ESDOM_CHAR		m_charPrev;
};

}; /* namespace ESDOM */

#endif /* INCLUDED_ESDOM_Parser_hpp */
