#include <ESDOM/ESDOM_define.hpp>
#include <ESDOM/API.hpp>

static bool WIN_NT;

ESDOM_EXPORT ESDOM::ESDOM_FILE ESDOM::openFile(const char *filename, int mode)
{
	if( !filename )
		return 0;
	HANDLE	hFile;
	DWORD	dwDesiredAccess = 0;
	DWORD	dwCreationDisposition;
	DWORD	dwFlagsAndAttributes = FILE_FLAG_SEQUENTIAL_SCAN;

	if( mode & FILE_WRITE )
		dwDesiredAccess |= GENERIC_WRITE;
	if( mode & FILE_READ )
		dwDesiredAccess |= GENERIC_READ;

	switch( mode & FILE_OPEN_TYPE )
	{
	case FILE_CREATE_NEW:
		dwCreationDisposition = CREATE_NEW;
		break;
	case FILE_CREATE_ALWAYS:
		dwCreationDisposition = CREATE_ALWAYS;
		break;
	case FILE_OPEN_EXISTING:
		dwCreationDisposition = OPEN_EXISTING;
		break;
//	case FILE_CREATE_TEMPORARY:
//		dwCreationDisposition = CREATE_NEW;
//		dwFlagsAndAttributes |= FILE_ATTRIBUTE_TEMPORARY|FILE_FLAG_DELETE_ON_CLOSE;
//		break;
	default:
		return 0;
	}
	hFile = CreateFileA( filename, dwDesiredAccess, FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL, dwCreationDisposition, dwFlagsAndAttributes, NULL );

	if( hFile == INVALID_HANDLE_VALUE )
	{
		return 0;
	}
	return hFile;
}

ESDOM_EXPORT ESDOM::ESDOM_FILE ESDOM::openFile(const ESDOM_CHAR *filename, int mode)
{
	if( !filename )
		return 0;
	HANDLE	hFile;
	DWORD	dwDesiredAccess = 0;
	DWORD	dwCreationDisposition;
	DWORD	dwFlagsAndAttributes = FILE_FLAG_SEQUENTIAL_SCAN;

	if( mode & FILE_WRITE )
		dwDesiredAccess |= GENERIC_WRITE;
	if( mode & FILE_READ )
		dwDesiredAccess |= GENERIC_READ;

	switch( mode & FILE_OPEN_TYPE )
	{
	case FILE_CREATE_NEW:
		dwCreationDisposition = CREATE_NEW;
		break;
	case FILE_CREATE_ALWAYS:
		dwCreationDisposition = CREATE_ALWAYS;
		break;
	case FILE_OPEN_EXISTING:
		dwCreationDisposition = OPEN_EXISTING;
		break;
//	case FILE_CREATE_TEMPORARY:
//		dwCreationDisposition = CREATE_NEW;
//		dwFlagsAndAttributes |= FILE_ATTRIBUTE_TEMPORARY|FILE_FLAG_DELETE_ON_CLOSE;
//		break;
	default:
		return 0;
	}
	hFile = CreateFileW( filename, dwDesiredAccess, FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL, dwCreationDisposition, dwFlagsAndAttributes, NULL );

	if( hFile == INVALID_HANDLE_VALUE )
	{
		return 0;
	}
	return hFile;
}

ESDOM_EXPORT bool ESDOM::closeFile(ESDOM_FILE FileHandle)
{
	if( WIN_NT )
	{
		typedef BOOL (WINAPI *UNLOCKFILEEX)(HANDLE, DWORD, DWORD, DWORD, LPOVERLAPPED);
		UNLOCKFILEEX	_UnlockFileEx = (UNLOCKFILEEX)GetProcAddress(GetModuleHandleA("KERNEL32.DLL"), "UnlockFileEx");
		OVERLAPPED	overlapped = {0};
		if( !_UnlockFileEx )
			return false;
		_UnlockFileEx(FileHandle, 0, 0, 1, &overlapped);
	}
	else
	{
		UnlockFile(FileHandle, 0, 0, DWORD(-1), 0 );
	}

	return CloseHandle(FileHandle) ? true : false;
}

ESDOM_EXPORT bool ESDOM::lockFile(ESDOM_FILE FileHandle, int mode)
{
	if( WIN_NT )
	{
		if( mode & FILE_LOCK_UN )
		{
			typedef BOOL (WINAPI *UNLOCKFILEEX)(HANDLE, DWORD, DWORD, DWORD, LPOVERLAPPED);
			UNLOCKFILEEX	_UnlockFileEx = (UNLOCKFILEEX)GetProcAddress(GetModuleHandleA("KERNEL32.DLL"), "UnlockFileEx");
			OVERLAPPED	overlapped = {0};
			if( !_UnlockFileEx )
				return false;
			return _UnlockFileEx((HANDLE)FileHandle, 0, 0, 1, &overlapped) ? true : false;
		}
		else
		{
			typedef BOOL (WINAPI *LOCKFILEEX)(HANDLE, DWORD, DWORD, DWORD, DWORD, LPOVERLAPPED);
			DWORD	dwFlags = 0;
			OVERLAPPED	overlapped = {0};
			LOCKFILEEX	_LockFileEx = (LOCKFILEEX)GetProcAddress(GetModuleHandleA("KERNEL32.DLL"), "LockFileEx");
			if( !_LockFileEx )
				return false;
			if( mode & FILE_LOCK_NB )
				dwFlags |= LOCKFILE_FAIL_IMMEDIATELY;
			if( mode & FILE_LOCK_EX )
				dwFlags |= LOCKFILE_EXCLUSIVE_LOCK;
			else if( !( mode & FILE_LOCK_SH ) )
				return true;
			return _LockFileEx((HANDLE)FileHandle, dwFlags, 0, 0, 1, &overlapped) ? true : false;
		}
	}
	else
	{
		if( mode & FILE_LOCK_UN )
			return UnlockFile( FileHandle, 0, 0,DWORD(-1), 0 ) ? true : false;
		return LockFile( FileHandle, 0, 0, DWORD(-1), 0 ) ? true : false;
	}
}

WINAPI DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lp)
{
	if( dwReason == DLL_PROCESS_ATTACH )
	{
		OSVERSIONINFO	osvi={0};
		osvi.dwOSVersionInfoSize = sizeof(osvi);
		GetVersionEx(&osvi);
		if( osvi.dwPlatformId == VER_PLATFORM_WIN32_NT )
			WIN_NT = true;
	}
	return TRUE;
}

#define MakeDllVersion(major, minor, revision, build) ((major<<24) + (minor<<16) + (revision<<8) + build)

extern "C" __declspec(dllexport) DWORD GetDllVersion(void)
{
	return MakeDllVersion(1, 1, 1, 6);
}

