#include <ESDOM/ESDOM_define.hpp>
#include <ESDOM/API.hpp>
#include <ESDOM/ESDOMException.hpp>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <unistd.h>

ESDOM_EXPORT ESDOM::ESDOM_FILE ESDOM::openFile(const char *filename, int mode)
{
	if( !filename )
		return 0;
	int fd;
	int oflag = 0;

	if( mode & FILE_WRITE )
	{
		if( mode & FILE_READ )
			oflag = O_RDWR;
		else
			oflag = O_WRONLY;
	}
	else if( mode & FILE_READ )
		oflag = O_RDONLY;


	switch( mode & FILE_OPEN_TYPE )
	{
	case FILE_CREATE_NEW:
		oflag |= O_CREAT | O_EXCL;
		break;
	case FILE_CREATE_ALWAYS:
		oflag |= O_CREAT;
		break;
	case FILE_OPEN_EXISTING:
		break;
//	case FILE_CREATE_TEMPORARY:
//		oflag |= O_CREAT | O_TEMPORARY;
//		break;
	}

	if( ( fd = open(filename, oflag, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH) ) != -1 )
		return fd;
	return 0;
}

ESDOM_EXPORT ESDOM::ESDOM_FILE ESDOM::openFile(const ESDOM_CHAR *filename, int mode)
{
	if( !filename )
		return 0;

	char *sb_filename;
	int length = strlen(filename);

	try{
		if( !( sb_filename = new char[strlen(filename)*3+1] ) )
			throw 0;
	}catch(...){
		throw ESDOMException(ESDOMException::OUT_OF_MEMORY);
	}
	char *ptr = sb_filename;
	for( ; length; length--, filename++ )
	{
		if( *filename < 0x0080 )
		{
			*ptr++ = (unsigned char)*filename;
			continue;
		} else if( *filename < 0x00000800 ) {
			*ptr++ = 0xc0 | ( 0x1f & ( *filename >> 6 ) );
			*ptr++ = 0x80 | ( 0x3f & *filename );
			continue;
		} else {// if( *filename < 0x00010000 ) {
			*ptr++ = 0xe0 | ( 0x0f & ( *filename >> 12 ) );
			*ptr++ = 0x80 | ( 0x3f & ( *filename >> 6 ) );
			*ptr++ = 0x80 | ( 0x3f & *filename );
			continue;
/*		} else if( *filename < 0x00200000 ) {
			*ptr++ = 0xf0 | ( 0x07 & ( *filename >> 18 ) );
			*ptr++ = 0x80 | ( 0x3f & ( *filename >> 12 ) );
			*ptr++ = 0x80 | ( 0x3f & ( *filename >> 6 ) );
			*ptr++ = 0x80 | ( 0x3f & *filename );
			continue;
		} else if( *filename < 0x04000000 ) {
			*ptr++ = 0xf8 | ( 0x03 & ( *filename >> 24 ) );
			*ptr++ = 0x80 | ( 0x3f & ( *filename >> 18 ) );
			*ptr++ = 0x80 | ( 0x3f & ( *filename >> 12 ) );
			*ptr++ = 0x80 | ( 0x3f & ( *filename >> 6 ) );
			*ptr++ = 0x80 | ( 0x3f & *filename );
			continue;
		} else if( *filename <= 0xffffffff ){
			*ptr++ = 0xfc | ( 0x01 & ( *filename >> 30 ) );
			*ptr++ = 0x80 | ( 0x3f & ( *filename >> 24 ) );
			*ptr++ = 0x80 | ( 0x3f & ( *filename >> 18 ) );
			*ptr++ = 0x80 | ( 0x3f & ( *filename >> 12 ) );
			*ptr++ = 0x80 | ( 0x3f & ( *filename >> 6 ) );
			*ptr++ = 0x80 | ( 0x3f & *filename );
			continue;
*/		}
	}
	*ptr = 0;


	int fd;
	int oflag = 0;

	if( mode & FILE_WRITE )
	{
		if( mode & FILE_READ )
			oflag = O_RDWR;
		else
			oflag = O_WRONLY;
	}
	else if( mode & FILE_READ )
		oflag = O_RDONLY;


	switch( mode & FILE_OPEN_TYPE )
	{
	case FILE_CREATE_NEW:
		oflag |= O_CREAT | O_EXCL;
		break;
	case FILE_CREATE_ALWAYS:
		oflag |= O_CREAT;
		break;
	case FILE_OPEN_EXISTING:
		break;
//	case FILE_CREATE_TEMPORARY:
//		oflag |= O_CREAT | O_TEMPORARY;
//		break;
	}

	if( ( fd = open(sb_filename, oflag, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH) ) != -1 )
	{
		delete[] sb_filename;
		return fd;
	}
	delete[] sb_filename;
	return 0;
}

ESDOM_EXPORT bool ESDOM::closeFile(ESDOM_FILE FileHandle)
{
	return close(FileHandle) ? false : true;
}

ESDOM_EXPORT bool ESDOM::lockFile(ESDOM_FILE FileHandle, int mode)
{
	int lockmode = 0;
	if( mode & FILE_LOCK_NB )
		lockmode |= LOCK_NB;
	if( mode & FILE_LOCK_UN )
		lockmode |= LOCK_UN;
	if( mode & FILE_LOCK_EX )
		lockmode |= LOCK_EX;
	if( mode & FILE_LOCK_SH )
		lockmode |= LOCK_SH;
	return flock( FileHandle, lockmode ) ? false : true;
}
