#include "Serializer.hpp"
#include "DocumentInst.hpp"
#include <ESDOM/NodeList.hpp>
#include <ESDOM/NamedNodeMap.hpp>
#include <stdio.h>

using namespace ESDOM;

bool Serializer::serialize(const Document &document, unsigned short encoding, ESDOM_FILE FileHandle)
{
	if( !FileHandle )
		return false;
	if( ESDOM_isRegFile(FileHandle) )
	{
		ESDOM_seek(FileHandle, 0, FILE_SEEK_SET);
		ESDOM_truncate(FileHandle);
	}
	m_document = document;
	m_encoding = encoding;
	m_dst = (void*)FileHandle;

	unsigned char bom[3];
	switch( encoding )
	{
	case UTF_16BE:
		bom[0] = 0xfe;
		bom[1] = 0xff;
		write(bom, 2 );
		break;
	case UTF_16LE:
		bom[0] = 0xff;
		bom[1] = 0xfe;
		write(bom, 2 );
		break;
	case UTF_16BE_NONBOM:
		m_encoding = UTF_16BE;
		break;
	case UTF_16LE_NONBOM:
		m_encoding = UTF_16LE;
		break;
	}
	return serialize();
}

bool Serializer::serialize()
{
	if( !m_document.m_instance || !((DocumentInst*)m_document.m_instance)->isValid() )
		return false;


	try
	{
		bool fHTML;
		DocumentType doctype = m_document.getDoctype();
		if( doctype && doctype.getNodeName().index( _ESDOM_TEXT("html") ) == 0 )
			fHTML = true;
		else
			fHTML = false;

		writeChars(_ESDOM_TEXT("<?xml version=\"") );
		writeChars(((DocumentInst*)m_document.m_instance)->getXmlVersion() );
		writeChars(_ESDOM_TEXT("\" encoding=\"") );
		if( UTF_16BE <= m_encoding && m_encoding <= UTF_16LE_NONBOM )
			writeChars(_ESDOM_TEXT("UTF-16\"") );
		else
			writeChars(_ESDOM_TEXT("UTF-8\"") );
		if( ((DocumentInst*)m_document.m_instance)->m_standalone > 0 )
			writeChars(_ESDOM_TEXT(" standalone=\"yes\"") );
		if( ((DocumentInst*)m_document.m_instance)->m_standalone == 0 )
			writeChars(_ESDOM_TEXT(" standalone=\"no\"") );
		writeChars(_ESDOM_TEXT("?>\n") );

		for( unsigned long i = 0, length = m_document.getChildNodes().getLength(); i < length; i++ )
		{
			Node node = m_document.getChildNodes().item(i);
			ESDOMString		buffer;
			bool			fSpace;
			unsigned long	pos;
			switch( node.getNodeType() )
			{
			case Node::PROCESSING_INSTRUCTION_NODE:
				writeChars(_ESDOM_TEXT("<?") );
				writeChars(node.getNodeName() );
				writeChars(_ESDOM_TEXT(" ") );
				buffer = node.getNodeValue();
				while( ( pos = buffer.index( _ESDOM_TEXT("?>") ) ) != (unsigned long)-1 )
				{
					writeChars(buffer.substr(0, pos+1 ) );
					writeChars(_ESDOM_TEXT(" ") );
					buffer = buffer.substr( pos+1 );
				}
				writeChars(_ESDOM_TEXT("?>\n") );
				break;
			case Node::DOCUMENT_TYPE_NODE:
				writeChars(_ESDOM_TEXT("<!DOCTYPE ") );
				writeChars(node.getNodeName() );
				writeChars(_ESDOM_TEXT(">\n") );
				break;
			case Node::COMMENT_NODE:
				writeChars(_ESDOM_TEXT("<!--") );
				buffer = node.getNodeValue();
				if( buffer[buffer.getLength()-1] == '-' )
					fSpace = true;
				else
					fSpace = false;
				while( ( pos = buffer.index( _ESDOM_TEXT("--") ) ) != (unsigned long)-1 )
				{
					writeChars(buffer.substr(0, pos+1 ) );
					writeChars(_ESDOM_TEXT(" ") );
					buffer = buffer.substr( pos+1 );
				}
				writeChars(buffer );
				if( fSpace )
					writeChars(_ESDOM_TEXT(" -->") );
				else
					writeChars(_ESDOM_TEXT("-->") );
				break;
			case Node::ELEMENT_NODE:
				serializeElement( node, fHTML );
				break;
			}
		}
		return true;
	}
	catch(...)
	{
		return false;
	}
}

void Serializer::serializeElement(const Element &element, bool fHTML)
{
	ESDOMString tagName = element.getNodeName();

	writeChars(_ESDOM_TEXT("<") );
	writeChars(tagName );

	NamedNodeMap attributes = element.getAttributes();
	unsigned long i, length;
	for( i = 0, length = attributes.getLength(); i < length; i++ )
	{
		writeChars(_ESDOM_TEXT(" ") );
		writeChars(attributes.item(i).getNodeName() );
		writeChars(_ESDOM_TEXT("=\"") );
		writeChars(char2entityReference( attributes.item(i).getNodeValue(), fHTML ) );
		writeChars(_ESDOM_TEXT("\"") );
	}
	if( !element.hasChildNodes() )
	{
		if( fHTML )
		{
			if( tagName == _ESDOM_TEXT("area") || tagName == _ESDOM_TEXT("base") || tagName == _ESDOM_TEXT("br")
				|| tagName == _ESDOM_TEXT("col") || tagName == _ESDOM_TEXT("frame") || tagName == _ESDOM_TEXT("hr")
				|| tagName == _ESDOM_TEXT("img") || tagName == _ESDOM_TEXT("input") || tagName == _ESDOM_TEXT("link")
				|| tagName == _ESDOM_TEXT("meta") )
				writeChars(_ESDOM_TEXT(" />") );
			else
			{
				writeChars(_ESDOM_TEXT("></") );
				writeChars(tagName );
				writeChars(_ESDOM_TEXT(">") );
			}
		}
		else
		{
			writeChars(_ESDOM_TEXT("/>") );
		}
		return;
	}
	writeChars(_ESDOM_TEXT(">") );

	NodeList childNodes = element.getChildNodes();
	for( i = 0; i < childNodes.getLength(); i++ )
	{
		bool			fSpace = false;
		unsigned long	pos;
		ESDOMString		buffer;

		switch( childNodes.item(i).getNodeType() )
		{
		case Node::ELEMENT_NODE:
			serializeElement( childNodes.item(i), fHTML );
			break;
		case Node::TEXT_NODE:
			writeChars(char2entityReference( childNodes.item(i).getNodeValue(), fHTML ) );
			break;
		case Node::COMMENT_NODE:
			writeChars(_ESDOM_TEXT("<!--") );
			buffer = childNodes.item(i).getNodeValue();
			if( buffer[buffer.getLength()-1] == '-' )
				fSpace = true;
			while( ( pos = buffer.index( _ESDOM_TEXT("--") ) ) != (unsigned long)-1 )
			{
				writeChars(buffer.substr(0, pos+1 ) );
				writeChars(_ESDOM_TEXT(" ") );
				buffer = buffer.substr( pos+1 );
			}
			writeChars(buffer );
			if( fSpace )
				writeChars(_ESDOM_TEXT(" -->") );
			else
				writeChars(_ESDOM_TEXT("-->") );
			break;
		case Node::CDATA_SECTION_NODE:
			writeChars(_ESDOM_TEXT("<![CDATA[") );
			buffer = childNodes.item(i).getNodeValue();
			while( ( pos = buffer.index( _ESDOM_TEXT("]]>") ) ) != (unsigned long)-1 )
			{
				writeChars(buffer.substr(0, pos+2 ) );
				writeChars(_ESDOM_TEXT(" >") );
				buffer = buffer.substr( pos+3 );
			}
			writeChars(buffer );
			writeChars(_ESDOM_TEXT("]]>") );
			break;
		}
	}
	writeChars(_ESDOM_TEXT("</") );
	writeChars(tagName );
	writeChars(_ESDOM_TEXT(">") );
}

ESDOMString Serializer::char2entityReference(const ESDOMString &string, bool fHTML)
{
	ESDOMString returnString;
	const ESDOM_CHAR *ptr = string;

	unsigned long prev = 0, current = 0;
	for( ; ptr[current]; current++ )
	{
		switch( ptr[current] )
		{
		case '&':
			if( current > prev )
				returnString += string.substr( prev, current-prev );
			returnString += "&amp;";
			prev = current +1;
			break;
		case '<':
			if( current > prev )
				returnString += string.substr( prev, current-prev );
			returnString += "&lt;";
			prev = current +1;
			break;
		case '>':
			if( current > prev )
				returnString += string.substr( prev, current-prev );
			returnString += "&gt;";
			prev = current +1;
			break;
		case '"':
			if( current > prev )
				returnString += string.substr( prev, current-prev );
			returnString += "&quot;";
			prev = current +1;
			break;
		case '\'':
			if( fHTML )
				break;
			if( current > prev )
				returnString += string.substr( prev, current-prev );
			returnString += "&apos;";
			prev = current +1;
			break;
		}
	}
	if( current > prev )
		returnString += string.substr( prev, current-prev );

	return returnString;
}

void Serializer::writeChars( const ESDOM_CHAR *string, unsigned long length )
{
	unsigned char	chars[3];
//	unsigned char	chars[6]; // for UCS4

	switch(m_encoding)
	{
	case UTF_8:
		for( ; length; length--, string++ )
		{
			if( *string < 0x0080 )
			{
				chars[0] = (unsigned char)*string;
				write( chars, 1 );
				continue;
			} else if( *string < 0x00000800 ) {
				chars[0] = (unsigned char)(0xc0 | ( 0x1f & ( *string >> 6 ) ));
				chars[1] = (unsigned char)(0x80 | ( 0x3f & *string ));
				write( chars, 2 );
				continue;
			} else if( *string < 0x00010000 ) {
				chars[0] = (unsigned char)(0xe0 | ( 0x0f & ( *string >> 12 ) ));
				chars[1] = (unsigned char)(0x80 | ( 0x3f & ( *string >> 6 ) ));
				chars[2] = (unsigned char)(0x80 | ( 0x3f & *string ));
				write( chars, 3 );
				continue;
/*			} else if( *string < 0x00200000 ) {
				chars[0] = (unsigned char)(0xf0 | ( 0x07 & ( *string >> 18 ) ));
				chars[1] = (unsigned char)(0x80 | ( 0x3f & ( *string >> 12 ) ));
				chars[2] = (unsigned char)(0x80 | ( 0x3f & ( *string >> 6 ) ));
				chars[3] = (unsigned char)(0x80 | ( 0x3f & *string ));
				write( chars, 4 );
				continue;
			} else if( *string < 0x04000000 ) {
				chars[0] = (unsigned char)(0xf8 | ( 0x03 & ( *string >> 24 ) ));
				chars[1] = (unsigned char)(0x80 | ( 0x3f & ( *string >> 18 ) ));
				chars[2] = (unsigned char)(0x80 | ( 0x3f & ( *string >> 12 ) ));
				chars[3] = (unsigned char)(0x80 | ( 0x3f & ( *string >> 6 ) ));
				chars[4] = (unsigned char)(0x80 | ( 0x3f & *string ));
				write( chars, 5 );
				continue;
			} else if( *string <= 0xffffffff ){
				chars[0] = (unsigned char)(0xfc | ( 0x01 & ( *string >> 30 ) ));
				chars[1] = (unsigned char)(0x80 | ( 0x3f & ( *string >> 24 ) ));
				chars[2] = (unsigned char)(0x80 | ( 0x3f & ( *string >> 18 ) ));
				chars[3] = (unsigned char)(0x80 | ( 0x3f & ( *string >> 12 ) ));
				chars[4] = (unsigned char)(0x80 | ( 0x3f & ( *string >> 6 ) ));
				chars[5] = (unsigned char)(0x80 | ( 0x3f & *string ));
				!write( chars, 6 );
				continue;
*/			}
			throw 0;
		}
		return;

	case UTF_16BE:
		for( ; length; length--, string++ )
		{
			chars[0] = (unsigned char)(*string / 0x0100);
			chars[1] = (unsigned char)(*string % 0x0100);
			write( chars, 2 );
		}
		return;

	case UTF_16LE:
		for( ; length; length--, string++ )
		{
			chars[0] = (unsigned char)(*string % 0x0100);
			chars[1] = (unsigned char)(*string / 0x0100);
			write( chars, 2 );
		}
		return;
	}
	throw 0;
}

void Serializer::write(unsigned char *buffer, unsigned long size)
{
	if( m_dst == stdout || m_dst == stderr )
	{
		if( fwrite( buffer, 1, size, (FILE*)m_dst ) == size )
			return;
		throw 0;
	}
	if( writeToFile( (ESDOM_FILE)m_dst, buffer, size ) == size )
		return;
	throw 0;
}
